@@ -738,8 +738,8 @@ dapls_evd_post_async_error_event(IN DAPL_EVD * evd_ptr,
 {
 	DAT_EVENT *event_ptr;
 
-	dapl_log(DAPL_DBG_TYPE_WARN,
-		 " WARNING: async event - %s evd=%p\n",
+	dapl_log(DAPL_DBG_TYPE_EXCEPTION,
+		 " async event - %s evd=%p\n",
 		 dapl_event_str(event_number), evd_ptr);
 
 	dapl_os_lock(&evd_ptr->header.lock);
@@ -975,7 +975,7 @@ dapls_evd_post_cr_event_ext(IN DAPL_SP * sp_ptr,
 		 * requestor that we cant help them.
 		 */
 		ia_ptr = sp_ptr->header.owner_ia;
-		ep_ptr = dapl_ep_alloc(ia_ptr, NULL);
+		ep_ptr = dapl_ep_alloc(ia_ptr, NULL, DAT_FALSE);
 		if (ep_ptr == NULL) {
 			dapls_cr_free(cr_ptr);
 			goto bail;
@@ -1086,6 +1086,7 @@ dapli_evd_cqe_to_event(IN DAPL_EVD * evd_ptr,
 	DAPL_COOKIE *cookie;
 	DAT_DTO_COMPLETION_STATUS dto_status;
 	DAPL_COOKIE_BUFFER *buffer;
+	DAPL_SRQ *srq_ptr = NULL;
 
 	/*
 	 * All that can be relied on if the status is bad is the status
@@ -1096,6 +1097,7 @@ dapli_evd_cqe_to_event(IN DAPL_EVD * evd_ptr,
 	cookie = (DAPL_COOKIE *) (uintptr_t) DAPL_GET_CQE_WRID(cqe_ptr);
 	dapl_os_assert((NULL != cookie));
 
+	/* In case of RECV SRQ DTO cookie->ep holds pointer to the SRQ */
 	ep_ptr = cookie->ep;
 	dapl_os_assert((NULL != ep_ptr));
 
@@ -1125,8 +1127,13 @@ dapli_evd_cqe_to_event(IN DAPL_EVD * evd_ptr,
 				break;
 			}
 #endif
-
-			if (DAPL_DTO_TYPE_RECV == cookie->val.dto.type)
+			if (DAPL_DTO_TYPE_RECV_SRQ == cookie->val.dto.type) {
+				/* in SRQ event we do NOT have ep pointer */
+				srq_ptr = (DAPL_SRQ *)ep_ptr;
+				buffer = &srq_ptr->recv_buffer;
+				ep_ptr = NULL;
+			}
+			else if (DAPL_DTO_TYPE_RECV == cookie->val.dto.type)
 				buffer = &ep_ptr->recv_buffer;
 			else
 				buffer = &ep_ptr->req_buffer;
@@ -1151,6 +1158,8 @@ dapli_evd_cqe_to_event(IN DAPL_EVD * evd_ptr,
 			}
 
 			dapls_cookie_dealloc(buffer, cookie);
+			if (srq_ptr)
+				dapl_os_atomic_dec(&srq_ptr->recv_count);
 			break;
 		}
 
@@ -1198,6 +1207,36 @@ dapli_evd_cqe_to_event(IN DAPL_EVD * evd_ptr,
 	    (dto_status != DAT_DTO_ERR_FLUSHED)) {
 		DAPL_EVD *evd_ptr;
 
+		/* In SRQ case we need to look up the EP */
+		if (ep_ptr == NULL) {
+			DAT_UINT32 qp_num = DAPL_GET_CQE_QP_NUM(cqe_ptr);
+			DAPL_IA *ia_ptr = (DAPL_IA *)srq_ptr->param.ia_handle;
+
+			dapl_os_lock(&ia_ptr->header.lock);
+			ep_ptr = (dapl_llist_is_empty(&ia_ptr->ep_list_head)
+				   ? NULL :
+				   dapl_llist_peek_head(&ia_ptr->ep_list_head));
+
+			while (ep_ptr != NULL) {
+				if (ep_ptr->qp_handle->qp->qp_num == qp_num)
+					break;
+
+				ep_ptr =
+				    dapl_llist_next_entry(&ia_ptr->ep_list_head,
+								&ep_ptr->header.
+								ia_list_entry);
+			}
+			dapl_os_unlock(&ia_ptr->header.lock);
+
+			if (ep_ptr == NULL ) {
+				dapl_dbg_log(DAPL_DBG_TYPE_DTO_COMP_ERR,
+					     " SRQ %p at ia %p - "
+					     "EP with qpn %d was not found\n",
+					     srq_ptr, ia_ptr, qp_num);
+				return;
+			}
+		}
+
 		/*
 		 * If we are connected, generate disconnect and generate an
 		 * event. We may be racing with other disconnect ops, so we
