@@ -479,6 +479,7 @@ struct dapl_srq
     DAPL_ATOMIC		srq_ref_count;
     DAPL_COOKIE_BUFFER	recv_buffer;
     DAPL_ATOMIC		recv_count;
+    ib_srq_handle_t	srq_handle;
 };
 
 /* DAPL_PZ maps to DAT_PZ_HANDLE */
@@ -580,6 +581,7 @@ typedef enum dapl_dto_type
 {
     DAPL_DTO_TYPE_SEND,
     DAPL_DTO_TYPE_RECV,
+    DAPL_DTO_TYPE_RECV_SRQ,
     DAPL_DTO_TYPE_RDMA_WRITE,
     DAPL_DTO_TYPE_RDMA_READ,
 #ifdef DAT_EXTENSIONS
