@@ -344,14 +344,16 @@ DAT_RETURN dapls_ib_open_hca(IN IB_HCA_NAME hca_name,
 	}
 
 	/* set inline max with env or default, get local lid and gid 0 */
-	if (hca_ptr->ib_hca_handle->device->transport_type == IBV_TRANSPORT_IWARP)
-		hca_ptr->ib_trans.max_inline_send =
-		    dapl_os_get_env_val("DAPL_MAX_INLINE",
-					INLINE_SEND_IWARP_DEFAULT);
-	else
-		hca_ptr->ib_trans.max_inline_send =
-		    dapl_os_get_env_val("DAPL_MAX_INLINE",
-					INLINE_SEND_IB_DEFAULT);
+	if (dapl_ib_inline_data(hca_ptr->ib_hca_handle)) {
+		if (hca_ptr->ib_hca_handle->device->transport_type == IBV_TRANSPORT_IWARP)
+			hca_ptr->ib_trans.ib_cm.max_inline =
+			    dapl_os_get_env_val("DAPL_MAX_INLINE",
+						INLINE_SEND_IWARP_DEFAULT);
+		else
+			hca_ptr->ib_trans.ib_cm.max_inline =
+			    dapl_os_get_env_val("DAPL_MAX_INLINE",
+						INLINE_SEND_IB_DEFAULT);
+	}
 
 	/* set CM timer defaults */
 	hca_ptr->ib_trans.max_cm_timeout =
@@ -393,7 +395,7 @@ DAT_RETURN dapls_ib_open_hca(IN IB_HCA_NAME hca_name,
 
 done:
 	/* set default IB MTU */
-	hca_ptr->ib_trans.mtu = dapl_ib_mtu(2048);
+	hca_ptr->ib_trans.ib_cm.mtu = dapl_ib_mtu(2048);
 
 	return DAT_SUCCESS;
 }
