@@ -57,6 +57,7 @@
 #define DAT_MCM_UD_QKEY		0x78655322
 #define DAT_MCM_PDATA_SIZE      64
 #define DAT_MCM_PROXY_DATA	40
+#define DAT_MCM_SEG_PO2		17
 
 #define ALIGN_64(o)  ((o + 64 - 1) & ~(64-1))
 #define ALIGN_P64(o) ((((uintptr_t)o) + 64 - 1)& ~(64-1))
@@ -185,11 +186,14 @@ static inline char * mcm_ib_async_str(IN int st)
 #define HOST_SOCK_DEV  1  /* host to HCA, any socket */
 #define MIC_SSOCK_DEV  2  /* MIC to HCA, same socket */
 #define MIC_XSOCK_DEV  3  /* MIC to HCA, cross socket */
+#define MIC_FULL_DEV   4  /* MIC to HCA, full proxy offload, no direct verbs */
 
 #define UND_EP(x) ((x)->ep_map <  1 || (x)->ep_map > 4)
 #define HST_EP(x) ((x)->ep_map == HOST_SOCK_DEV)
-#define MXS_EP(x) ((x)->ep_map == MIC_XSOCK_DEV)
 #define MSS_EP(x) ((x)->ep_map == MIC_SSOCK_DEV)
+#define MXS_EP(x) ((x)->ep_map == MIC_XSOCK_DEV)
+#define MFO_EP(x) ((x)->ep_map == MIC_FULL_DEV)
+#define MXF_EP(x) ((x)->ep_map == MIC_XSOCK_DEV || (x)->ep_map == MIC_FULL_DEV)
 
 static inline char * mcm_map_str(IN uint8_t ep_map)
 {
@@ -198,8 +202,9 @@ static inline char * mcm_map_str(IN uint8_t ep_map)
 		"HST",
 		"MSS",
 		"MXS",
+		"MFO"
 	};
-	return ((ep_map < 1 || ep_map > 3) ? "???" : map[ep_map]);
+	return ((ep_map < 1 || ep_map > 4) ? "???" : map[ep_map]);
 }
 
 /* MCM address, 28 bytes */
@@ -910,7 +915,11 @@ static inline void mcm_hton_wc_rx(struct mcm_wc_rx *m_wc_rx, struct mcm_wr_rx *m
 	m_wc_rx->wc.wr_id = htonll(m_wr_rx->org_id);
 	m_wc_rx->wc.status = htonl(status);
 	m_wc_rx->wc.byte_len = htonl(m_wr_rx->sg[0].length);
-	if (m_wr_rx->wr.send_flags & IBV_WR_RDMA_WRITE)
+	if (m_wr_rx->wr.opcode == IBV_WR_RDMA_WRITE_WITH_IMM) {
+		m_wc_rx->wc.wc_flags = htonl(IBV_WC_WITH_IMM);
+		m_wc_rx->wc.opcode = htonl(IBV_WC_RDMA_WRITE);
+	}
+	else if (m_wr_rx->wr.opcode == IBV_WR_RDMA_WRITE)
 		m_wc_rx->wc.opcode = htonl(IBV_WC_RDMA_WRITE);
 	else
 		m_wc_rx->wc.opcode = htonl(IBV_WC_SEND);
@@ -926,6 +935,7 @@ static inline void mcm_ntoh_wc_rx(struct mcm_wc_rx *m_wc_rx)
 	m_wc_rx->wc.status = ntohl(m_wc_rx->wc.status);
 	m_wc_rx->wc.byte_len = ntohl(m_wc_rx->wc.byte_len);
 	m_wc_rx->wc.opcode = ntohl(m_wc_rx->wc.opcode);
+	m_wc_rx->wc.wc_flags = ntohl(m_wc_rx->wc.wc_flags);
 }
 
 static inline void mcm_const_mix_wr(struct dat_mix_wr *mwr, struct ibv_send_wr *iwr)
