@@ -61,10 +61,12 @@ dapls_ib_qp_alloc(IN DAPL_IA * ia_ptr,
 {
 	DAT_EP_ATTR *attr;
 	DAPL_EVD *rcv_evd, *req_evd;
+	DAPL_SRQ *dapl_srq = (DAPL_SRQ *)ep_ptr->param.srq_handle;
 	ib_cq_handle_t rcv_cq, req_cq;
 	ib_pd_handle_t ib_pd_handle;
+	ib_srq_handle_t ib_srq_handle = NULL;
 	int ret = EINVAL;
-	int max_inline = ia_ptr->hca_ptr->ib_trans.max_inline_send;
+	int max_inline = ia_ptr->hca_ptr->ib_trans.ib_cm.max_inline;
 	struct ibv_qp_init_attr qp_create;
 #ifdef _OPENIB_CMA_
 	dp_ib_cm_handle_t conn;
@@ -72,6 +74,8 @@ dapls_ib_qp_alloc(IN DAPL_IA * ia_ptr,
 
 	attr = &ep_ptr->param.ep_attr;
 	ib_pd_handle = ((DAPL_PZ *) ep_ptr->param.pz_handle)->pd_handle;
+	if (dapl_srq)
+		ib_srq_handle = dapl_srq->srq_handle;
 	rcv_evd = (DAPL_EVD *) ep_ptr->param.recv_evd_handle;
 	req_evd = (DAPL_EVD *) ep_ptr->param.request_evd_handle;
 
@@ -167,6 +171,7 @@ dapls_ib_qp_alloc(IN DAPL_IA * ia_ptr,
 	qp_create.cap.max_inline_data = max_inline;
 	qp_create.qp_type = IBV_QPT_RC;
 	qp_create.qp_context = (void *)ep_ptr;
+	qp_create.srq = ib_srq_handle;
 
 #ifdef DAT_EXTENSIONS 
 	if ((int)attr->service_type == (int)DAT_IB_SERVICE_TYPE_UD) {
@@ -175,9 +180,16 @@ dapls_ib_qp_alloc(IN DAPL_IA * ia_ptr,
 #endif
 		qp_create.qp_type = IBV_QPT_UD;
 		if (attr->max_message_size >
-		    (128 << ia_ptr->hca_ptr->ib_trans.mtu)) {
+		    (128 << ia_ptr->hca_ptr->ib_trans.ib_cm.mtu)) {
 			goto err;
 		}
+		ep_ptr->qp_handle->ah =
+			(ib_ah_handle_t*) dapl_os_alloc(sizeof(ib_ah_handle_t) * DCM_AH_SPACE);
+
+		if (!ep_ptr->qp_handle->ah)
+			goto err;
+
+		dapl_os_memzero(ep_ptr->qp_handle->ah, sizeof(ib_ah_handle_t) * DCM_AH_SPACE);
 	}
 #endif
 
@@ -220,9 +232,15 @@ dapls_ib_qp_alloc(IN DAPL_IA * ia_ptr,
 #endif
 	ep_ptr->qp_handle->qp =	ibv_create_qp(ib_pd_handle, &qp_create);
 	if (!ep_ptr->qp_handle->qp) {
-		dapl_log(1," qp_alloc ERR %d %s line %d on device %s\n",
-			   errno, strerror(errno), __LINE__ ,
+		dapl_log(1," qp_alloc ERR %d %s on device %s\n",
+			   errno, strerror(errno),
 			   ibv_get_device_name(ia_ptr->hca_ptr->ib_trans.ib_dev));
+		dapl_log(1," qp_attr: SQ %d,%d cq %p RQ %d,%d cq %p SRQ %p [inl %d typ %d]\n",
+			    qp_create.cap.max_send_wr, qp_create.cap.max_send_sge,
+			    req_cq, qp_create.cap.max_recv_wr,
+			    qp_create.cap.max_recv_sge, rcv_cq,
+			    qp_create.srq, qp_create.cap.max_inline_data,
+			    qp_create.qp_type);
 		ret = errno;
 		goto err;
 	}
@@ -263,7 +281,7 @@ skip_qp:
 	} else {
 		/* NON-MIC: need QPt, in case of shadowed QP's on remote MIC's */
 		/* Prep for HST -> MXS: xfers via remote PI instead of direct */
-		ia_ptr->hca_ptr->ib_trans.max_inline_send =
+		ia_ptr->hca_ptr->ib_trans.ib_cm.max_inline =
 				DAPL_MAX(sizeof(struct mcm_wr_rx), max_inline);
 		/* create CQ for peer PI, HST->MXS case */
 		if (mcm_create_pi_cq(ep_ptr->qp_handle, MCM_WRC_QLEN))
@@ -274,7 +292,7 @@ skip_qp:
 		qp_create.cap.max_recv_sge = 1;
 		qp_create.cap.max_send_wr = DAPL_MAX(MCM_WRC_QLEN, attr->max_request_dtos);
 		qp_create.cap.max_send_sge = attr->max_request_iov;
-		qp_create.cap.max_inline_data = ia_ptr->hca_ptr->ib_trans.max_inline_send;
+		qp_create.cap.max_inline_data = ia_ptr->hca_ptr->ib_trans.ib_cm.max_inline;
 
 		ep_ptr->qp_handle->qp2 = ibv_create_qp(ib_pd_handle, &qp_create);
 		if (!ep_ptr->qp_handle->qp2) {
@@ -325,6 +343,10 @@ err:
 		if (ep_ptr->qp_handle->qp2)
 			ibv_destroy_qp(ep_ptr->qp_handle->qp2);
 #endif
+		if (ep_ptr->qp_handle->ah)
+			dapl_os_free(ep_ptr->qp_handle->ah,
+				     sizeof(ib_ah_handle_t) * DCM_AH_SPACE);
+
 		dapl_os_free(ep_ptr->qp_handle, sizeof(struct dcm_ib_qp));
 	}
 	ep_ptr->qp_handle = IB_INVALID_HANDLE;
@@ -400,6 +422,17 @@ DAT_RETURN dapls_ib_qp_free(IN DAPL_IA * ia_ptr, IN DAPL_EP * ep_ptr)
 		dapl_os_unlock(&ep_ptr->header.lock);
 	}
 
+	/* destroy any UD address handles */
+	if (ep_ptr->qp_handle->ah) {
+		int i;
+
+		for (i = 0; i < DCM_AH_SPACE; i++) {
+			if (ep_ptr->qp_handle->ah[i])
+				ibv_destroy_ah(ep_ptr->qp_handle->ah[i]);
+		}
+		dapl_os_free(ep_ptr->qp_handle->ah, sizeof(ib_ah_handle_t) * DCM_AH_SPACE);
+	}
+
 	dapl_os_free(ep_ptr->qp_handle, sizeof(struct dcm_ib_qp));
 	ep_ptr->qp_handle = IB_INVALID_HANDLE;
 	return DAT_SUCCESS;
@@ -574,7 +607,7 @@ out:
 #else
 uint8_t dapls_get_sl(DAPL_HCA *hca_ptr, uint16_t dlid)
 {
-	return hca_ptr->ib_trans.sl;
+	return hca_ptr->ib_trans.ib_cm.sl;
 }
 #endif
 
@@ -609,10 +642,10 @@ dapls_modify_qp_state(IN struct ibv_qp		*qp_handle,
 
 		qp_attr.dest_qp_num = ntohl(qpn);
 		qp_attr.rq_psn = 1;
-		qp_attr.path_mtu = ia_ptr->hca_ptr->ib_trans.mtu;
-		qp_attr.min_rnr_timer =	ia_ptr->hca_ptr->ib_trans.rnr_timer;
+		qp_attr.path_mtu = ia_ptr->hca_ptr->ib_trans.ib_cm.mtu;
+		qp_attr.min_rnr_timer =	ia_ptr->hca_ptr->ib_trans.ib_cm.rnr_timer;
 #ifdef _OPENIB_MCM_
-		qp_attr.max_dest_rd_atomic = ia_ptr->hca_ptr->ib_trans.rd_atom_in;
+		qp_attr.max_dest_rd_atomic = ia_ptr->hca_ptr->ib_trans.ib_cm.rd_atom_in;
 #else
 		qp_attr.max_dest_rd_atomic = ep_ptr->param.ep_attr.max_rdma_read_in;
 #endif
@@ -625,7 +658,7 @@ dapls_modify_qp_state(IN struct ibv_qp		*qp_handle,
 
 		/* address handle. RC and UD */
 		qp_attr.ah_attr.dlid = ntohs(lid);
-		if (gid && ia_ptr->hca_ptr->ib_trans.global) {
+		if (gid && ia_ptr->hca_ptr->ib_trans.ib_cm.global) {
 			dapl_dbg_log(DAPL_DBG_TYPE_EP, 
 				     " QPS_RTR: GID Subnet 0x" F64x " ID 0x" F64x "\n", 
 				     (unsigned long long)htonll(gid->global.subnet_prefix),
@@ -637,9 +670,9 @@ dapls_modify_qp_state(IN struct ibv_qp		*qp_handle,
 			qp_attr.ah_attr.grh.dgid.global.interface_id = 
 				gid->global.interface_id;
 			qp_attr.ah_attr.grh.hop_limit =
-				ia_ptr->hca_ptr->ib_trans.hop_limit;
+				ia_ptr->hca_ptr->ib_trans.ib_cm.hop_limit;
 			qp_attr.ah_attr.grh.traffic_class =
-				ia_ptr->hca_ptr->ib_trans.tclass;
+				ia_ptr->hca_ptr->ib_trans.ib_cm.tclass;
 		}
 		qp_attr.ah_attr.sl = dapls_get_sl(ia_ptr->hca_ptr, lid);
 		qp_attr.ah_attr.src_path_bits = 0;
@@ -660,13 +693,13 @@ dapls_modify_qp_state(IN struct ibv_qp		*qp_handle,
 				IBV_QP_RETRY_CNT |
 				IBV_QP_RNR_RETRY | IBV_QP_MAX_QP_RD_ATOMIC;
 			qp_attr.timeout =
-				ia_ptr->hca_ptr->ib_trans.ack_timer;
+				ia_ptr->hca_ptr->ib_trans.ib_cm.ack_timer;
 			qp_attr.retry_cnt =
-				ia_ptr->hca_ptr->ib_trans.ack_retry;
+				ia_ptr->hca_ptr->ib_trans.ib_cm.ack_retry;
 			qp_attr.rnr_retry =
-				ia_ptr->hca_ptr->ib_trans.rnr_retry;
+				ia_ptr->hca_ptr->ib_trans.ib_cm.rnr_retry;
 #ifdef _OPENIB_MCM_
-			qp_attr.max_rd_atomic = ia_ptr->hca_ptr->ib_trans.rd_atom_out;
+			qp_attr.max_rd_atomic = ia_ptr->hca_ptr->ib_trans.ib_cm.rd_atom_out;
 #else
 			qp_attr.max_rd_atomic = ep_ptr->param.ep_attr.max_rdma_read_out;
 #endif
@@ -711,7 +744,7 @@ dapls_modify_qp_state(IN struct ibv_qp		*qp_handle,
 			qp_attr.qkey = DAT_UD_QKEY;
 		}
 
-		qp_attr.pkey_index = ia_ptr->hca_ptr->ib_trans.pkey_idx;
+		qp_attr.pkey_index = ia_ptr->hca_ptr->ib_trans.ib_cm.pkey_idx;
 		qp_attr.port_num = ia_ptr->hca_ptr->port_num;
 
 		dapl_dbg_log(DAPL_DBG_TYPE_EP,
@@ -732,7 +765,7 @@ dapls_modify_qp_state(IN struct ibv_qp		*qp_handle,
 			" modify_qp_state: ERR type %d qpn 0x%x gid %p (%d) lid 0x%x"
 			" port %d state %d mtu %d rd %d rnr %d sl %d\n",
 			qp_handle->qp_type, ntohl(qpn), gid,
-			ia_ptr->hca_ptr->ib_trans.global,
+			ia_ptr->hca_ptr->ib_trans.ib_cm.global,
 			ntohs(lid), ia_ptr->hca_ptr->port_num,
 			ep_ptr->qp_state,
 			qp_attr.path_mtu, qp_attr.max_dest_rd_atomic,
@@ -751,7 +784,7 @@ dapls_modify_qp_ud(IN DAPL_HCA *hca, IN struct ibv_qp *qp)
 	/* modify QP, setup and prepost buffers */
 	dapl_os_memzero((void *)&qp_attr, sizeof(qp_attr));
 	qp_attr.qp_state = IBV_QPS_INIT;
-        qp_attr.pkey_index = hca->ib_trans.pkey_idx;
+        qp_attr.pkey_index = hca->ib_trans.ib_cm.pkey_idx;
         qp_attr.port_num = hca->port_num;
 #ifdef _OPENIB_MCM_
         qp_attr.qkey = DAT_MCM_UD_QKEY; /* MCM gets different key */
@@ -815,8 +848,8 @@ dapls_create_ah(IN DAPL_HCA		*hca,
 				ntohll(gid->global.subnet_prefix);
 		qp_attr.ah_attr.grh.dgid.global.interface_id = 
 				ntohll(gid->global.interface_id);
-		qp_attr.ah_attr.grh.hop_limit =	hca->ib_trans.hop_limit;
-		qp_attr.ah_attr.grh.traffic_class = hca->ib_trans.tclass;
+		qp_attr.ah_attr.grh.hop_limit =	hca->ib_trans.ib_cm.hop_limit;
+		qp_attr.ah_attr.grh.traffic_class = hca->ib_trans.ib_cm.tclass;
 	}
 	qp_attr.ah_attr.sl = dapls_get_sl(hca, lid);
 	qp_attr.ah_attr.src_path_bits = 0;
