@@ -231,6 +231,46 @@ retry:
 }
 #endif
 
+#ifdef _OPENIB_MCM_
+static int cpuinfo_atoi(const char *v_str)
+{
+	char r_buf[500];
+	char *f_path = "/proc/cpuinfo";
+	char *token = ":";
+	int i, ii, fd, len, v_len, r_len;
+	int val = 0;
+
+	fd = open(f_path, O_RDONLY);
+	if (fd < 0)
+		return val;
+
+	v_len = strlen(v_str);
+	r_len = sizeof(r_buf) - 1;
+	len = read(fd, r_buf, r_len);
+
+	if (len < 1)
+		return val;
+
+	/* get value pattern followed by : followed by value */
+	for (i=0; i < len; i++) {
+		for (ii=0; ii < v_len && i < len; ii++, i++) {
+			if ((v_str[ii] == r_buf[i]) && (ii == v_len-1))
+				for (; i < len; i++) {
+					if (!strncmp(&r_buf[i], token, 1)) {
+						val = atoi(&r_buf[i+1]);
+						i = len;
+					}
+				}
+			else if (v_str[ii] != r_buf[i])
+				break;
+		}
+	}
+
+	close(fd);
+	return val;
+}
+#endif
+
 enum ibv_mtu dapl_ib_mtu(int mtu)
 {
 	switch (mtu) {
@@ -287,6 +327,22 @@ const char *dapl_ib_port_str(enum ibv_port_state state)
 	}
 }
 
+const char *dapl_ib_port_num_str(unsigned long num)
+{
+	switch (num) {
+	case 1:
+		return "1";
+	case 2:
+		return "2";
+	case 3:
+		return "3";
+	case 4:
+		return "4";
+	default:
+		return "UNKNOWN";
+	}
+}
+
 /*
  * dapls_ib_query_hca
  *
@@ -311,6 +367,7 @@ DAT_RETURN dapls_ib_query_hca(IN DAPL_HCA * hca_ptr,
 			      OUT DAT_EP_ATTR * ep_attr,
 			      OUT DAT_SOCK_ADDR6 * ip_addr)
 {
+	ib_hca_transport_t *tp = &hca_ptr->ib_trans;
 	struct ibv_device_attr dev_attr;
 	struct ibv_port_attr port_attr;
 
@@ -327,6 +384,30 @@ DAT_RETURN dapls_ib_query_hca(IN DAPL_HCA * hca_ptr,
 	if (hca_ptr->ib_hca_handle == NULL) /* no open device, query mode */
 		return DAT_SUCCESS;
 
+#ifdef _OPENIB_MCM_
+		if (tp->self.node)
+			tp->na.mode = "PROXY";
+		else
+			tp->na.mode = "DIRECT";
+
+		tp->na.read = "FALSE";
+		sprintf(tp->ver_str, "%d", DAT_MIX_VER);
+
+		if (!tp->pr_attr.cpu_family) {
+			if (tp->self.node) {
+				dapli_mix_get_attr(tp, &tp->pr_attr);
+			} else {
+				tp->pr_attr.cpu_family = cpuinfo_atoi("cpu family");
+				tp->pr_attr.cpu_model = cpuinfo_atoi("model");
+			}
+			sprintf(tp->fam_str, "%d", tp->pr_attr.cpu_family);
+			sprintf(tp->mod_str, "%d", tp->pr_attr.cpu_model);
+		}
+#else
+		tp->na.mode = "DIRECT";
+		tp->na.read = "TRUE";
+#endif
+
 	/* query verbs for this device and port attributes */
 	if (ibv_query_device(hca_ptr->ib_hca_handle, &dev_attr) ||
 	    ibv_query_port(hca_ptr->ib_hca_handle,
@@ -348,7 +429,7 @@ DAT_RETURN dapls_ib_query_hca(IN DAPL_HCA * hca_ptr,
 	if (ia_attr != NULL) {
 		(void)dapl_os_memzero(ia_attr, sizeof(*ia_attr));
 		strncpy(ia_attr->adapter_name,
-		        ibv_get_device_name(hca_ptr->ib_trans.ib_dev),
+		        ibv_get_device_name(tp->ib_dev),
 		        DAT_NAME_MAX_LENGTH - 1);
 		ia_attr->adapter_name[DAT_NAME_MAX_LENGTH - 1] = '\0';
 		ia_attr->vendor_name[DAT_NAME_MAX_LENGTH - 1] = '\0';
@@ -398,52 +479,35 @@ DAT_RETURN dapls_ib_query_hca(IN DAPL_HCA * hca_ptr,
 		ia_attr->extension_version = DAT_IB_EXTENSION_VERSION;
 #endif
 		/* save key device attributes for CM exchange */
-		hca_ptr->ib_trans.rd_atom_in  = dev_attr.max_qp_rd_atom;
-		hca_ptr->ib_trans.rd_atom_out = dev_attr.max_qp_init_rd_atom;
-		
-		hca_ptr->ib_trans.mtu = DAPL_MIN(port_attr.active_mtu,
-						 hca_ptr->ib_trans.mtu);
-		hca_ptr->ib_trans.ack_timer =
-		    DAPL_MAX(dev_attr.local_ca_ack_delay,
-			     hca_ptr->ib_trans.ack_timer);
+		tp->ib_cm.rd_atom_in  = dev_attr.max_qp_rd_atom;
+		tp->ib_cm.rd_atom_out = dev_attr.max_qp_init_rd_atom;
+		tp->ib_cm.mtu = DAPL_MIN(port_attr.active_mtu, tp->ib_cm.mtu);
+		tp->ib_cm.ack_timer = DAPL_MAX(dev_attr.local_ca_ack_delay, tp->ib_cm.ack_timer);
 
 		/* set provider/transport specific named attributes */
-		hca_ptr->ib_trans.na.dev = ia_attr->adapter_name;
-		hca_ptr->ib_trans.na.mtu = dapl_ib_mtu_str(hca_ptr->ib_trans.mtu);
-		hca_ptr->ib_trans.na.port = dapl_ib_port_str(port_attr.state);
-		hca_ptr->ib_trans.guid = ntohll(ibv_get_device_guid(hca_ptr->ib_trans.ib_dev));
-		sprintf(hca_ptr->ib_trans.guid_str, "%04x:%04x:%04x:%04x",
-			(unsigned) (hca_ptr->ib_trans.guid >> 48) & 0xffff,
-			(unsigned) (hca_ptr->ib_trans.guid >> 32) & 0xffff,
-			(unsigned) (hca_ptr->ib_trans.guid >> 16) & 0xffff,
-			(unsigned) (hca_ptr->ib_trans.guid >>  0) & 0xffff);
-#ifdef _OPENIB_MCM_
-		hca_ptr->ib_trans.sys_guid = dev_attr.sys_image_guid; /* network order */
-		if (hca_ptr->ib_trans.self.node)
-			hca_ptr->ib_trans.na.mode = "PROXY";
-		else
-			hca_ptr->ib_trans.na.mode = "DIRECT";
-		hca_ptr->ib_trans.na.read = "FALSE";
+		tp->na.dev = ia_attr->adapter_name;
+		tp->na.mtu = dapl_ib_mtu_str(tp->ib_cm.mtu);
+		tp->na.port = dapl_ib_port_str(port_attr.state);
+		tp->na.port_num = dapl_ib_port_num_str(hca_ptr->port_num);
+		if (!tp->guid)
+			tp->guid = ntohll(ibv_get_device_guid(tp->ib_dev));
+
+		sprintf(tp->guid_str, "%04x:%04x:%04x:%04x",
+			(unsigned) (tp->guid >> 48) & 0xffff,
+			(unsigned) (tp->guid >> 32) & 0xffff,
+			(unsigned) (tp->guid >> 16) & 0xffff,
+			(unsigned) (tp->guid >>  0) & 0xffff);
 
-		if (!hca_ptr->ib_trans.pr_attr.cpu_family && hca_ptr->ib_trans.scif_ep) {
-			dapli_mix_get_attr(&hca_ptr->ib_trans, &hca_ptr->ib_trans.pr_attr);
-			sprintf(hca_ptr->ib_trans.fam_str, "%d", hca_ptr->ib_trans.pr_attr.cpu_family);
-			sprintf(hca_ptr->ib_trans.mod_str, "%d", hca_ptr->ib_trans.pr_attr.cpu_model);
-		}
-#else
-		hca_ptr->ib_trans.na.mode = "DIRECT";
-		hca_ptr->ib_trans.na.read = "TRUE";
-#endif
 		if (hca_ptr->ib_hca_handle->device->transport_type != IBV_TRANSPORT_IB)
 			goto skip_ib;
 
-               /* set SL, PKEY values, defaults = 0 */
-               hca_ptr->ib_trans.pkey_idx = 0;
-               hca_ptr->ib_trans.pkey = htons(dapl_os_get_env_val("DAPL_IB_PKEY", 0));
-               hca_ptr->ib_trans.sl = dapl_os_get_env_val("DAPL_IB_SL", 0);
+		/* set SL, PKEY values, defaults = 0 */
+		tp->ib_cm.pkey_idx = 0;
+		tp->ib_cm.pkey = htons(dapl_os_get_env_val("DAPL_IB_PKEY", 0));
+		tp->ib_cm.sl = dapl_os_get_env_val("DAPL_IB_SL", 0);
 
 		/* index provided, get pkey; pkey provided, get index */
-		if (hca_ptr->ib_trans.pkey) {
+		if (tp->ib_cm.pkey) {
 			int i; uint16_t pkey = 0;
 			for (i=0; i < dev_attr.max_pkeys; i++) {
 				if (ibv_query_pkey(hca_ptr->ib_hca_handle,
@@ -452,8 +516,8 @@ DAT_RETURN dapls_ib_query_hca(IN DAPL_HCA * hca_ptr,
 					i = dev_attr.max_pkeys;
 					break;
 				}
-				if (pkey == hca_ptr->ib_trans.pkey) {
-					hca_ptr->ib_trans.pkey_idx = i;
+				if (pkey == tp->ib_cm.pkey) {
+					tp->ib_cm.pkey_idx = i;
 					break;
 				}
 			}
@@ -461,7 +525,7 @@ DAT_RETURN dapls_ib_query_hca(IN DAPL_HCA * hca_ptr,
 				dapl_log(DAPL_DBG_TYPE_ERR,
 					 " ERR: new pkey(0x%x), query (%s)"
 					 " err or key !found, using default pkey_idx=0\n",
-					 ntohs(hca_ptr->ib_trans.pkey), strerror(errno));
+					 ntohs(tp->ib_cm.pkey), strerror(errno));
 			}
 		}
 skip_ib:
@@ -470,36 +534,35 @@ skip_ib:
 #ifndef _OPENIB_CMA_
 		if (port_attr.link_layer != IBV_LINK_LAYER_INFINIBAND &&
 		    port_attr.link_layer != IBV_LINK_LAYER_UNSPECIFIED)
-			hca_ptr->ib_trans.global = 1;
+			tp->ib_cm.global = 1;
 
 		dapl_log(DAPL_DBG_TYPE_UTIL,
 			 " query_hca: port.link_layer = 0x%x, global = %d\n",
-			 port_attr.link_layer, hca_ptr->ib_trans.global);
+			 port_attr.link_layer, tp->ib_cm.global);
 #endif
 #endif
 
 #ifdef _WIN32
 #ifndef _OPENIB_CMA_
 		if (port_attr.transport != IBV_TRANSPORT_IB)
-			hca_ptr->ib_trans.global = 1;
+			tp->ib_cm.global = 1;
 
 		dapl_log(DAPL_DBG_TYPE_UTIL,
 			 " query_hca: port.transport %d ib_trans.global %d\n",
-			 port_attr.transport, hca_ptr->ib_trans.global);
+			 port_attr.transport, tp->ib_cm.global);
 #endif
 #endif
 
 		dapl_log(DAPL_DBG_TYPE_UTIL,
 			     " query_hca: (%x.%x) eps %d, sz %d evds %d,"
-			     " sz %d mtu %d - pkey %x p_idx %d sl %d gl %d\n",
+			     " sz %d mtu %d pk %x pi %d sl %d gl %d inl %d\n",
 			     ia_attr->hardware_version_major,
 			     ia_attr->hardware_version_minor,
 			     ia_attr->max_eps, ia_attr->max_dto_per_ep,
 			     ia_attr->max_evds, ia_attr->max_evd_qlen,
-			     128 << hca_ptr->ib_trans.mtu,
-			     ntohs(hca_ptr->ib_trans.pkey),
-			     hca_ptr->ib_trans.pkey_idx,
-			     hca_ptr->ib_trans.sl, hca_ptr->ib_trans.global);
+			     128 << tp->ib_cm.mtu, ntohs(tp->ib_cm.pkey),
+			     tp->ib_cm.pkey_idx, tp->ib_cm.sl,
+			     tp->ib_cm.global, tp->ib_cm.max_inline);
 
 		dapl_log(DAPL_DBG_TYPE_UTIL,
 			     " query_hca: msg %llu rdma %llu iov %d lmr %d rmr %d"
@@ -507,7 +570,7 @@ skip_ib:
 			     ia_attr->max_message_size, ia_attr->max_rdma_size,
 			     ia_attr->max_iov_segments_per_dto,
 			     ia_attr->max_lmrs, ia_attr->max_rmrs,
-			     hca_ptr->ib_trans.ack_timer,
+			     tp->ib_cm.ack_timer,
 			     ia_attr->max_lmr_block_size,
 			     ia_attr->ia_address_ptr);
 	}
@@ -528,7 +591,7 @@ skip_ib:
 			     " query_hca: MAX msg %llu mtu %d qsz %d iov %d"
 			     " rdma i%d,o%d\n",
 			     ep_attr->max_message_size,
-			     128 << hca_ptr->ib_trans.mtu,
+			     128 << tp->ib_cm.mtu,
 			     ep_attr->max_recv_dtos, 
 			     ep_attr->max_recv_iov,
 			     ep_attr->max_rdma_read_in,
@@ -637,11 +700,16 @@ void dapli_async_event_cb(struct _ib_hca_transport *hca)
 		case IBV_EVENT_SRQ_LIMIT_REACHED:
 		case IBV_EVENT_SQ_DRAINED:
 		{
+			DAPL_DBG_TYPE dbg_type = DAPL_DBG_TYPE_ERR;
 			struct dapl_ep *ep_ptr =
 				event.element.qp->qp_context;
 
-			dapl_log(DAPL_DBG_TYPE_ERR,
-				 "dapl async_event QP (%p) ERR %d\n",
+			if (event.event_type == IBV_EVENT_QP_LAST_WQE_REACHED &&
+			    ep_ptr->param.srq_handle) {
+				dbg_type = DAPL_DBG_TYPE_EVD;
+			}
+
+			dapl_log(dbg_type, "dapl async_event QP (%p) Event %d\n",
 				 ep_ptr, event.event_type);
 
 			/* report up if async callback still setup */
@@ -725,14 +793,9 @@ DAT_NAMED_ATTR ib_attrs[] = {
 	{
 	 "DAT_IB_PORT_STATUS", "UNKNOWN"}
 	,
-#ifdef _OPENIB_MCM_
 	{
-	 "DAT_IB_PROXY_CPU_FAMILY", "UNKNOWN"}
+	 "DAT_IB_PORT_NUMBER", "UNKNOWN"}
 	,
-	{
-	 "DAT_IB_PROXY_CPU_MODEL", "UNKNOWN"}
-	,
-#endif
 #ifdef DAT_EXTENSIONS
 	{
 	 "DAT_EXTENSION_INTERFACE", "TRUE"}
@@ -778,6 +841,17 @@ DAT_NAMED_ATTR ib_attrs[] = {
 	 DAT_ATTR_COUNTERS, "TRUE"}
 	,
 #endif				/* DAPL_COUNTERS */
+#ifdef _OPENIB_MCM_
+	{
+	 "DAT_IB_PROXY_CPU_FAMILY", "UNKNOWN"}
+	,
+	{
+	 "DAT_IB_PROXY_CPU_MODEL", "UNKNOWN"}
+	,
+	{
+	 "DAT_IB_PROXY_VERSION", "UNKNOWN"}
+	,
+#endif  /* _OPENIB_MCM_, end of list */
 #endif
 };
 
@@ -786,7 +860,6 @@ DAT_NAMED_ATTR ib_attrs[] = {
 void dapls_query_provider_specific_attr(IN DAPL_IA * ia_ptr,
 					IN DAT_PROVIDER_ATTR * attr_ptr)
 {
-
 	attr_ptr->num_provider_specific_attr = SPEC_ATTR_SIZE(ib_attrs);
 	attr_ptr->provider_specific_attr = ib_attrs;
 
@@ -800,9 +873,14 @@ void dapls_query_provider_specific_attr(IN DAPL_IA * ia_ptr,
 	ib_attrs[4].value = ia_ptr->hca_ptr->ib_trans.guid_str;
 	ib_attrs[5].value = ia_ptr->hca_ptr->ib_trans.na.mtu;
 	ib_attrs[6].value = ia_ptr->hca_ptr->ib_trans.na.port;
+	ib_attrs[7].value = ia_ptr->hca_ptr->ib_trans.na.port_num;
 #ifdef _OPENIB_MCM_
-	ib_attrs[7].value = ia_ptr->hca_ptr->ib_trans.fam_str;
-	ib_attrs[8].value = ia_ptr->hca_ptr->ib_trans.mod_str;
+{
+	int i = attr_ptr->num_provider_specific_attr;
+	ib_attrs[i-3].value = ia_ptr->hca_ptr->ib_trans.fam_str;
+	ib_attrs[i-2].value = ia_ptr->hca_ptr->ib_trans.mod_str;
+	ib_attrs[i-1].value = ia_ptr->hca_ptr->ib_trans.ver_str;
+}
 #endif
 }
 
