@@ -642,7 +642,7 @@ static int mcm_send(ib_hca_transport_t *tp, dat_mcm_msg_t *msg, DAT_PVOID p_data
         wr.opcode = IBV_WR_SEND;
         wr.wr_id = (unsigned long)tp->s_hd;
 	wr.send_flags = (wr.wr_id % tp->burst) ? 0 : IBV_SEND_SIGNALED;
-	if (len <= tp->max_inline_send)
+	if (len <= tp->ib_cm.max_inline)
 		wr.send_flags |= IBV_SEND_INLINE; 
 
         sge.length = len;
@@ -749,6 +749,7 @@ dp_ib_cm_handle_t dapls_cm_create(DAPL_HCA *hca, DAPL_EP *ep)
 	cm->msg.ver = htons(DAT_MCM_VER);
 	cm->msg.s_id = htonl(dapl_os_getpid()); /* process id for src id */
 	cm->msg.sys_guid = hca->ib_trans.sys_guid;
+	cm->msg.seg_sz = DAT_MCM_SEG_PO2;
 	
 	/* ACTIVE: init source address QP info from local EP */
 	if (ep) {
@@ -817,6 +818,11 @@ void dapls_cm_free(dp_ib_cm_handle_t cm)
 	dapl_ep_unlink_cm(cm->ep, cm);
 }
 
+DAT_RETURN dapls_ud_cm_free(DAPL_EP *ep_ptr, dp_ib_cm_handle_t cm_ptr)
+{
+	return DAT_NOT_IMPLEMENTED;
+}
+
 /* ACTIVE/PASSIVE: queue up connection object on CM list */
 void dapli_queue_conn(dp_ib_cm_handle_t cm)
 {
@@ -1083,6 +1089,9 @@ void mcm_connect_rtu(dp_ib_cm_handle_t cm, dat_mcm_msg_t *msg)
 	}
 
 	/* CM_REP: save remote address information to EP and CM */
+	if (msg->seg_sz) /* set po2 seg_sz, if provided */
+		cm->msg.seg_sz = msg->seg_sz;
+
 	cm->msg.d_id = msg->s_id;
 	dapl_os_memcpy(&ep->remote_ia_address, &msg->saddr2, sizeof(dat_mcm_addr_t));
 	dapl_os_memcpy(&cm->msg.daddr2, &msg->saddr2, sizeof(dat_mcm_addr_t));
@@ -1180,16 +1189,18 @@ void mcm_connect_rtu(dp_ib_cm_handle_t cm, dat_mcm_msg_t *msg)
 			mcm_create_wc_q(ep->qp_handle, ep->qp_handle->wrc_rem.wr_end + 1);
 			mcm_hton_wrc((mcm_wrc_info_t*)cm->msg.p_proxy, &ep->qp_handle->wrc);
 			ep->qp_handle->ep_map = cm->msg.daddr1.ep_map;
+			ep->qp_handle->seg_sz = (1 << cm->msg.seg_sz);
 
 			/* post 0-byte rcv for inbound WC's via RW_imm */
 			if (mcm_post_rcv_wc(ep->qp_handle, MCM_WRC_QLEN))
 				goto bail;
 
-			dapl_log(DAPL_DBG_TYPE_CM, "CONN_RTU: WR_rem %p sz %d, WC %p sz %d\n",
+			dapl_log(DAPL_DBG_TYPE_CM, "CONN_RTU: WR_rem %p sz %d, WC %p sz %d, sg %d\n",
 				 ep->qp_handle->wrc_rem.wr_addr,
 				 ep->qp_handle->wrc_rem.wr_end+1,
 				 ep->qp_handle->wrc.wc_addr,
-				 ep->qp_handle->wrc.wc_end+1);
+				 ep->qp_handle->wrc.wc_end+1,
+				 ep->qp_handle->seg_sz);
 		}
 	}
 	dapl_os_unlock(&cm->ep->header.lock);
@@ -1260,6 +1271,8 @@ static void mcm_accept(ib_cm_srvc_handle_t cm, dat_mcm_msg_t *msg)
 	acm->msg.p_size = msg->p_size;
 	acm->msg.d_id = msg->s_id;
 	acm->msg.rd_in = msg->rd_in;
+	if (msg->seg_sz) /* set po2 seg_sz, if provided */
+		acm->msg.seg_sz = msg->seg_sz;
 
 	/* CR saddr1 is CM daddr1 info, need EP for local saddr1 */
 	dapl_os_memcpy(&acm->msg.daddr1, &msg->saddr1, sizeof(dat_mcm_addr_t));
@@ -1500,19 +1513,21 @@ dapli_accept_usr(DAPL_EP *ep, DAPL_CR *cr, DAT_COUNT p_size, DAT_PVOID p_data)
 			mcm_create_wc_q(ep->qp_handle, ep->qp_handle->wrc_rem.wr_end + 1);
 			mcm_hton_wrc((mcm_wrc_info_t*)cm->msg.p_proxy, &ep->qp_handle->wrc);
 			ep->qp_handle->ep_map = cm->msg.daddr1.ep_map;
+			ep->qp_handle->seg_sz = (1 << cm->msg.seg_sz);
 
 			/* post 0-byte rcv for inbound WC's via RW_imm */
 			if (mcm_post_rcv_wc(ep->qp_handle, MCM_WRC_QLEN))
 					goto bail;
 
 			dapl_log(DAPL_DBG_TYPE_CM,
-				 "ACCEPT_USR: WR_rem %p rkey %x sz %d, WC %p rkey %x sz %d\n",
+				 "ACCEPT_USR: WR_rem %p rkey %x sz %d, WC %p rkey %x sz %d sg %d\n",
 				 ep->qp_handle->wrc_rem.wr_addr,
 				 ep->qp_handle->wrc_rem.wr_rkey,
 				 ep->qp_handle->wrc_rem.wr_end+1,
 				 ep->qp_handle->wrc.wc_addr,
 				 ep->qp_handle->wrc.wc_rkey,
-				 ep->qp_handle->wrc.wc_end+1);
+				 ep->qp_handle->wrc.wc_end+1,
+				 ep->qp_handle->seg_sz);
 		}
 	}
 	dapl_os_unlock(&ep->header.lock);
