@@ -143,7 +143,22 @@ int dapli_mix_open(ib_hca_transport_t *tp, char *name, int port, int query_only)
 	msg.hdr.req_id = dapl_os_getpid();
 	msg.port = port;
 	strcpy((char*)&msg.name, name);
-	memcpy(&msg.dev_attr, (void*)&tp->ack_timer, sizeof(dat_mix_dev_attr_t));
+
+	/* send any overridden attributes to proxy */
+	msg.dev_attr.ack_timer = tp->ib_cm.ack_timer;
+	msg.dev_attr.ack_retry = tp->ib_cm.ack_retry;
+	msg.dev_attr.rnr_timer = tp->ib_cm.rnr_timer;
+	msg.dev_attr.rnr_retry = tp->ib_cm.rnr_retry ;
+	msg.dev_attr.global = tp->ib_cm.global;
+	msg.dev_attr.hop_limit = tp->ib_cm.hop_limit;
+	msg.dev_attr.tclass = tp->ib_cm.tclass;
+	msg.dev_attr.sl = tp->ib_cm.sl;
+	msg.dev_attr.mtu = tp->ib_cm.mtu;
+	msg.dev_attr.rd_atom_in = tp->ib_cm.rd_atom_in;
+	msg.dev_attr.rd_atom_out = tp->ib_cm.rd_atom_out;
+	msg.dev_attr.pkey_idx = tp->ib_cm.pkey_idx;
+	msg.dev_attr.pkey = tp->ib_cm.pkey;
+	msg.dev_attr.max_inline = tp->ib_cm.max_inline;
 
 	len = sizeof(dat_mix_open_t);
 	ret = scif_send(tp->scif_ep, &msg, len, SCIF_SEND_BLOCK);
@@ -174,7 +189,25 @@ int dapli_mix_open(ib_hca_transport_t *tp, char *name, int port, int query_only)
 	}
 	/* save address to transport object, keeps IA queries local */
 	memcpy((void*)&tp->addr, (void*)&msg.dev_addr, sizeof(dat_mcm_addr_t));
+
+	/* save actual attributes and device ID */
+	tp->ib_cm.ack_timer = msg.dev_attr.ack_timer;
+	tp->ib_cm.ack_retry = msg.dev_attr.ack_retry;
+	tp->ib_cm.rnr_timer = msg.dev_attr.rnr_timer;
+	tp->ib_cm.rnr_retry = msg.dev_attr.rnr_retry;
+	tp->ib_cm.global = msg.dev_attr.global;
+	tp->ib_cm.hop_limit = msg.dev_attr.hop_limit;
+	tp->ib_cm.tclass = msg.dev_attr.tclass;
+	tp->ib_cm.sl = msg.dev_attr.sl;
+	tp->ib_cm.mtu = msg.dev_attr.mtu;
+	tp->ib_cm.rd_atom_in = msg.dev_attr.rd_atom_in;
+	tp->ib_cm.rd_atom_out = msg.dev_attr.rd_atom_out;
+	tp->ib_cm.pkey_idx = msg.dev_attr.pkey_idx;
+	tp->ib_cm.pkey = msg.dev_attr.pkey;
+	tp->ib_cm.max_inline = msg.dev_attr.max_inline;
+
 	tp->dev_id = msg.hdr.req_id;
+
 	dapl_log(DAPL_DBG_TYPE_EXTENSION,
 		 " mix_open reply (msg %p, ln %d) EPs %d %d %d - dev_id %d\n",
 		 &msg, len, tp->scif_ep, tp->scif_ev_ep,
