@@ -42,7 +42,7 @@
  *
  *    $Id: $
  *
- *	Copyright (c) 2005 Intel Corporation.  All rights reserved.
+ *	Copyright (c) 2005-2015 Intel Corporation.  All rights reserved.
  *
  **************************************************************************/
 
@@ -470,6 +470,11 @@ void dapls_cm_free(dp_ib_cm_handle_t cm_ptr)
 	dapl_ep_unlink_cm(cm_ptr->ep, cm_ptr);
 }
 
+DAT_RETURN dapls_ud_cm_free(DAPL_EP *ep_ptr, dp_ib_cm_handle_t cm_ptr)
+{
+	return DAT_NOT_IMPLEMENTED;
+}
+
 /*
  * ACTIVE/PASSIVE: called from CR thread or consumer via ep_disconnect
  *                 or from ep_free. 
@@ -732,13 +737,11 @@ static void dapli_socket_connect_rtu(dp_ib_cm_handle_t cm_ptr)
 		int err = dapl_socket_errno();
 		dapl_log(DAPL_DBG_TYPE_CM_WARN,
 			 " CONN_REP_PENDING: sk %d ERR 0x%x, rcnt=%d, v=%d ->"
-			 " %s PORT L-%x R-%x PID L-%x%x R-%x%x %d\n",
+			 " %s PORT L-%x R-%x %d\n",
 			 cm_ptr->socket, err, len, ntohs(cm_ptr->msg.ver),
 			 inet_ntoa(((struct sockaddr_in *)&cm_ptr->addr)->sin_addr),
 			 ntohs(((struct sockaddr_in *)&cm_ptr->msg.daddr.so)->sin_port),
 			 ntohs(((struct sockaddr_in *)&cm_ptr->addr)->sin_port),
-			 cm_ptr->msg.resv[1],cm_ptr->msg.resv[0],
-			 cm_ptr->msg.resv[3],cm_ptr->msg.resv[2],
 			 cm_ptr->retry);
 
 		/* Retry; corner case where server tcp stack resets under load */
@@ -1297,11 +1300,6 @@ dapli_socket_accept_usr(DAPL_EP * ep_ptr,
 	sl = sizeof(local.daddr.so);
 	getsockname(cm_ptr->socket, (struct sockaddr *)&local.daddr.so, &sl);
 
-#ifdef DAPL_DBG
-	/* DBG: Active PID [0], PASSIVE PID [2] */
-	*(uint16_t*)&cm_ptr->msg.resv[2] = htons((uint16_t)dapl_os_getpid()); 
-	dapl_os_memcpy(local.resv, cm_ptr->msg.resv, 4); 
-#endif
 	cm_ptr->hca = ia_ptr->hca_ptr;
 	dapl_os_lock(&cm_ptr->lock);
 	cm_ptr->state = DCM_ACCEPTED;
@@ -1968,7 +1966,7 @@ void dapls_print_cm_list(IN DAPL_IA *ia_ptr)
 				(DAPL_LLIST_ENTRY*)&cr->local_entry);
 
 		printf( "  CONN[%d]: sp %p ep %p sock %d %s %s %s %s %s %s "
-			" PORT L-%x R-%x PID L-%x%x R-%x%x\n",
+			" PORT L-%x R-%x \n",
 			i, cr->sp, cr->ep, cr->socket,
 			cr->msg.saddr.ib.qp_type == IBV_QPT_RC ? "RC" : "UD",
 			dapl_cm_state_str(cr->state), dapl_cm_op_str(ntohs(cr->msg.op)),
@@ -1986,10 +1984,7 @@ void dapls_print_cm_list(IN DAPL_IA *ia_ptr)
 
 			ntohs(cr->msg.op) == DCM_REQ ? /* remote port */
 				ntohs(((struct sockaddr_in *)&cr->addr)->sin_port) :
-				ntohs(((struct sockaddr_in *)&cr->msg.daddr.so)->sin_port),
-
-			cr->sp ? cr->msg.resv[3] : cr->msg.resv[1], cr->sp ? cr->msg.resv[2] : cr->msg.resv[0],
-			cr->sp ? cr->msg.resv[1] : cr->msg.resv[3], cr->sp ? cr->msg.resv[0] : cr->msg.resv[2]);
+				ntohs(((struct sockaddr_in *)&cr->msg.daddr.so)->sin_port) );
 		i++;
 	}
 	printf("\n");
