@@ -274,7 +274,7 @@ DAT_RETURN dapls_ib_open_hca(IN IB_HCA_NAME hca_name,
 {
 	struct ibv_device **dev_list;
 	struct ibv_port_attr port_attr;
-	int i;
+	int i, nd = 0;
 	DAT_RETURN dat_status = DAT_SUCCESS;
 
 	dapl_log(DAPL_DBG_TYPE_UTIL, " open_hca: %s %s - %p in %s\n",
@@ -289,25 +289,28 @@ DAT_RETURN dapls_ib_open_hca(IN IB_HCA_NAME hca_name,
 	}
 
         /* Get list of all IB devices, find match, open */
-	dev_list = ibv_get_device_list(NULL);
+	dev_list = ibv_get_device_list(&nd);
 	if (!dev_list) {
 		dapl_log(DAPL_DBG_TYPE_ERR,
 			 " open_hca: ibv_get_device_list() failed\n",
 			 hca_name);
 		return DAT_INTERNAL_ERROR;
 	}
+	hca_ptr->ib_trans.ib_dev = NULL;
+	for (i = 0; i < nd; ++i) {
+		if (!hca_ptr->ib_trans.guid &&
+		    dev_list[i]->transport_type == IBV_TRANSPORT_IB)
+			hca_ptr->ib_trans.guid = ibv_get_device_guid(dev_list[i]);
 
-	for (i = 0; dev_list[i]; ++i) {
-		hca_ptr->ib_trans.ib_dev = dev_list[i];
-		if (!strcmp(ibv_get_device_name(hca_ptr->ib_trans.ib_dev),
-			    hca_name))
-			goto found;
+		if (!strcmp(dev_list[i]->name, hca_name))
+			hca_ptr->ib_trans.ib_dev = dev_list[i];
+	}
+	if (hca_ptr->ib_trans.ib_dev == NULL) {
+		dapl_log(DAPL_DBG_TYPE_ERR, " open_hca: device %s not found\n", hca_name);
+		dat_status = DAT_PROVIDER_NOT_FOUND;
+		goto err;
 	}
-	dapl_log(DAPL_DBG_TYPE_ERR,
-		 " open_hca: device %s not found\n", hca_name);
-	goto err;
 
-found:
 	dapl_dbg_log(DAPL_DBG_TYPE_UTIL, " open_hca: Found dev %s %016llx\n",
 		     ibv_get_device_name(hca_ptr->ib_trans.ib_dev),
 		     (unsigned long long)
@@ -349,23 +352,26 @@ found:
 	}
 
 	/* set RC tunables via enviroment or default */
-	hca_ptr->ib_trans.max_inline_send =
-	    dapl_os_get_env_val("DAPL_MAX_INLINE", INLINE_SEND_DEFAULT);
-	hca_ptr->ib_trans.ack_retry =
+	if (dapl_ib_inline_data(hca_ptr->ib_hca_handle)) {
+		hca_ptr->ib_trans.ib_cm.max_inline =
+			dapl_os_get_env_val("DAPL_MAX_INLINE",
+					    INLINE_SEND_DEFAULT);
+	}
+	hca_ptr->ib_trans.ib_cm.ack_retry =
 	    dapl_os_get_env_val("DAPL_ACK_RETRY", SCM_ACK_RETRY);
-	hca_ptr->ib_trans.ack_timer =
+	hca_ptr->ib_trans.ib_cm.ack_timer =
 	    dapl_os_get_env_val("DAPL_ACK_TIMER", SCM_ACK_TIMER);
-	hca_ptr->ib_trans.rnr_retry =
+	hca_ptr->ib_trans.ib_cm.rnr_retry =
 	    dapl_os_get_env_val("DAPL_RNR_RETRY", SCM_RNR_RETRY);
-	hca_ptr->ib_trans.rnr_timer =
+	hca_ptr->ib_trans.ib_cm.rnr_timer =
 	    dapl_os_get_env_val("DAPL_RNR_TIMER", SCM_RNR_TIMER);
-	hca_ptr->ib_trans.global =
+	hca_ptr->ib_trans.ib_cm.global =
 	    dapl_os_get_env_val("DAPL_GLOBAL_ROUTING", SCM_GLOBAL);
-	hca_ptr->ib_trans.hop_limit =
+	hca_ptr->ib_trans.ib_cm.hop_limit =
 	    dapl_os_get_env_val("DAPL_HOP_LIMIT", SCM_HOP_LIMIT);
-	hca_ptr->ib_trans.tclass =
+	hca_ptr->ib_trans.ib_cm.tclass =
 	    dapl_os_get_env_val("DAPL_TCLASS", SCM_TCLASS);
-	hca_ptr->ib_trans.mtu =
+	hca_ptr->ib_trans.ib_cm.mtu =
 	    dapl_ib_mtu(dapl_os_get_env_val("DAPL_IB_MTU", SCM_IB_MTU));
 
 	if (flags & DAPL_OPEN_QUERY)
