@@ -216,17 +216,21 @@ DAT_RETURN dapls_ib_open_hca(IN IB_HCA_NAME hca_name,
 			 hca_name);
 		return DAT_INTERNAL_ERROR;
 	}
-
+	hca_ptr->ib_trans.ib_dev = NULL;
 	for (i = 0; i < nd; ++i) {
-		if (!strcmp(dev_list[i]->name, hca_name)) {
+		if (!hca_ptr->ib_trans.guid &&
+		    dev_list[i]->transport_type == IBV_TRANSPORT_IB)
+			hca_ptr->ib_trans.guid = ibv_get_device_guid(dev_list[i]);
+
+		if (!strcmp(dev_list[i]->name, hca_name))
 			hca_ptr->ib_trans.ib_dev = dev_list[i];
-			goto found;
-		}
 	}
-	dapl_log(DAPL_DBG_TYPE_ERR, " open_hca: %s not found\n", hca_name);
-	dat_status = DAT_PROVIDER_NOT_FOUND;
-	goto err;
-found:
+	if (hca_ptr->ib_trans.ib_dev == NULL) {
+		dapl_log(DAPL_DBG_TYPE_ERR, " open_hca: device %s not found\n", hca_name);
+		dat_status = DAT_PROVIDER_NOT_FOUND;
+		goto err;
+	}
+
 	hca_ptr->ib_hca_handle = ibv_open_device(hca_ptr->ib_trans.ib_dev);
 	if (!hca_ptr->ib_hca_handle) {
 		dapl_log(DAPL_DBG_TYPE_ERR,
@@ -269,23 +273,26 @@ found:
 	}
 
 	/* set RC tunables via enviroment or default */
-	hca_ptr->ib_trans.max_inline_send =
-	    dapl_os_get_env_val("DAPL_MAX_INLINE", INLINE_SEND_IB_DEFAULT);
-	hca_ptr->ib_trans.ack_retry =
+	if (dapl_ib_inline_data(hca_ptr->ib_hca_handle)) {
+		hca_ptr->ib_trans.ib_cm.max_inline =
+			dapl_os_get_env_val("DAPL_MAX_INLINE",
+					     INLINE_SEND_IB_DEFAULT);
+	}
+	hca_ptr->ib_trans.ib_cm.ack_retry =
 	    dapl_os_get_env_val("DAPL_ACK_RETRY", DCM_ACK_RETRY);
-	hca_ptr->ib_trans.ack_timer =
+	hca_ptr->ib_trans.ib_cm.ack_timer =
 	    dapl_os_get_env_val("DAPL_ACK_TIMER", DCM_ACK_TIMER);
-	hca_ptr->ib_trans.rnr_retry =
+	hca_ptr->ib_trans.ib_cm.rnr_retry =
 	    dapl_os_get_env_val("DAPL_RNR_RETRY", DCM_RNR_RETRY);
-	hca_ptr->ib_trans.rnr_timer =
+	hca_ptr->ib_trans.ib_cm.rnr_timer =
 	    dapl_os_get_env_val("DAPL_RNR_TIMER", DCM_RNR_TIMER);
-	hca_ptr->ib_trans.global =
+	hca_ptr->ib_trans.ib_cm.global =
 	    dapl_os_get_env_val("DAPL_GLOBAL_ROUTING", DCM_GLOBAL);
-	hca_ptr->ib_trans.hop_limit =
+	hca_ptr->ib_trans.ib_cm.hop_limit =
 	    dapl_os_get_env_val("DAPL_HOP_LIMIT", DCM_HOP_LIMIT);
-	hca_ptr->ib_trans.tclass =
+	hca_ptr->ib_trans.ib_cm.tclass =
 	    dapl_os_get_env_val("DAPL_TCLASS", DCM_TCLASS);
-	hca_ptr->ib_trans.mtu =
+	hca_ptr->ib_trans.ib_cm.mtu =
 	    dapl_ib_mtu(dapl_os_get_env_val("DAPL_IB_MTU", DCM_IB_MTU));
 
 	if (flags & DAPL_OPEN_QUERY)
@@ -304,6 +311,9 @@ found:
 	if ((dapl_os_lock_init(&hca_ptr->ib_trans.plock)) != DAT_SUCCESS)
 		goto bail;
 
+	if ((dapl_os_lock_init(&hca_ptr->ib_trans.ilock)) != DAT_SUCCESS)
+		goto bail;
+
 	/* EVD events without direct CQ channels, CNO support */
 	hca_ptr->ib_trans.ib_cq =
 	    ibv_create_comp_channel(hca_ptr->ib_hca_handle);
@@ -360,11 +370,11 @@ done:
 	       &hca_ptr->ib_trans.addr,
 	       sizeof(union dcm_addr));
 
-	dapl_dbg_log(DAPL_DBG_TYPE_UTIL,
+	dapl_log(DAPL_DBG_TYPE_UTIL,
 		 "%s open: dev %s port %d, GID %s, LID %x qpn %x sl %d\n",
 		 PROVIDER_NAME, hca_name, hca_ptr->port_num,
 		 inet_ntop(AF_INET6, &hca_ptr->ib_trans.addr.ib.gid,
-			   gid_str, sizeof(gid_str)),
+			  gid_str, sizeof(gid_str)),
 		 ntohs(ucm_ia->ib.lid), ntohl(ucm_ia->ib.qpn),
 		 ucm_ia->ib.sl, ucm_ia->ib.qp_type);
 
@@ -421,6 +431,7 @@ DAT_RETURN dapls_ib_close_hca(IN DAPL_HCA * hca_ptr)
 
 	dapl_os_lock_destroy(&hca_ptr->ib_trans.lock);
 	dapl_os_lock_destroy(&hca_ptr->ib_trans.llock);
+	dapl_os_lock_destroy(&hca_ptr->ib_trans.ilock);
 	destroy_os_signal(hca_ptr);
 	ucm_service_destroy(hca_ptr);
 done:
@@ -447,7 +458,7 @@ done:
 static void ucm_service_destroy(IN DAPL_HCA *hca)
 {
 	ib_hca_transport_t *tp = &hca->ib_trans;
-	int msg_size = sizeof(ib_cm_msg_t);
+	int i, msg_size = sizeof(ib_cm_msg_t);
 
 	if (tp->mr_sbuf)
 		ibv_dereg_mr(tp->mr_sbuf);
@@ -468,26 +479,32 @@ static void ucm_service_destroy(IN DAPL_HCA *hca)
 		ibv_destroy_comp_channel(tp->rch);
 
  	if (tp->ah) {
-		int i;
-
-		for (i = 0;i < 0xffff; i++) {
+		for (i=0; i<DCM_AH_SPACE; i++) {
 			if (tp->ah[i])
 				ibv_destroy_ah(tp->ah[i]);
 		}
-		dapl_os_free(tp->ah, (sizeof(*tp->ah) * 0xffff));
+		dapl_os_free(tp->ah, (sizeof(*tp->ah) * DCM_AH_SPACE));
 	}
 
 	if (tp->pd)
 		ibv_dealloc_pd(tp->pd);
 
 	if (tp->sid)
-		dapl_os_free(tp->sid, (sizeof(*tp->sid) * 0xffff));
+		dapl_os_free(tp->sid, UCM_SID_SPACE/UCM_SID_ENTRY);
 
 	if (tp->rbuf)
 		dapl_os_free(tp->rbuf, (msg_size * tp->qpe));
 
 	if (tp->sbuf)
 		dapl_os_free(tp->sbuf, (msg_size * tp->qpe));
+
+	if (tp->cm_idxr) {
+		for (i=0; i<=tp->cm_idxr_cur; i++) {
+			dapl_os_free(tp->cm_idxr[i],
+				     UCM_ENTRY_SIZE(tp->cm_entry_bits));
+			tp->cm_idxr[i] = 0;
+		}
+	}
 }
 
 static int ucm_service_create(IN DAPL_HCA *hca)
@@ -496,43 +513,93 @@ static int ucm_service_create(IN DAPL_HCA *hca)
 	ib_hca_transport_t *tp = &hca->ib_trans;
 	struct ibv_recv_wr recv_wr, *recv_err;
         struct ibv_sge sge;
-	int i, mlen = sizeof(ib_cm_msg_t);
+	int i, array_sz, entry_sz, mlen = sizeof(ib_cm_msg_t);
 	int hlen = sizeof(struct ibv_grh); /* hdr included with UD recv */
 	char *rbuf;
 
-	dapl_dbg_log(DAPL_DBG_TYPE_UTIL, " ucm_create: \n");
-
 	/* setup CM timers and queue sizes */
 	tp->retries = dapl_os_get_env_val("DAPL_UCM_RETRY", DCM_RETRY_CNT);
 	tp->rep_time = dapl_os_get_env_val("DAPL_UCM_REP_TIME", DCM_REP_TIME);
 	tp->rtu_time = dapl_os_get_env_val("DAPL_UCM_RTU_TIME", DCM_RTU_TIME);
-	tp->cm_timer = DAPL_MIN(tp->rep_time,tp->rtu_time);
+	tp->wait_time = dapl_os_get_env_val("DAPL_UCM_WAIT_TIME", DCM_WAIT_TIME);
 	tp->qpe = dapl_os_get_env_val("DAPL_UCM_QP_SIZE", DCM_QP_SIZE);
 	tp->cqe = dapl_os_get_env_val("DAPL_UCM_CQ_SIZE", DCM_CQ_SIZE);
 	tp->burst = dapl_os_get_env_val("DAPL_UCM_TX_BURST", DCM_TX_BURST);
-	tp->pd = ibv_alloc_pd(hca->ib_hca_handle);
-        if (!tp->pd) 
-                goto bail;
-        
-        dapl_log(DAPL_DBG_TYPE_UTIL,
-                        " create_service: pd %p ctx %p handle 0x%x\n",
-                         tp->pd, tp->pd->context, tp->pd->handle);
+	tp->dreq_cnt = dapl_os_get_env_val("DAPL_UCM_DREQ_RETRY", DCM_DREQ_CNT);
+	tp->drep_time = dapl_os_get_env_val("DAPL_UCM_DREP_TIME", DCM_DREP_TIME);
+	tp->cm_timer = dapl_os_get_env_val("DAPL_UCM_TIMER", DCM_CM_TIMER);
+	/* default = 11-bit, 2KB entries; 18 bit, 256KB total */
+	tp->cm_entry_bits = dapl_os_get_env_val("DAPL_UCM_ENTRY_BITS", UCM_ENTRY_BITS);
+	tp->cm_array_bits = DAPL_MAX(dapl_os_get_env_val("DAPL_UCM_ARRAY_BITS", UCM_ARRAY_BITS), tp->cm_entry_bits);
+	array_sz = UCM_ARRAY_SIZE(tp->cm_array_bits, tp->cm_entry_bits);
+	entry_sz = UCM_ENTRY_SIZE(tp->cm_entry_bits);
 
-    	tp->rch = ibv_create_comp_channel(hca->ib_hca_handle);
-	if (!tp->rch) 
+	tp->pd = ibv_alloc_pd(hca->ib_hca_handle);
+        if (!tp->pd) {
+		dapl_log(DAPL_DBG_TYPE_ERR,
+			 "UCM: CM service: ERR ibv_pd (%s)\n",
+			 strerror(errno));
 		goto bail;
+        }
+        
+        dapl_log(DAPL_DBG_TYPE_CM,
+                 "CM (%d+%d)- pd %p Timers(s): req %d rtu %d wait %d -"
+        	 " idx(%d,%d): Array %d Entry %d = %d\n",
+                 hlen, mlen, tp->pd, tp->rep_time/1000, tp->rtu_time/1000,
+                 tp->wait_time/1000, tp->cm_array_bits,
+                 tp->cm_entry_bits, array_sz, entry_sz,
+                 array_sz * entry_sz);
+
+        /* default == 2K idx size, grow to 256K total CM slots  */
+        tp->cm_idxr = dapl_os_alloc(sizeof(void*) * array_sz);
+        if (!tp->cm_idxr) {
+        	dapl_log(DAPL_DBG_TYPE_ERR,
+        		 "UCM: CM service: ERR (%s) idx_array alloc %d\n",
+        		 strerror(errno), sizeof(void*) * array_sz);
+        	goto bail;
+        }
+        (void)dapl_os_memzero(tp->cm_idxr, sizeof(void*) * array_sz);
+
+        /* allocate first index array for cm entries, 2K by default */
+        tp->cm_idxr[0] = dapl_os_alloc(sizeof(void*) * entry_sz);
+        if (!tp->cm_idxr[0]) {
+        	dapl_log(DAPL_DBG_TYPE_ERR,
+        		 "UCM: CM service: ERR (%s) idx_entry alloc %d\n",
+        		 strerror(errno), sizeof(void*) * entry_sz);
+        	goto bail;
+        }
+        (void)dapl_os_memzero(tp->cm_idxr[0], sizeof(void*) * entry_sz);
+
+        tp->rch = ibv_create_comp_channel(hca->ib_hca_handle);
+	if (!tp->rch) {
+		dapl_log(DAPL_DBG_TYPE_ERR,
+			 "UCM: CM service: ERR ibv_comp_channel (%s)\n",
+			 strerror(errno));
+		goto bail;
+	}
 	dapls_config_comp_channel(tp->rch);
 
 	tp->scq = ibv_create_cq(hca->ib_hca_handle, tp->cqe, hca, NULL, 0);
-	if (!tp->scq) 
+	if (!tp->scq) {
+		dapl_log(DAPL_DBG_TYPE_ERR,
+			 "UCM: CM service: ERR ibv_cq_s (%s)\n",
+			 strerror(errno));
 		goto bail;
-        
+	}
 	tp->rcq = ibv_create_cq(hca->ib_hca_handle, tp->cqe, hca, tp->rch, 0);
-	if (!tp->rcq) 
+	if (!tp->rcq) {
+		dapl_log(DAPL_DBG_TYPE_ERR,
+			 "UCM: CM service: ERR ibv_cq_r (%s)\n",
+			 strerror(errno));
 		goto bail;
+	}
 
-	if(ibv_req_notify_cq(tp->rcq, 0))
-		goto bail; 
+	if(ibv_req_notify_cq(tp->rcq, 0)) {
+		dapl_log(DAPL_DBG_TYPE_ERR,
+			 "UCM: CM service: ERR ibv_notify (%s)\n",
+			 strerror(errno));
+		goto bail;
+	}
  
 	dapl_os_memzero((void *)&qp_create, sizeof(qp_create));
 	qp_create.qp_type = IBV_QPT_UD;
@@ -540,43 +607,63 @@ static int ucm_service_create(IN DAPL_HCA *hca)
 	qp_create.recv_cq = tp->rcq;
 	qp_create.cap.max_send_wr = qp_create.cap.max_recv_wr = tp->qpe;
 	qp_create.cap.max_send_sge = qp_create.cap.max_recv_sge = 1;
-	qp_create.cap.max_inline_data = tp->max_inline_send;
+	qp_create.cap.max_inline_data = tp->ib_cm.max_inline;
 	qp_create.qp_context = (void *)hca;
 
 	tp->qp = ibv_create_qp(tp->pd, &qp_create);
-	if (!tp->qp) 
-                goto bail;
+	if (!tp->qp) {
+		dapl_log(DAPL_DBG_TYPE_ERR,
+			 "UCM: CM service: ERR ibv_qp (%s)\n",
+			 strerror(errno));
+		goto bail;
+	}
 
-	tp->ah = (ib_ah_handle_t*) dapl_os_alloc(sizeof(ib_ah_handle_t) * 0xffff);
-	tp->sid = (uint8_t*) dapl_os_alloc(sizeof(uint8_t) * 0xffff);
+	tp->ah = (ib_ah_handle_t*) dapl_os_alloc(sizeof(ib_ah_handle_t) * DCM_AH_SPACE);
+	tp->sid = (uint8_t*) dapl_os_alloc(UCM_SID_SPACE/UCM_SID_ENTRY);
 	tp->rbuf = (void*) dapl_os_alloc((mlen + hlen) * tp->qpe);
 	tp->sbuf = (void*) dapl_os_alloc(mlen * tp->qpe);
 	tp->s_hd = tp->s_tl = 0;
 
-	if (!tp->ah || !tp->rbuf || !tp->sbuf || !tp->sid)
+	if (!tp->ah || !tp->rbuf || !tp->sbuf) {
+		dapl_log(DAPL_DBG_TYPE_ERR,
+			 "UCM: CM service: ERR malloc ah,bufs(%s)\n",
+			 strerror(errno));
 		goto bail;
+	}
 
-	(void)dapl_os_memzero(tp->ah, (sizeof(ib_ah_handle_t) * 0xffff));
-	(void)dapl_os_memzero(tp->sid, (sizeof(uint8_t) * 0xffff));
-	tp->sid[0] = 1; /* resv slot 0, 0 == no ports available */
+	(void)dapl_os_memzero(tp->ah, (sizeof(ib_ah_handle_t) * DCM_AH_SPACE));
+	(void)dapl_os_memzero(tp->sid, UCM_SID_SPACE/UCM_SID_ENTRY);
+	tp->sid[0] = 0x1; /* resv slot 0 */
 	(void)dapl_os_memzero(tp->rbuf, ((mlen + hlen) * tp->qpe));
 	(void)dapl_os_memzero(tp->sbuf, (mlen * tp->qpe));
 
 	tp->mr_sbuf = ibv_reg_mr(tp->pd, tp->sbuf, 
 				 (mlen * tp->qpe),
 				 IBV_ACCESS_LOCAL_WRITE);
-	if (!tp->mr_sbuf)
+	if (!tp->mr_sbuf) {
+		dapl_log(DAPL_DBG_TYPE_ERR,
+			 "UCM: CM service: ERR ibv_mr sbuf (%s)\n",
+			 strerror(errno));
 		goto bail;
+	}
 
 	tp->mr_rbuf = ibv_reg_mr(tp->pd, tp->rbuf, 
 				 ((mlen + hlen) * tp->qpe),
 				 IBV_ACCESS_LOCAL_WRITE);
-	if (!tp->mr_rbuf)
+	if (!tp->mr_rbuf) {
+		dapl_log(DAPL_DBG_TYPE_ERR,
+			 "UCM: CM service: ERR ibv_mr rbuf (%s)\n",
+			 strerror(errno));
 		goto bail;
+	}
 	
 	/* modify UD QP: init, rtr, rts */
-	if ((dapls_modify_qp_ud(hca, tp->qp)) != DAT_SUCCESS)
+	if ((dapls_modify_qp_ud(hca, tp->qp)) != DAT_SUCCESS) {
+		dapl_log(DAPL_DBG_TYPE_ERR,
+			 "UCM: CM service: ERR modify_qp_ud (%s)\n",
+			 strerror(errno));
 		goto bail;
+	}
 
 	/* post receive buffers, setup head, tail pointers */
 	recv_wr.next = NULL;
@@ -589,8 +676,12 @@ static int ucm_service_create(IN DAPL_HCA *hca)
 	for (i = 0; i < tp->qpe; i++) {
 		recv_wr.wr_id = (uintptr_t) (rbuf + hlen);
 		sge.addr = (uintptr_t) rbuf;
-		if (ibv_post_recv(tp->qp, &recv_wr, &recv_err))
+		if (ibv_post_recv(tp->qp, &recv_wr, &recv_err)) {
+			dapl_log(DAPL_DBG_TYPE_ERR,
+				 "UCM: CM service: ERR ibv_post_rcv (%s)\n",
+				 strerror(errno));
 			goto bail;
+		}
 		rbuf += sge.length;
 	}
 
