@@ -50,6 +50,7 @@ int mcm_rtu_ms = 2000;
 int mcm_dreq_ms = 1000;
 int mcm_proxy_in = 1;
 
+extern int mix_buffer_sg_po2;
 extern int mcm_rx_entries;
 extern uint64_t system_guid;
 extern char gid_str[INET6_ADDRSTRLEN];
@@ -220,7 +221,8 @@ int mcm_init_cm_service(mcm_ib_dev_t *md)
 	qp_create.recv_cq = md->rcq;
 	qp_create.cap.max_send_wr = qp_create.cap.max_recv_wr = md->qpe;
 	qp_create.cap.max_send_sge = qp_create.cap.max_recv_sge = 1;
-	qp_create.cap.max_inline_data = mlen + sizeof(struct ibv_grh);
+	if (md->indata)
+		qp_create.cap.max_inline_data = mlen + sizeof(struct ibv_grh);
 	qp_create.qp_context = (void *)md;
 
 	md->qp = ibv_create_qp(md->pd, &qp_create);
@@ -283,11 +285,12 @@ int mcm_init_cm_service(mcm_ib_dev_t *md)
 	/* save qp_num as part of ia_address, network order */
 	md->addr.qpn = htonl(md->qp->qp_num);
 
-	mlog(0, " IB LID 0x%x PORT %d GID %s QPN 0x%x: mic%d -> %s - %s, mic_ver %d\n",
+	mlog(0, " IB LID 0x%x PORT %d GID %s QPN 0x%x: mic%d -> %s - %s, mic_ver %d %s\n",
 		ntohs(md->addr.lid), md->port,
 		inet_ntop(AF_INET6, md->addr.gid, gid_str, sizeof(gid_str)),
 		md->qp->qp_num, md->mc->scif_id - 1, md->ibdev->name,
-	        md->addr.ep_map == MIC_SSOCK_DEV ? "MSS":"MXS", md->mc->ver);
+	        md->addr.ep_map == MIC_SSOCK_DEV ? "MSS":"MXS", md->mc->ver,
+	        md->mc->ver == MIX_COMP ? "WARN: OLD MPSS -> RUN COMP MODE":"");
 
 	return 0;
 bail:
@@ -530,6 +533,7 @@ mcm_cm_t *m_cm_create(mcm_scif_dev_t *smd, mcm_qp_t *m_qp, dat_mcm_addr_t *r_add
 	cm->md = smd->md;
 	cm->msg.ver = htons(DAT_MCM_VER);
 	cm->msg.sqpn = smd->md->addr.qpn; /* ucm, in network order */
+	cm->msg.seg_sz = mix_buffer_sg_po2;
 #ifdef MPXYD_LOCAL_SUPPORT
 	cm->msg.sys_guid = system_guid; /* network order */
 #else
@@ -798,7 +802,8 @@ static int mcm_send(mcm_ib_dev_t *md, dat_mcm_msg_t *msg, DAT_PVOID p_data, DAT_
         wr.opcode = IBV_WR_SEND;
         wr.wr_id = (unsigned long)md->s_hd;
 	wr.send_flags = (wr.wr_id % md->signal) ? 0 : IBV_SEND_SIGNALED;
-	wr.send_flags |= IBV_SEND_INLINE;
+	if (md->indata)
+		wr.send_flags |= IBV_SEND_INLINE;
 
         sge.length = len;
         sge.lkey = md->mr_sbuf->lkey;
