@@ -45,10 +45,11 @@ int mix_max_msg_mb = 16;
 int mix_inline_threshold = 256;
 int mix_eager_completion = 1;
 int mcm_ib_inline = MCM_IB_INLINE;
-int mcm_rw_signal = 10;
-int mcm_rr_signal = 10;
-int mcm_rr_max = 50;
-int mcm_wrc_max = 5;
+int mcm_rw_signal = 1;
+int mcm_rr_signal = 1;
+int mcm_rr_max = 40;	/* PO rdma read in-process limit */
+int mcm_wr_max = 10; 	/* PO WR batch processing limit */
+int mcm_wc_max = 10; 	/* PO-PI poll_cq WC processing limit */
 int mcm_tx_entries = MCM_WRC_QLEN; /* power of 2, default = 512 */
 int mcm_rx_entries = MCM_WRC_QLEN;
 int mcm_rx_cq_size = MCM_WRC_QLEN;
@@ -105,7 +106,7 @@ void mix_close_device(mcm_ib_dev_t *md, mcm_scif_dev_t *smd)
 	while (smd->th_ref_cnt) {
 		mlog(1, " waiting for SMD %p ref_cnt (%d) = 0\n", smd, smd->th_ref_cnt);
 		mpxy_unlock(&md->slock);
-		sleep_usec(1000);
+		sched_yield();
 		mpxy_lock(&md->slock);
 	}
 	mpxy_destroy_smd(smd);
@@ -540,7 +541,8 @@ static int mix_cq_create(mcm_scif_dev_t *smd, dat_mix_cq_t *pmsg)
 	pmsg->cq_id = new_mcq->cq_id;
 	pmsg->cq_ctx = (uint64_t)new_mcq;
 	pmsg->hdr.status = MIX_SUCCESS;
-	mlog(8, " new cq_id %d, mpxyd_ctx=%p\n", pmsg->cq_id, pmsg->cq_ctx);
+	mlog(8, " new cq_id %d, mpxyd_ctx %p, mcm_ctx %p\n",
+		pmsg->cq_id, pmsg->cq_ctx, new_mcq->cq_ctx);
 	MCNTR(smd->md, MCM_CQ_CREATE);
 	goto resp;
 err:
@@ -663,7 +665,8 @@ int m_qp_create_pi(mcm_scif_dev_t *smd, struct mcm_qp *m_qp)
 	qp_create.cap.max_recv_sge = 1;
 	qp_create.cap.max_send_wr = mcm_tx_entries;
 	qp_create.cap.max_send_sge = 1;
-	qp_create.cap.max_inline_data = mcm_ib_inline;
+	if (smd->md->indata)
+		qp_create.cap.max_inline_data = mcm_ib_inline;
 	qp_create.qp_type = IBV_QPT_RC;
 	qp_create.send_cq = m_qp->m_cq_tx->ib_cq; /* signal rate adjusted to avoid CQ overrun */
 	qp_create.recv_cq = m_qp->m_cq_rx->ib_cq;
@@ -792,7 +795,8 @@ static int m_qp_create(mcm_scif_dev_t *smd,
 	m_qp->qp_attr2.max_recv_sge = attr->cap.max_recv_sge;
 	m_qp->qp_attr2.max_send_wr = attr->cap.max_send_wr;
 	m_qp->qp_attr2.max_send_sge = attr->cap.max_send_sge;
-	m_qp->qp_attr2.max_inline_data = attr->cap.max_inline_data;
+	if (smd->md->indata)
+		m_qp->qp_attr2.max_inline_data = attr->cap.max_inline_data;
 
 	/* don't post recv msgs until RTU, if local MXS or local MSS to remote MXS */
 	if (smd->md->addr.ep_map == MIC_XSOCK_DEV)
@@ -853,7 +857,8 @@ static int mix_qp_create(mcm_scif_dev_t *smd, dat_mix_qp_t *pmsg)
 	qp_create.cap.max_recv_sge = 1;
 	qp_create.cap.max_send_wr = mcm_tx_entries; /* proxy-out WR separate from client WR */
 	qp_create.cap.max_send_sge = min(pmsg->qp_t.max_send_sge, DAT_MIX_SGE_MAX);
-	qp_create.cap.max_inline_data = mcm_ib_inline;
+	if (smd->md->indata)
+		qp_create.cap.max_inline_data = mcm_ib_inline;
 	qp_create.qp_type = IBV_QPT_RC;
 
 	mlog(8, " QP_t - client max_wr %d, proxy_out max_wr %d\n",
@@ -1661,12 +1666,13 @@ static int mix_cm_rep_out(mcm_scif_dev_t *smd, dat_mix_cm_t *pmsg, scif_epd_t sc
 	mcm_ntoh_wrc(&m_cm->m_qp->wrc_rem, (mcm_wrc_info_t *)m_cm->msg.p_proxy); /* save peer PI WRC info */
 
 	mlog(2, " WRC: m_qp %p - WR 0x%Lx rkey 0x%x ln %d, sz %d end %d"
-		" WC 0x%Lx rkey 0x%x ln %d, sz %d end %d\n",
+		" WC 0x%Lx rkey 0x%x ln %d, sz %d end %d sg_po2 %d\n",
 	     m_cm->m_qp, m_cm->m_qp->wrc.wr_addr, m_cm->m_qp->wrc.wr_rkey,
 	     m_cm->m_qp->wrc.wr_len, m_cm->m_qp->wrc.wr_sz,
 	     m_cm->m_qp->wrc.wr_end, m_cm->m_qp->wrc.wc_addr,
 	     m_cm->m_qp->wrc.wc_rkey, m_cm->m_qp->wrc.wc_len,
-	     m_cm->m_qp->wrc.wc_sz, m_cm->m_qp->wrc.wc_end);
+	     m_cm->m_qp->wrc.wc_sz, m_cm->m_qp->wrc.wc_end,
+	     m_cm->msg.seg_sz);
 
 	mlog(2, " WRC_rem: m_qp %p - WR 0x%Lx rkey 0x%x ln %d, sz %d end %d"
 		" WC 0x%Lx rkey 0x%x ln %d, sz %d end %d\n",
@@ -1905,7 +1911,7 @@ int mix_cm_disc_in(mcm_cm_t *m_cm)
 /* Post SEND message request, IB send or rdma write, proxy out to remote direct endpoint */
 static int mix_proxy_out(mcm_scif_dev_t *smd, dat_mix_sr_t *pmsg, mcm_qp_t *m_qp)
 {
-	int len, ret, l_start, l_end, retries, wc_err = IBV_WC_GENERAL_ERR;
+	int len, ret = 0, l_start, l_end, retries, wc_err = IBV_WC_GENERAL_ERR;
 	struct mcm_wr *m_wr;
 
 	if (!(pmsg->hdr.flags & MIX_OP_INLINE))
@@ -1954,7 +1960,7 @@ retry_mr:
 		l_start = 64;
 	l_end = l_start + len;
 
-	if (l_start < smd->m_tl && l_end > smd->m_tl) {
+	if (pb_full(l_start, l_end, smd->m_hd, smd->m_tl, len, smd->m_len)) {
 		if (!retries) {
 			MCNTR(smd->md, MCM_MX_MR_STALL);
 			write(smd->md->mc->tx_pipe[1], "w", sizeof("w"));
@@ -1984,11 +1990,11 @@ retry_mr:
 			mpxy_unlock(&smd->tblock);
 			goto bail;
 		}
-		mpxy_unlock(&m_qp->txlock);
 		mpxy_unlock(&smd->tblock);
-		sleep_usec(10000);
-		mpxy_lock(&smd->tblock);
+		mpxy_unlock(&m_qp->txlock);
+		sched_yield();
 		mpxy_lock(&m_qp->txlock);
+		mpxy_lock(&smd->tblock);
 		goto retry_mr;
 	}
 
@@ -1998,7 +2004,10 @@ retry_mr:
 			smd->m_buf, smd->m_hd, smd->m_tl,
 			smd->m_buf + smd->m_len, l_start, len, retries);
 	}
-
+#ifdef MCM_PROFILE
+	if (l_end < smd->m_hd)
+		smd->m_hd_ro++;
+#endif
 	m_wr->sg->addr = (uint64_t)(smd->m_buf + l_start);
 	m_wr->sg->lkey = smd->m_mr->lkey;
 	m_wr->sg->length = len;
@@ -2041,7 +2050,7 @@ retry_mr:
 	}
 	mlog(4, " inline data rcv'ed %d bytes\n", len);
 
-	if (len <= mcm_ib_inline)
+	if ((smd->md->indata) && (len <= mcm_ib_inline))
 		m_wr->wr.send_flags |= IBV_SEND_INLINE;
 
 	/* signal TX, CQ thread, this WR data ready, post pending */
@@ -2065,6 +2074,7 @@ bail:
 		wc.vendor_err = ret;
 		mix_dto_event(m_qp->ib_qp2->send_cq->cq_context, &wc, 1);
 	}
+	sched_yield();
 	return ret;
 }
 
@@ -2108,7 +2118,7 @@ static int mix_post_send(mcm_scif_dev_t *smd, dat_mix_sr_t *pmsg)
 		pmsg->wr.num_sge, pmsg->wr.opcode, pmsg->wr.send_flags,
 		m_qp->post_cnt,	m_qp->post_sig_cnt, m_qp->comp_cnt,
 		pmsg->hdr.flags & MIX_OP_INLINE ? 1:0,
-		pmsg->wr.opcode == IBV_WR_SEND ? "SND":"WR",
+		pmsg->wr.opcode == IBV_WR_SEND ? "SND":"RW",
 		m_qp->wrc_rem.wr_len ? "PROXY_OUT_IN":"PROXY_OUT");
 
 	return (mix_proxy_out(smd, pmsg, m_qp));
