@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2014 Intel Corporation. All rights reserved.
+ * Copyright (c) 2012-2015 Intel Corporation. All rights reserved.
  *
  * This software is available to you under the OpenIB.org BSD license
  * below:
@@ -35,9 +35,10 @@
 #include "mpxyd.h"
 
 extern int log_level;
-extern int mcm_wrc_max;
+extern int mcm_wc_max;
 extern int mcm_rr_signal;
 extern int mcm_rr_max;
+extern int mcm_wr_max;
 extern int mcm_rx_entries;
 extern int mcm_ib_signal_rate;
 extern uint64_t system_guid;
@@ -66,13 +67,20 @@ void m_pi_destroy_bpool(struct mcm_qp *m_qp)
 		free((void*)m_qp->wrc.wr_addr);
 		m_qp->wrc.wr_addr = 0;
 	}
-
 	m_pi_destroy_wc_q(m_qp);
 
 	if (m_qp->sr_buf) {
 		free(m_qp->sr_buf);
 		m_qp->sr_buf = 0;
 	}
+	if (m_qp->wc_buf_rx_mr) {
+		ibv_dereg_mr(m_qp->wc_buf_rx_mr);
+		m_qp->wc_buf_rx_mr = NULL;
+	}
+	if (m_qp->wc_buf_rx) {
+		free(m_qp->wc_buf_rx);
+		m_qp->wc_buf_rx = 0;
+	}
 }
 
 int m_pi_create_wr_q(struct mcm_qp *m_qp, int entries)
@@ -122,7 +130,32 @@ int m_pi_create_wr_q(struct mcm_qp *m_qp, int entries)
 	mlog(4, " SCIF_mr for wr_rbuf addr %p, off 0x%llx, len %d, entries %d\n",
 		m_qp->wrc.wr_addr, m_qp->wr_off_r, ALIGN_PAGE(m_qp->wrc.wr_len), entries);
 
+	if (!m_qp->smd->md->indata) {
+		if (posix_memalign((void **)&m_qp->wc_buf_rx, 4096, entries * sizeof(mcm_wc_rx_t))) {
+			mlog(0, "failed to allocate wc_buf_rx, m_qp=%p, wr_rx_len=%d, entries=%d\n",
+				 m_qp, entries * sizeof(mcm_wc_rx_t), entries);
+			goto err;
+		}
+		memset(m_qp->wc_buf_rx, 0, entries * sizeof(mcm_wc_rx_t));
+
+		m_qp->wc_buf_rx_mr = ibv_reg_mr(m_qp->smd->md->pd, (void*)m_qp->wc_buf_rx,
+						entries * sizeof(mcm_wc_rx_t),
+					        IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE);
+		if (!m_qp->wc_buf_rx_mr) {
+			mlog(0, " IB_register addr=%p,%d failed %s\n",
+				m_qp->wc_buf_rx_mr->addr,
+				entries * sizeof(mcm_wc_rx_t),
+				strerror(errno));
+			goto err;
+		}
+		mlog(4, " no inline support: WC_buf_rx pool %p, LEN %d\n",
+			m_qp->wc_buf_rx, entries * sizeof(mcm_wc_rx_t) );
+		mlog(4, " no inline support: WC_buf_rx_mr = %x\n", m_qp->wc_buf_rx_mr);
+	}
 	return 0;
+err:
+	m_pi_destroy_bpool(m_qp);
+	return -1;
 }
 
 int m_pi_create_wc_q(struct mcm_qp *m_qp, int entries)
@@ -274,8 +307,21 @@ static int m_pi_buf_hd(mcm_scif_dev_t *smd, int m_idx, struct mcm_wr_rx *m_wr_rx
 	smd->m_buf_wc_r[smd->m_buf_hd_r].m_idx = m_idx;
 	smd->m_buf_wc_r[smd->m_buf_hd_r].done = 0;
 #ifdef MCM_PROFILE
+	smd->m_buf_wc_r[smd->m_buf_hd_r].hd = smd->m_hd_r;
+	smd->m_buf_wc_r[smd->m_buf_hd_r].tl = smd->m_tl_r;
+	smd->m_buf_wc_r[smd->m_buf_hd_r].ref++;
 	smd->m_buf_wc_r[smd->m_buf_hd_r].ts = mcm_ts_us();
 	smd->m_buf_wc_r[smd->m_buf_hd_r].wr = (void *) m_wr_rx;
+	if ((smd->m_hd_ro_r != smd->m_tl_ro_r) && (smd->m_hd_r >= smd->m_tl_r)) {
+		struct mcm_qp *m_qp = (struct mcm_qp*)m_wr_rx->context;
+		mlog(0, " ERR: [%d:%d:%d] PI_buf HD(%d) passed TL(%d): w_tl %d w_hd %d:"
+			" wc_r[%d].m_idx=%x, m_tl %x m_hd %x\n",
+			smd->md->mc->scif_id, smd->entry.tid, m_qp->r_entry.tid,
+			smd->m_hd_ro_r, smd->m_tl_ro_r, smd->m_buf_tl_r + 1,
+			smd->m_buf_hd_r, smd->m_buf_hd_r, m_idx,
+			smd->m_tl_r, smd->m_hd_r);
+		mcm_check_io();
+	}
 #endif
 	return 0;
 }
@@ -283,17 +329,23 @@ static int m_pi_buf_hd(mcm_scif_dev_t *smd, int m_idx, struct mcm_wr_rx *m_wr_rx
 /* called with smd->rblock */
 static void m_pi_buf_tl(mcm_scif_dev_t *smd, int m_idx, struct mcm_wr_rx *m_wr_rx)
 {
-	int s_idx, idx;
+	int idx;
 	int busy = 0, match = 0;
-
+#ifdef MCM_PROFILE
+	off_t tl_sav = smd->m_tl_r;
+	uint32_t now = mcm_ts_us();
+#endif
 	idx = (smd->m_buf_tl_r + 1) & smd->m_buf_end_r; /* tl == hd is empty */
-	s_idx = idx;
 
 	/* mark m_idx complete, move proxy buffer tail until busy slot */
 	while ((!match || !busy) && smd->m_buf_tl_r != smd->m_buf_hd_r) {
 		if (smd->m_buf_wc_r[idx].m_idx == m_idx) {
 			smd->m_buf_wc_r[idx].done = 1;
 			match = 1;
+#ifdef MCM_PROFILE
+			smd->m_buf_wc_r[idx].ref--;
+			smd->m_buf_wc_r[idx].ts = now - smd->m_buf_wc_r[idx].ts;
+#endif
 		}
 		if (smd->m_buf_wc_r[idx].done && !busy) {
 			smd->m_tl_r = smd->m_buf_wc_r[idx].m_idx;
@@ -308,33 +360,18 @@ static void m_pi_buf_tl(mcm_scif_dev_t *smd, int m_idx, struct mcm_wr_rx *m_wr_r
 
 		idx = (idx + 1) & smd->m_buf_end_r;
 	}
+
 #ifdef MCM_PROFILE
-	if ((log_level < 4) || (smd->m_buf_wc_r[s_idx].done))
-		return;
+	if (tl_sav > smd->m_tl_r)
+		smd->m_tl_ro_r++;
 
-	if (smd->m_buf_wc_r[s_idx].done) {
-		mlog(0x10," [%d:%d] InOrder: m_wc %p: tl %d hd %d wc[%d].m_idx=0x%x "
-			  "%s m_idx 0x%x %s wr %p \n",
-			  smd->md->mc->scif_id, smd->entry.tid, smd->m_buf_wc_r,
-			  s_idx, smd->m_buf_hd_r, s_idx, smd->m_buf_wc_r[s_idx].m_idx,
-			  smd->m_buf_wc_r[s_idx].m_idx == m_idx ? "==":"!=",
-			  m_idx, smd->m_buf_wc_r[s_idx].done ? "DONE":"BUSY",
-			  smd->m_buf_wc_r[s_idx].wr);
-		return;
-	}
-
-	for (idx = s_idx;;) {
-		uint32_t now = mcm_ts_us();
-
-		mlog(4," [%d:%d] OutOfOrder: tl %d hd %d wc[%d].m_idx=0x%x %s m_idx 0x%x %s %d us\n",
+	if (!match) {
+		mlog(0, " [%d:%d] ERR: m_tl 0x%x m_hd 0x%x"
+		        "- m_wc: tl %d hd %d - m_idx=0x%x NO MATCH\n",
 			smd->md->mc->scif_id, smd->entry.tid,
-			s_idx, smd->m_buf_hd_r, idx, smd->m_buf_wc_r[idx].m_idx,
-			smd->m_buf_wc_r[idx].m_idx == m_idx ? "==":"!=",
-			m_idx, smd->m_buf_wc_r[idx].done ? "DONE":"BUSY",
-			now - smd->m_buf_wc_r[idx].ts);
-		if (idx == smd->m_buf_hd_r)
-			break;
-		idx = (idx + 1) & smd->m_buf_end_r;
+			smd->m_tl_r, smd->m_hd_r, smd->m_buf_tl_r,
+			(smd->m_buf_hd_r + 1) & smd->m_buf_end_r, m_idx);
+		mcm_check_io();
 	}
 #endif
 }
@@ -434,6 +471,7 @@ static int m_pi_send_wc(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx, int status
 	struct ibv_sge sge;
 	struct wrc_idata wrc;
 	struct mcm_wc_rx wc_rx;
+	struct mcm_wc_rx *wc_rx_ptr;
 	struct ibv_qp *ib_qp;
 	int wc_idx, ret;
 
@@ -457,32 +495,44 @@ static int m_pi_send_wc(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx, int status
 	wrc.id = (uint16_t)wc_idx;  /* imm_data for proxy_out rcv engine */
 	wrc.type = M_WC_TYPE;
 	wrc.flags = 0;
-	mcm_hton_wc_rx(&wc_rx, wr_rx, m_qp->wr_tl_r, status);
+
+	if (m_qp->smd->md->indata) {
+		wc_rx_ptr = &wc_rx;
+		sge.lkey = 0; /* inline doesn't need registered */
+	} else {
+		wc_rx_ptr = (struct mcm_wc_rx *)
+			    (m_qp->wc_buf_rx + (sizeof(struct mcm_wc_rx) * wc_idx));
+		sge.lkey = m_qp->wc_buf_rx_mr->lkey;
+	}
+	sge.addr = (uint64_t)(uintptr_t) wc_rx_ptr;
+	sge.length = (uint32_t) sizeof(struct mcm_wc_rx);
+
+	mcm_hton_wc_rx(wc_rx_ptr, wr_rx, m_qp->wr_tl_r, status);
 
 	/* P2P on same system, keep it local */
 	if (htonll(m_qp->cm->msg.sys_guid) == system_guid)
 		return (m_pi_send_wc_local(m_qp, wr_rx, wc_idx));
 
 	/* send back a WC with error */
-	wr.next = 0;
+	memset(&wr, 0, sizeof(struct ibv_send_wr));
 	wr.opcode = IBV_WR_RDMA_WRITE_WITH_IMM;
 	wr.imm_data = htonl(*(uint32_t *)&wrc);
 	wr.num_sge = 1;
-	wr.send_flags = IBV_SEND_SIGNALED | IBV_SEND_INLINE; /* mcm_wc_rx, 160 bytes */
+	wr.send_flags |= IBV_SEND_SIGNALED;
 	wr.wr_id = WRID_SET(wr_rx, WRID_RX_RW_IMM);
 	wr.wr.rdma.rkey = m_qp->wrc_rem.wc_rkey;
 	wr.wr.rdma.remote_addr = (uint64_t)(uintptr_t)((struct mcm_wc_rx *)
 				 (m_qp->wrc_rem.wc_addr + (m_qp->wrc_rem.wc_sz * wc_idx)));
 	wr.sg_list = &sge;
-	sge.addr = (uint64_t)(uintptr_t) &wc_rx;
-	sge.length = (uint32_t) sizeof(struct mcm_wc_rx);
-	sge.lkey = 0; /* inline doesn't need registered */
+
+	if (m_qp->smd->md->indata)
+		wr.send_flags |= IBV_SEND_INLINE;
 
 	mlog(4, " WC: RW_imm post: wr_id[%d] %Lx sglist %p sge %d op %d flgs %x"
-		" idata %x WR_rem = raddr %p rkey %x ln %d\n",
+		" idata %x WR_rem = raddr %p rkey %x ln %d op %x\n",
 		wr_rx->w_idx, wr.wr_id, wr.sg_list, wr.num_sge, wr.opcode,
 		wr.send_flags, ntohl(wr.imm_data), wr.wr.rdma.remote_addr,
-		wr.wr.rdma.rkey, sge.length);
+		wr.wr.rdma.rkey, sge.length, wr_rx->wr.opcode);
 
 	/* MXS -> MSS or HST, PI service will be on QP1 */
 	if (MXS_EP(&m_qp->smd->md->addr) &&
@@ -547,8 +597,8 @@ static void m_pi_post_writeto(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_sig, str
 		MCNTR(smd->md, MCM_QP_READ_DONE);
 
 		/* if SR or RW_imm, need a posted receive */
-		if ((wr_rx->wr.opcode & IBV_WR_SEND) ||
-		    (wr_rx->wr.opcode & IBV_WR_RDMA_WRITE_WITH_IMM)) {
+		if ((wr_rx->wr.opcode == IBV_WR_SEND) ||
+		    (wr_rx->wr.opcode == IBV_WR_RDMA_WRITE_WITH_IMM)) {
 			m_sr = m_pi_get_sr(m_qp, wr_rx->w_idx);
 			if (!m_sr) {
 				mlog(0, " WARNING: SR stalled, no RCV messages posted"
@@ -562,11 +612,12 @@ static void m_pi_post_writeto(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_sig, str
 			num_sge = m_sr->num_sge;
 			sg_len = m_sr->sg[0].length;
 			r_off = m_sr->sg[0].addr; /* post recv buffer address */
-			mlog(4, " WR SR or RW_IMM: m_sr[%d] %p -> scif r_off %Lx ln %d\n",
+			mlog(4, " WR %s: m_sr[%d] %p -> scif r_off %Lx ln %d\n",
+				(wr_rx->wr.opcode == IBV_WR_SEND) ? "SR":"RW_IMM",
 				m_sr->s_idx, m_sr, r_off, sg_len);
 		}
 		/* need to translate to rdma write dst */
-		if (!(wr_rx->wr.opcode & IBV_WR_SEND)) {
+		if (!(wr_rx->wr.opcode == IBV_WR_SEND)) {
 			num_sge = 1;
 			sg_len = wr_rx->sg[2].length;
 			r_off = m_pi_mr_trans(smd, wr_rx->wr.wr.rdma.remote_addr,
@@ -648,7 +699,6 @@ static void m_pi_post_writeto(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_sig, str
 
 		wr_idx = (wr_idx + 1) & m_qp->wrc.wr_end; /* next WR */
 	}
-	write(smd->md->mc->rx_pipe[1], "w", sizeof "w"); /* signal rx_thread */
 	return;
 bail:
 	/* report error via WC back to proxy-out */
@@ -672,11 +722,10 @@ void m_pi_req_event(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx, struct ibv_wc
 {
 	mlog(4, " WR_rx[%d] %p %s complete po-addr=%p ln=%d, key=%x ctx=%Lx\n",
 		wr_rx->w_idx, wr_rx,
-		wc->opcode == IBV_WC_RDMA_READ ? "RR":"RW_IMM WC",
+		wr_rx->wr.opcode == IBV_WR_RDMA_READ ? "RR":"RW_IMM WC",
 		wr_rx->sg[0].addr, wr_rx->sg[0].length,
 		wr_rx->sg[0].lkey, wr_rx->context);
 
-	mpxy_lock(&m_qp->rxlock);
 	if (wc->status && (wc->status != IBV_WC_WR_FLUSH_ERR)) {
 		char *sbuf = (char*)wr_rx->sg[1].addr;
 
@@ -694,6 +743,7 @@ void m_pi_req_event(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx, struct ibv_wc
 		     wr_rx->sg[0].addr, wr_rx->sg[0]. lkey,wr_rx->sg[0].length);
 
 		/* send WC with ERR to RW initiator, hold rxlock */
+		mpxy_lock(&m_qp->rxlock);
 		if (m_pi_send_wc(m_qp, wr_rx, wc->status))
 			mlog(0, "WR ERR: proxy-in to proxy-out WC send failed\n");
 		mpxy_unlock(&m_qp->rxlock);
@@ -701,16 +751,21 @@ void m_pi_req_event(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx, struct ibv_wc
 	}
 	/* RR complete, ready for SCIF_writeto to complete RW or SR */
 	if (type == WRID_RX_RR) {
+		mpxy_lock(&m_qp->rxlock);
 		m_pi_post_writeto(m_qp, wr_rx, wc);
+		mpxy_unlock(&m_qp->rxlock);
+		write(m_qp->smd->md->mc->rx_pipe[1], "w", sizeof "w"); /* signal rx_thread */
+		sched_yield();
 	} else {
 		mlog(4, " WR_rx[%d] %p flgs %x %s complete - WR tl %d tl_wt %d hd %d\n",
 			  wr_rx->w_idx, wr_rx, wr_rx->flags,
 			  wc->opcode == IBV_WC_RDMA_READ ? "RR":"RW_IMM WC",
 			  m_qp->wr_tl_r, m_qp->wr_tl_r_wt, m_qp->wr_hd_r);
 
+		mpxy_lock(&m_qp->rxlock);
 		m_qp->pi_rw_cnt--; /* pending WC post_send */
+		mpxy_unlock(&m_qp->rxlock);
 	}
-	mpxy_unlock(&m_qp->rxlock);
 }
 
 /* called with m_qp->rxlock */
@@ -738,19 +793,19 @@ static void m_pi_post_read(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx)
 		l_start = 64;
 	l_end = l_start + l_len;
 
-	if (l_start < smd->m_tl_r && l_end > smd->m_tl_r) {
+	if (pb_full(l_start, l_end, smd->m_hd_r, smd->m_tl_r, l_len, smd->m_len_r)) {
 		if (!(wr_rx->flags & M_READ_PAUSED)) {
 			wr_rx->flags |= M_READ_PAUSED;
 			m_qp->stall_cnt_rr++;
 			MCNTR(smd->md, MCM_MX_RR_STALL);
 
-			mlog(0, " WARN[%d:%d:%d] WR_rx[%d] org_id %Lx RR stall (%d)"
+			mlog(1, " WARN[%d:%d:%d] WR_rx[%d] org_id %Lx RR stall (%d)"
 				" low mem (%p-%p) hd 0x%x tl 0x%x ln %x,%d\n",
 				smd->md->mc->scif_id, smd->entry.tid,
 				m_qp->r_entry.tid, wr_rx->w_idx, wr_rx->org_id,
 				m_qp->stall_cnt_rr, smd->m_buf_r, smd->m_buf_r + smd->m_len_r,
 				smd->m_hd_r, smd->m_tl_r, l_len, l_len);
-			mlog(0, " wr[%d] %p RR(%d,%d,%d): flgs %x tl %d tl_wt %d hd %d\n",
+			mlog(1, " wr[%d] %p RR(%d,%d,%d): flgs %x tl %d tl_wt %d hd %d\n",
 				wr_rx->w_idx, wr_rx, m_qp->post_cnt_rr, m_qp->stall_cnt_rr,
 				m_qp->pi_rr_cnt,  wr_rx->flags,	m_qp->wr_tl_r, m_qp->wr_tl_r_wt,
 				m_qp->wr_hd_r);
@@ -780,12 +835,12 @@ static void m_pi_post_read(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx)
 	if (wr_rx->flags & M_READ_PAUSED) {
 		m_qp->stall_cnt_rr--;
 		wr_rx->flags &= ~M_READ_PAUSED;
-		mlog(0x1, "[%d:%d:%d] WR_rx[%d] RR released (%d) got memory (%p-%p)"
-			" hd 0x%x tl 0x%x ln %x,%d\n",
+		mlog(0x1, "[%d:%d:%d] WR_rx[%d] RR (%d) got memory (%p-%p)"
+			" hd 0x%x tl 0x%x (0x%x-0x%x) ln %x,%d\n",
 			smd->md->mc->scif_id, smd->entry.tid, m_qp->r_entry.tid,
 			wr_rx->w_idx, m_qp->stall_cnt_rr,
 			smd->m_buf_r, smd->m_buf_r + smd->m_len_r,
-			smd->m_hd_r, smd->m_tl_r, l_len, l_len);
+			smd->m_hd_r, smd->m_tl_r, l_start, l_end, l_len, l_len);
 	} else if (m_qp->stall_cnt_rr) {
 		wr_rx->flags |= M_READ_PAUSED;
 		m_qp->stall_cnt_rr++;
@@ -829,10 +884,15 @@ static void m_pi_post_read(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx)
 	    (!((m_qp->post_cnt_rr+1) % mcm_rr_signal))) {
 		ib_wr.send_flags = IBV_SEND_SIGNALED;
 		wr_rx->m_idx = ((rbuf + (l_len - 1)) - smd->m_buf_r);
-		if (m_pi_buf_hd(smd, wr_rx->m_idx, wr_rx))
+		if (m_pi_buf_hd(smd, wr_rx->m_idx, wr_rx)) {
+			mpxy_unlock(&smd->rblock);
 			goto buf_err;
+		}
 	}
-
+#ifdef MCM_PROFILE
+	if (l_end < smd->m_hd_r)
+		smd->m_hd_ro_r++;
+#endif
 	/*
 	 * update shared proxy-in buffer hd, save end of buffer idx
 	 * and save ref m_idx for out of order completions across QP's
@@ -874,6 +934,14 @@ static void m_pi_post_read(struct mcm_qp *m_qp, struct mcm_wr_rx *wr_rx)
 	return;
 bail:
 	mpxy_lock(&smd->rblock);
+	mlog(0x0, "[%d:%d:%d] ERR: WR[%d] %p RR(%d,%d,%d): wr_id %Lx qn %x f=%x,%x ln %d "
+		   "RA %Lx %x to LA %Lx %x tl %d hd %d, m=%x\n",
+		smd->md->mc->scif_id, smd->entry.tid, m_qp->r_entry.tid,
+		wr_rx->w_idx, wr_rx, m_qp->post_cnt_rr, m_qp->stall_cnt_rr,
+		m_qp->pi_rr_cnt, ib_wr.wr_id, ib_qp->qp_num, ib_wr.send_flags,
+		wr_rx->flags, l_len, ib_wr.wr.rdma.remote_addr,
+		ib_wr.wr.rdma.rkey, ib_wr.sg_list->addr, ib_wr.sg_list->lkey,
+		m_qp->wr_tl_r, m_qp->wr_hd_r, wr_rx->m_idx);
 	m_pi_buf_tl(smd, wr_rx->m_idx, wr_rx); /* return buffer slot */
 	mpxy_unlock(&smd->rblock);
 buf_err:
@@ -912,10 +980,15 @@ void m_pi_rcv_event(struct mcm_qp *m_qp, wrc_idata_t *wrc)
 
 		mpxy_lock(&m_qp->rxlock);
 		m_qp->wc_tl_rem = wr_rx->w_idx; /* remote WC tail update in WR */
+#ifdef MCM_PROFILE
+		if (wrc->id == m_qp->wr_hd_r)
+			mlog(0," ERR: RX imm_data: WR id %d duplicate!!!\n", wrc->id);
+#endif
 		m_qp->wr_hd_r = wrc->id; /* new WR took slot, move hd_r */
 		wr_rx->w_idx = wrc->id;   /* my idx slot, to move tl */
 		m_pi_post_read(m_qp, wr_rx);
 		mpxy_unlock(&m_qp->rxlock);
+		write(m_qp->smd->md->mc->tx_pipe[1], "w", sizeof "w");
 
 	} else if (wrc->type == M_WC_TYPE) {
 		struct mcm_wc_rx *m_wc;
@@ -940,12 +1013,12 @@ void m_pi_rcv_event(struct mcm_qp *m_qp, wrc_idata_t *wrc)
 
 /* Proxy-in service - RX thread
  *
- *  <- Work request in (RW_imm - WR idata), remote initiated RW
- *  <- Work completion in (RW_imm - WC idata), local initiated RW
+ *  <- Work request in (RW_imm - WR idata), remote initiated RW, local RR
+ *  <- Work completion in (RW_imm - WC idata), local initiated RW, remote RR
  */
 void m_rcv_event(struct mcm_cq *m_cq, int *events)
 {
-	struct ibv_wc wc[mcm_wrc_max];
+	struct ibv_wc wc[mcm_wc_max/2];
 	struct ibv_cq *ib_cq;
 	struct mcm_qp *m_qp;
 	void *cq_ctx;
@@ -957,14 +1030,12 @@ void m_rcv_event(struct mcm_cq *m_cq, int *events)
 
 	wc_cnt = 0;
 retry:
-	if (wc_cnt >= mcm_wrc_max) {
-		if (wc[0].status == 0)
-			mlog(0x10," m_cq %p processed max %d, exit\n", m_cq, wc_cnt);
+	if (wc_cnt >= mcm_wc_max) {
 		*events += 1;  /* pending */
 		return;
 	}
 
-	ret = ibv_poll_cq(m_cq->ib_cq, mcm_wrc_max, wc);
+	ret = ibv_poll_cq(m_cq->ib_cq, mcm_wc_max/2, wc);
 	if (ret <= 0) {
 		if (!ret && !notify) {
 			ibv_req_notify_cq(m_cq->ib_cq, 0);
@@ -1070,7 +1141,7 @@ void m_pi_pending_wr(struct mcm_qp *m_qp, int *data)
 	}
 
 	mpxy_lock(&m_qp->rxlock);
-	wr_max = mcm_rr_max * 2;
+	wr_max = mcm_wr_max;
 	wr_idx = m_qp->wr_tl_r_wt; /* last write_to marker */
 	wr_cnt = 0;
 
@@ -1099,7 +1170,7 @@ void m_pi_pending_wr(struct mcm_qp *m_qp, int *data)
 			mlog(0x10, " [%d,%d,%d] WR_rx[%d] wr %p scif_wt DONE! flgs  0x%x"
 				   " tl %d w_tl %d hd %d org_id %Lx m_idx %x\n",
 				   smd->md->mc->scif_id, smd->entry.tid, m_qp->r_entry.tid,
-				   wr_rx->w_idx, wr_rx,  wr_rx->flags,
+				   wr_rx->w_idx, wr_rx, wr_rx->flags,
 				   m_qp->wr_tl_r, m_qp->wr_tl_r_wt, m_qp->wr_hd_r,
 				   wr_rx->wr.wr_id, wr_rx->m_idx);
 
@@ -1112,26 +1183,28 @@ void m_pi_pending_wr(struct mcm_qp *m_qp, int *data)
 			m_qp->wr_tl_r_wt = wr_rx->w_idx;  /* writeto pending tail */
 
 			/* if SR operation, send local RX event */
-			if ((wr_rx->wr.opcode & IBV_WR_SEND) ||
-			    (wr_rx->wr.opcode & IBV_WR_RDMA_WRITE_WITH_IMM)) {
+			if ((wr_rx->wr.opcode == IBV_WR_SEND) ||
+			    (wr_rx->wr.opcode == IBV_WR_RDMA_WRITE_WITH_IMM)) {
 				struct dat_mix_wc wc;
 				struct mcm_sr *m_sr;
 
 				m_sr = (struct mcm_sr *)(m_qp->sr_buf + (m_qp->sr_sz * wr_rx->s_idx));
 
-				mlog(4, " SR[%d] rx_event: m_sr %p idx %d wr_idx %d wr_id %Lx imm %x ln %d\n",
-					wr_rx->s_idx,m_sr, m_sr->s_idx, m_sr->w_idx,
-					m_sr->wr_id, wr_rx->wr.imm_data, m_sr->len);
-
 				wc.wr_id = m_sr->wr_id;
 				wc.imm_data = wr_rx->wr.imm_data;
 				wc.byte_len = m_sr->len;
 				wc.status = IBV_WC_SUCCESS;
-				wc.opcode =  wr_rx->wr.opcode & IBV_WR_SEND ?
-					     IBV_WC_RECV:IBV_WC_RECV_RDMA_WITH_IMM;
 				wc.vendor_err = 0;
-				mix_dto_event(m_qp->m_cq_rx, &wc, 1);
+				wc.wc_flags = 0;
+				wc.opcode = (wr_rx->wr.opcode == IBV_WR_SEND) ? IBV_WC_RECV:IBV_WC_RECV_RDMA_WITH_IMM;
 
+				mlog(4, " SR[%d] rx_event: m_sr %p idx %d wr_idx %d"
+					" wr_id %Lx imm %x ln %d wr.op %x wc.op %x\n",
+					wr_rx->s_idx,m_sr, m_sr->s_idx, m_sr->w_idx,
+					m_sr->wr_id, wr_rx->wr.imm_data, m_sr->len,
+					wr_rx->wr.opcode, wc.opcode);
+
+				mix_dto_event(m_qp->m_cq_rx, &wc, 1);
 				m_pi_free_sr(m_qp, m_sr);
 
 			}
@@ -1198,7 +1271,6 @@ done:
 	/* wt pending, stalled RR's, pending RR's, wr queued */
 	*data += ((m_qp->post_cnt_wt || m_qp->stall_cnt_rr || m_qp->pi_rr_cnt || m_qp->wr_tl_r != m_qp->wr_hd_r) ? 1:0);
 	mpxy_unlock(&m_qp->rxlock);
-
 	if (smd->destroy) {
 		mlog(0, "  SMD destroy - QP %p hd %d tl %d, pending data %d\n",
 			m_qp, m_qp->wr_hd_r, m_qp->wr_tl_r, data);
