@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2014 Intel Corporation. All rights reserved.
+ * Copyright (c) 2012-2015 Intel Corporation. All rights reserved.
  *
  * This software is available to you under the OpenIB.org BSD license
  * below:
@@ -35,22 +35,38 @@
 #include "mpxyd.h"
 
 extern int log_level;
+extern int mcm_wc_max;
 extern int mcm_rw_signal;
+extern int mcm_wr_max;
 extern int mcm_ib_inline;
 extern int mix_eager_completion;
 extern int mix_inline_threshold;
 extern uint64_t system_guid;
 
+#ifdef MCM_PROFILE
+extern void mcm_check_io();
+#endif
+
 /* buffer pool for proxy outbount RDMA work request entries, SCIF registration, scif_ep */
 void m_po_destroy_bpool(struct mcm_qp *m_qp)
 {
 	if (m_qp->wr_off)
 		scif_unregister(m_qp->smd->scif_tx_ep, m_qp->wr_off, ALIGN_PAGE(m_qp->wr_len));
 
-	if (m_qp->wr_buf)
+	if (m_qp->wr_buf) {
 		free(m_qp->wr_buf);
+		m_qp->wr_buf = 0;
+	}
 
-	m_qp->wr_buf = 0;
+	if (m_qp->wr_buf_rx_mr) {
+		ibv_dereg_mr(m_qp->wr_buf_rx_mr);
+		m_qp->wr_buf_rx_mr = NULL;
+	}
+
+	if (m_qp->wr_buf_rx) {
+		free(m_qp->wr_buf_rx);
+		m_qp->wr_buf_rx = 0;
+	}
 }
 
 int m_po_create_bpool(struct mcm_qp *m_qp, int entries)
@@ -81,6 +97,29 @@ int m_po_create_bpool(struct mcm_qp *m_qp, int entries)
 	mlog(4, " SCIF_mr for wr_buf addr %p, off 0x%llx, len %d, entries %d\n",
 		m_qp->wr_buf, m_qp->wr_off, ALIGN_PAGE(m_qp->wr_len), entries);
 
+	if (!m_qp->smd->md->indata) {
+		if (posix_memalign((void **)&m_qp->wr_buf_rx, 4096, entries * sizeof(mcm_wr_rx_t))) {
+			mlog(0, "failed to allocate wr_buf_rx, m_qp=%p, wr_rx_len=%d, entries=%d\n",
+				 m_qp, entries * sizeof(mcm_wr_rx_t), entries);
+			goto err;
+		}
+		memset(m_qp->wr_buf, 0, entries * sizeof(mcm_wr_rx_t));
+
+		m_qp->wr_buf_rx_mr = ibv_reg_mr(m_qp->smd->md->pd, (void*)m_qp->wr_buf_rx,
+						entries * sizeof(mcm_wr_rx_t),
+					        IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE);
+		if (!m_qp->wr_buf_rx_mr) {
+			mlog(0, " IB_register addr=%p,%d failed %s\n",
+				m_qp->wr_buf_rx_mr->addr,
+				entries * sizeof(mcm_wr_rx_t),
+				strerror(errno));
+			goto err;
+		}
+		mlog(4, " no inline support: WR_buf_rx pool %p, LEN %d\n",
+			m_qp->wr_buf, entries * sizeof(mcm_wr_rx_t));
+		mlog(4, " no inline support: WR_buf_rx_mr = %x\n", m_qp->wr_buf_rx_mr);
+	}
+
 	return 0;
 err:
 	m_po_destroy_bpool(m_qp);
@@ -106,8 +145,21 @@ int m_po_buf_hd(mcm_scif_dev_t *smd, int m_idx, struct mcm_wr *wr)
 	smd->m_buf_wc[smd->m_buf_hd].m_idx = m_idx;
 	smd->m_buf_wc[smd->m_buf_hd].done = 0;
 #ifdef MCM_PROFILE
+	smd->m_buf_wc[smd->m_buf_hd].hd = smd->m_hd;
+	smd->m_buf_wc[smd->m_buf_hd].tl = smd->m_tl;
+	smd->m_buf_wc[smd->m_buf_hd].ref++;
 	smd->m_buf_wc[smd->m_buf_hd].ts = mcm_ts_us();
 	smd->m_buf_wc[smd->m_buf_hd].wr = (void *) wr;
+	if ((smd->m_hd_ro != smd->m_tl_ro) && (smd->m_hd >= smd->m_tl)) {
+		struct mcm_qp *m_qp = (struct mcm_qp*)wr->context;
+		mlog(0, " WARNING: [%d:%d:%d] PO_buf HD(%d) passed TL(%d): w_tl %d w_hd %d:"
+			" wc_r[%d].m_idx=%x, m_tl %x m_hd %x\n",
+			smd->md->mc->scif_id, smd->entry.tid, m_qp->r_entry.tid,
+			smd->m_hd_ro, smd->m_tl_ro, smd->m_buf_tl + 1,
+			smd->m_buf_hd, smd->m_buf_hd, m_idx,
+			smd->m_tl, smd->m_hd);
+		mcm_check_io();
+	}
 #endif
 	return 0;
 }
@@ -115,11 +167,13 @@ int m_po_buf_hd(mcm_scif_dev_t *smd, int m_idx, struct mcm_wr *wr)
 /* called with smd->tblock */
 static void m_po_buf_tl(mcm_scif_dev_t *smd, int m_idx)
 {
-	int s_idx, idx;
+	int idx;
 	int busy = 0, match = 0, hits = 0;
-
+#ifdef MCM_PROFILE
+	off_t tl_sav = smd->m_tl;
+	uint32_t now = mcm_ts_us();
+#endif
 	idx = (smd->m_buf_tl + 1) & smd->m_buf_end; /* tl == hd is empty */
-	s_idx = idx;
 
 	mlog(0x10," [%d:%d] m_tl 0x%Lx m_hd 0x%Lx - "
 		  "m_wc %p: tl %d hd %d - m_idx=0x%x free\n",
@@ -132,6 +186,10 @@ static void m_po_buf_tl(mcm_scif_dev_t *smd, int m_idx)
 		if (smd->m_buf_wc[idx].m_idx == m_idx) {
 			smd->m_buf_wc[idx].done = 1;
 			match = 1;
+#ifdef MCM_PROFILE
+			smd->m_buf_wc[idx].ref--;
+			smd->m_buf_wc[idx].ts = now - smd->m_buf_wc[idx].ts;
+#endif
 		}
 		if (smd->m_buf_wc[idx].done && !busy) {
 			smd->m_tl = smd->m_buf_wc[idx].m_idx;
@@ -149,59 +207,20 @@ static void m_po_buf_tl(mcm_scif_dev_t *smd, int m_idx)
 	}
 
 #ifdef MCM_PROFILE
-{
-	int match = 0;
-	uint32_t now = mcm_ts_us();
+	if (tl_sav > smd->m_tl)
+		smd->m_tl_ro++;
 
-	if (smd->m_buf_wc[s_idx].done) {
-		mlog(0x10," [%d:%d] InOrder: m_wc %p: tl %d hd %d wc[%d].m_idx=0x%x "
-			  "%s m_idx 0x%x %s wr %p hits %d - %d us\n",
-			  smd->md->mc->scif_id, smd->entry.tid, smd->m_buf_wc,
-			  s_idx, smd->m_buf_hd, s_idx, smd->m_buf_wc[s_idx].m_idx,
-			  smd->m_buf_wc[s_idx].m_idx == m_idx ? "==":"!=",
-			  m_idx, smd->m_buf_wc[s_idx].done ? "DONE":"BUSY",
-			  smd->m_buf_wc[s_idx].wr, hits, now - smd->m_buf_wc[s_idx].ts);
-		return;
-	}
-
-	for (idx = s_idx;;) {
-		if (smd->m_buf_wc[idx].m_idx == m_idx)
-			match++;
-
-		if ((!smd->m_buf_wc[idx].done) || (smd->m_buf_wc[idx].m_idx == m_idx)) {
-			mlog(0x10," [%d:%d] OutOfOrder: m_wc %p: tl %d hd %d wc[%d].m_idx=0x%x"
-				  " %s m_idx 0x%x %s wr %p hits %d - %d us\n",
-				  smd->md->mc->scif_id, smd->entry.tid, smd->m_buf_wc,
-			          s_idx, smd->m_buf_hd, idx, smd->m_buf_wc[idx].m_idx,
-			          smd->m_buf_wc[idx].m_idx == m_idx ? "==":"!=",
-			          m_idx, smd->m_buf_wc[idx].done ? "DONE":"BUSY",
-			          smd->m_buf_wc[idx].wr, hits, now - smd->m_buf_wc[idx].ts);
-		}
-		if (idx == smd->m_buf_hd)
-			break;
-
-		idx = (idx + 1) & smd->m_buf_end;
-	}
 	if (!match) {
-		mlog(0x1," [%d:%d] WARN: m_tl 0x%Lx m_hd 0x%Lx"
-			 "- m_wc %p: tl %d hd %d - m_idx=0x%x not found\n",
-			 smd->md->mc->scif_id, smd->entry.tid,
-			 smd->m_tl, smd->m_hd, smd->m_buf_wc, smd->m_buf_tl,
-			 (smd->m_buf_hd + 1) & smd->m_buf_end, m_idx);
+		mlog(0, " [%d:%d] ERR: m_tl 0x%x m_hd 0x%x"
+			"- m_wc %p: tl %d hd %d - m_idx=0x%x NO MATCH\n",
+			smd->md->mc->scif_id, smd->entry.tid,
+			smd->m_tl, smd->m_hd, smd->m_buf_wc, smd->m_buf_tl,
+			(smd->m_buf_hd + 1) & smd->m_buf_end, m_idx);
+		mcm_check_io();
 	}
-	if (match > 1) {
-		mlog(0x1," [%d:%d] WARN: m_tl 0x%Lx m_hd 0x%Lx"
-			 "- m_wc %p: tl %d hd %d - m_idx=0x%x duplicate\n",
-			 smd->md->mc->scif_id, smd->entry.tid,
-			 smd->m_tl, smd->m_hd, smd->m_buf_wc, smd->m_buf_tl,
-			 (smd->m_buf_hd + 1) & smd->m_buf_end, m_idx);
-	}
-
-}
 #endif
 }
 
-
 /*
  * Proxy-out to Proxy-in - Endpoints are on same platform
  */
@@ -229,6 +248,7 @@ static int m_po_send_wr_local(struct mcm_qp *m_qp, struct mcm_wr *m_wr, int wr_i
  * With inline large enough we don't have to register local side WR data
  *
  * called with m_qp->txlock held
+ *
  */
 static int m_po_send_pi(struct mcm_qp *m_qp, struct mcm_wr *m_wr, int wr_idx)
 {
@@ -237,6 +257,7 @@ static int m_po_send_pi(struct mcm_qp *m_qp, struct mcm_wr *m_wr, int wr_idx)
 	struct ibv_send_wr wr;
 	struct ibv_sge sge;
 	struct mcm_wr_rx wr_rx;
+	struct mcm_wr_rx *wr_rx_ptr;
 	int ret;
 	struct wrc_idata wrc;
 
@@ -244,39 +265,52 @@ static int m_po_send_pi(struct mcm_qp *m_qp, struct mcm_wr *m_wr, int wr_idx)
 	if (m_qp->cm->msg.sys_guid == system_guid)
 		return (m_po_send_wr_local(m_qp, m_wr, wr_idx));
 
+	if (m_qp->smd->md->indata) {
+		wr_rx_ptr = &wr_rx;
+		sge.lkey = 0; /* inline doesn't need registered */
+	} else {
+		wr_rx_ptr = (struct mcm_wr_rx *)
+			    (m_qp->wr_buf_rx + (sizeof(struct mcm_wr_rx) * wr_idx));
+		sge.lkey = m_qp->wr_buf_rx_mr->lkey;
+	}
+	sge.addr = (uint64_t)(uintptr_t) wr_rx_ptr;
+	sge.length = (uint32_t) sizeof(struct mcm_wr_rx);
+
 	/* proxy m_wr over to remote m_wr_rem slot, remote will initiate RR and send back WC */
 	m_wr->flags |= M_SEND_PI;
-	mcm_hton_wr_rx(&wr_rx, m_wr, m_qp->wc_tl); /* build rx_wr for wire transfer, send it */
+	mcm_hton_wr_rx(wr_rx_ptr, m_wr, m_qp->wc_tl); /* build rx_wr for wire transfer, send it */
 
 	wrc.id = (uint16_t)wr_idx;  /* setup imm_data for proxy_in rcv engine */
 	wrc.type = M_WR_TYPE;
 	wrc.flags = 0;
 
+	memset(&wr, 0, sizeof(struct ibv_send_wr));
 	wr.wr_id = WRID_SET(m_wr, WRID_TX_RW_IMM);
-	wr.next = 0;
 	wr.sg_list = &sge;
 	wr.num_sge = 1;
 	wr.opcode = IBV_WR_RDMA_WRITE_WITH_IMM;
 
+	if (m_qp->smd->md->indata)
+		wr.send_flags = IBV_SEND_INLINE;
+
 	if (m_wr->flags & M_SEND_MP_SIG) {
 		m_qp->post_sig_cnt++;  /* sig event pending */
-#if MCM_PROFILE
-		m_wr->wr.wr.atomic.swap = mcm_ts_us();
-		m_wr->wr.wr.atomic.rkey = ((m_wr->w_idx - m_qp->last_wr_pst) & m_qp->wr_end); /* queued */
-		m_qp->last_wr_pst = m_wr->w_idx;
+#if MCM_PROFILE /* MCM_QP_PO_PI_RW */
+		if ((m_wr->flags & M_SEND_FS) && (m_wr->sg[0].length > 1024)) {
+			m_wr->wr.wr.atomic.swap = mcm_ts_us();
+			m_wr->wr.wr.atomic.rkey = ((m_wr->w_idx - m_qp->last_wr_pst) & m_qp->wr_end); /* queued */
+			m_qp->last_wr_pst = m_wr->w_idx;
+		}
 #endif
 	}
 
-	wr.send_flags = m_wr->wr.send_flags | IBV_SEND_INLINE | IBV_SEND_SIGNALED; /* m_wr_rx, 148 bytes */
+	wr.send_flags |= IBV_SEND_SIGNALED;
 	wr.imm_data = htonl(*(uint32_t *)&wrc);
 	wr.wr.rdma.rkey = m_qp->wrc_rem.wr_rkey;
 	wr.wr.rdma.remote_addr =
 		(uint64_t)(uintptr_t)
 		((struct mcm_wr_rx *) (m_qp->wrc_rem.wr_addr + (m_qp->wrc_rem.wr_sz * wr_idx)));
 
-	sge.addr = (uint64_t)(uintptr_t) &wr_rx;
-	sge.length = (uint32_t) sizeof(struct mcm_wr_rx); /* 160 byte WR */
-	sge.lkey = 0; /* inline doesn't need registered */
 
 	/* MXS -> MSS or HST, PI service will be on QP1 */
 	if (MXS_EP(&m_qp->smd->md->addr) &&
@@ -319,7 +353,7 @@ static int m_po_send_pi(struct mcm_qp *m_qp, struct mcm_wr *m_wr, int wr_idx)
 	return 0;
 }
 
-void m_po_pending_wr(struct mcm_qp *m_qp, int *data, int *events)
+void m_po_pending_wr(struct mcm_qp *m_qp, int *data)
 {
 	mcm_scif_dev_t *smd = m_qp->smd;
 	struct mcm_wr *m_wr;
@@ -332,11 +366,11 @@ void m_po_pending_wr(struct mcm_qp *m_qp, int *data, int *events)
 		mpxy_unlock(&m_qp->txlock);
 		return;
 	}
-	wr_max = 40;
+	wr_max = mcm_wr_max;
 	wr_idx = m_qp->wr_tl_rf;
 
 	while (wr_max && m_qp->wr_pp) {
-		cn_signal = 0; poll_cnt = 100;
+		cn_signal = 0; poll_cnt = 50;
 		m_wr = (struct mcm_wr *)(m_qp->wr_buf + (m_qp->wr_sz * wr_idx));
 
 		mlog(0x4, " CHECK: qp %p hd %d tl %d idx %d wr %p wr_id %p,"
@@ -486,9 +520,9 @@ void m_po_pending_wr(struct mcm_qp *m_qp, int *data, int *events)
 		wr_idx = (wr_idx + 1) & m_qp->wr_end; /* next, hd == done */
 
 		if (smd->destroy) {
-			mlog(0, " SMD destroy - QP %p hd %d tl %d pst %d,%d cmp %d, pp %d, events %d\n",
+			mlog(0, " SMD destroy - QP %p hd %d tl %d pst %d,%d cmp %d, pp %d, data %d\n",
 				m_qp, m_qp->wr_hd, m_qp->wr_tl,m_qp->post_cnt,
-				m_qp->post_sig_cnt, m_qp->comp_cnt, m_qp->wr_pp, events);
+				m_qp->post_sig_cnt, m_qp->comp_cnt, m_qp->wr_pp, data);
 			mlog(0, " wr %p wr_id %p org_id %p sglist %p sge %d ln %d op %d flgs"
 				" %x idata 0x%x raddr %p rkey %x m_flgs %x\n",
 			     m_wr, m_wr->wr.wr_id, m_wr->org_id, m_wr->wr.sg_list,
@@ -500,20 +534,18 @@ void m_po_pending_wr(struct mcm_qp *m_qp, int *data, int *events)
 	}
 done:
 	*data += m_qp->wr_pp;
-	*events +=  m_qp->post_sig_cnt - m_qp->comp_cnt;
+	*data +=  m_qp->post_sig_cnt - m_qp->comp_cnt;
 	mpxy_unlock(&m_qp->txlock);
 
 	if (smd->destroy) {
 		mlog(0, " SMD destroy - QP %p hd %d tl %d pst %d,%d cmp %d,"
-			" pending data %d, events %d\n",
+			" pending data %d, data %d\n",
 			m_qp, m_qp->wr_hd, m_qp->wr_tl,m_qp->post_cnt,
-			m_qp->post_sig_cnt, m_qp->comp_cnt, data, events);
+			m_qp->post_sig_cnt, m_qp->comp_cnt, data, data);
 	}
 }
 
-#if MCM_PROFILE
-static uint32_t last_rf = 0;
-#endif
+static uint32_t po_ts = 0;
 
 /* initiate proxy data transfer, operation channel */
 int m_po_proxy_data(mcm_scif_dev_t *smd, dat_mix_sr_t *pmsg, struct mcm_qp *m_qp)
@@ -531,6 +563,14 @@ int m_po_proxy_data(mcm_scif_dev_t *smd, dat_mix_sr_t *pmsg, struct mcm_qp *m_qp
 
 	total_offset = 0;
 
+#if MCM_PROFILE
+	if (!m_qp->last_ps) {
+		m_qp->last_ps = mcm_ts_us();
+	} else {
+		mcm_qp_prof_ts(m_qp, MCM_QP_PO_POST, m_qp->last_ps, 1, 1);
+		m_qp->last_ps = mcm_ts_us();
+	}
+#endif
 	if (pmsg->wr.opcode == IBV_WR_SEND)
 		MCNTR(smd->md, MCM_MX_SEND);
 	else
@@ -541,7 +581,6 @@ int m_po_proxy_data(mcm_scif_dev_t *smd, dat_mix_sr_t *pmsg, struct mcm_qp *m_qp
 	while (((m_qp->wr_hd + 1) & m_qp->wr_end) == m_qp->wr_tl) {
 		if (!retries) {
 			MCNTR(smd->md, MCM_MX_WR_STALL);
-			write(smd->md->mc->tx_pipe[1], "w", sizeof("w"));
 		}
 		if (!(++retries % 100)) {
 			mlog(1, " WARN: DTO delay: no PO WRs: sz %d, hd %d tl %d io %d"
@@ -557,8 +596,9 @@ int m_po_proxy_data(mcm_scif_dev_t *smd, dat_mix_sr_t *pmsg, struct mcm_qp *m_qp
 			wc_err = IBV_WC_RETRY_EXC_ERR;
 			goto bail;
 		}
+		write(smd->md->mc->tx_pipe[1], "w", sizeof("w"));
 		mpxy_unlock(&m_qp->txlock);
-		sleep_usec(10000);
+		sched_yield();
 		mpxy_lock(&m_qp->txlock);
 	}
 	if (retries) {
@@ -611,20 +651,22 @@ retry_mr:
 				l_start = 64;
 			l_end = l_start + seg_len;
 
-			if (l_start < smd->m_tl && l_end > smd->m_tl) {
+			if (pb_full(l_start, l_end, smd->m_hd, smd->m_tl, seg_len, smd->m_len)) {
 				if (!retries) {
+					po_ts = mcm_ts_us();
 					MCNTR(smd->md, MCM_MX_MR_STALL);
 					write(smd->md->mc->tx_pipe[1], "w", sizeof("w"));
+					write(smd->md->mc->rx_pipe[1], "w", sizeof("w"));
 				}
 				if (!(++retries % 100)) {
-					mlog(1, " [%d:%d:%d] WARN: DTO delay, no PO memory,"
-						" %x hd 0x%x tl 0x%x %x,"
+					mlog(1, " [%d:%d:%d] WARN: no PO memory,"
+						" 0x%x hd 0x%x tl 0x%x 0x%x 0x%x,"
 						" need 0x%x-0x%x ln %d %d<-%d,"
 						" retries = %d -> %s\n",
 						m_qp->smd->md->mc->scif_id,
 						m_qp->smd->entry.tid, m_qp->r_entry.tid,
 						smd->m_buf, smd->m_hd, smd->m_tl,
-						smd->m_buf + smd->m_len,
+						smd->m_len, smd->m_buf + smd->m_len,
 						l_start, l_end, seg_len, l_len,
 						pmsg->sge[i].length, retries,
 						mcm_map_str(m_qp->cm->msg.daddr1.ep_map));
@@ -638,27 +680,30 @@ retry_mr:
 				if (retries == 1000) {
 					ret = ENOMEM;
 					wc_err = IBV_WC_RETRY_EXC_ERR;
+					mpxy_unlock(&smd->tblock);
 					goto bail;
 				}
-				mpxy_unlock(&m_qp->txlock);
 				mpxy_unlock(&smd->tblock);
-				sleep_usec(10000);
-				mpxy_lock(&smd->tblock);
+				mpxy_unlock(&m_qp->txlock);
+				sleep_usec(1000);
 				mpxy_lock(&m_qp->txlock);
+				mpxy_lock(&smd->tblock);
 				goto retry_mr;
 			}
 			if (retries) {
 				mlog(1, " MEM stalled: %x hd 0x%x tl 0x%x %x"
-					" got 0x%x-0x%x ln %d %d<-%d retried %d\n",
+					" got 0x%x-0x%x ln %d %d<-%d retried %d %dus\n",
 					smd->m_buf, smd->m_hd, smd->m_tl,
 					smd->m_buf + smd->m_len,
 					l_start, l_end, seg_len, l_len,
-					pmsg->sge[i].length, retries);
+					pmsg->sge[i].length, retries, mcm_ts_us() - po_ts);
 			}
-
+#ifdef MCM_PROFILE
+			if (l_end < smd->m_hd)
+				smd->m_hd_ro++;
+#endif
 			l_off = smd->m_offset + l_start;
 			smd->m_hd = l_end;
-			mpxy_unlock(&smd->tblock);
 
 			mlog(4, " SCIF_readfrom[%d] l_off %p, r_off %p,"
 				" l_start 0x%x l_end 0x%x seg_len %d,"
@@ -670,20 +715,20 @@ retry_mr:
 #if MCM_PROFILE
 			m_wr->wr.wr.atomic.swap = mcm_ts_us();
 			m_wr->wr.wr.atomic.rkey = m_qp->wr_pp;
-			if (last_rf) {
+			if (m_qp->last_rf) {
 				uint32_t now = mcm_ts_us();
-				if ((now - last_rf) > 100000) {
-					mlog(0x4, " WARN: delayed post (%d us):"
+				if ((now - m_qp->last_rf) > 100000) {
+					mlog(0x1, " WARN: delayed post (%d us):"
 						   " WR[%d] hd %d tl %d io %d"
 						   " pst %d,%d cmp %d wr_pp %d\n",
-						now - last_rf, m_wr->w_idx,
-						m_qp->wr_hd, m_qp->wr_tl,
-						seg_len, m_qp->post_cnt,
-						m_qp->post_sig_cnt,
+						now - m_qp->last_rf,
+						m_wr->w_idx, m_qp->wr_hd,
+						m_qp->wr_tl, seg_len,
+						m_qp->post_cnt, m_qp->post_sig_cnt,
 						m_qp->comp_cnt, m_qp->wr_pp);
 				}
 			}
-			last_rf = mcm_ts_us();
+			m_qp->last_rf = mcm_ts_us();
 #endif
 			if (seg_len < 256)
 				ret = scif_readfrom(smd->scif_tx_ep, l_off, seg_len, r_off, SCIF_RMA_USECPU);
@@ -692,6 +737,7 @@ retry_mr:
 
 			if (ret) {
 				mlog(0, " ERR: scif_readfrom, ret %d\n", ret);
+				mpxy_unlock(&smd->tblock);
 				goto bail;
 			}
 			MCNTR(smd->md, MCM_SCIF_READ_FROM);
@@ -739,10 +785,10 @@ retry_mr:
 					m_wr->wr.send_flags |= IBV_SEND_SIGNALED;
 					m_wr->flags |= M_SEND_MP_SIG;
 					m_wr->m_idx = (sbuf + (m_wr->wr.sg_list->length - 1)) - smd->m_buf;
-					mpxy_lock(&smd->tblock);
-					if (m_po_buf_hd(smd, m_wr->m_idx, m_wr))
+					if (m_po_buf_hd(smd, m_wr->m_idx, m_wr)) {
+						mpxy_unlock(&smd->tblock);
 						goto bail;
-					mpxy_unlock(&smd->tblock);
+					}
 					mlog(0x10, "[%d:%d:%d] %s_RF_post_sig: qp %p wr %p wr_id %p flgs 0x%x,"
 						" pcnt %d sg_rate %d hd %d tl %d sz %d m_idx %x\n",
 						m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid,
@@ -753,7 +799,7 @@ retry_mr:
 						m_wr->wr.sg_list->length, m_wr->m_idx);
 				}
 
-				if (pmsg->len <= mcm_ib_inline)
+				if ((smd->md->indata) && (pmsg->len <= mcm_ib_inline))
 					m_wr->wr.send_flags |= IBV_SEND_INLINE;
 
 				mlog(4, " SCIF_fence[%d] l_off %p, ln %d,%d,%d wr_id %p"
@@ -767,18 +813,17 @@ retry_mr:
 							SCIF_SIGNAL_LOCAL);
 				if (ret) {
 					mlog(0," ERR: scif_fence_sig, ret %d \n", ret);
+					mpxy_unlock(&smd->tblock);
 					goto bail;
 				}
 				m_qp->wr_pp++;
 				MCNTR(smd->md, MCM_SCIF_SIGNAL);
 				MCNTR(smd->md, MCM_MX_WRITE_SEG);
 
-				mpxy_unlock(&m_qp->txlock);
-				write(smd->md->mc->tx_pipe[1], "w", sizeof("w"));
-				mpxy_lock(&m_qp->txlock);
-
-				if (!len) /* done */
+				if (!len) { /* done */
+					mpxy_unlock(&smd->tblock);
 					break;
+				}
 
 				/* get next WR */
 				retries = 0;
@@ -786,6 +831,7 @@ retry_mr:
 					if (!retries) {
 						MCNTR(smd->md, MCM_MX_WR_STALL);
 						write(smd->md->mc->tx_pipe[1], "w", sizeof("w"));
+						write(smd->md->mc->rx_pipe[1], "w", sizeof("w"));
 					}
 					if (!(++retries % 100)) {
 						mlog(1, " WARN: DTO delay: no PO WR: sz %d, hd %d tl %d io %d"
@@ -799,11 +845,14 @@ retry_mr:
 					if (retries == 1000) {
 						ret = ENOMEM;
 						wc_err = IBV_WC_RETRY_EXC_ERR;
+						mpxy_unlock(&smd->tblock);
 						goto bail;
 					}
+					mpxy_unlock(&smd->tblock);
 					mpxy_unlock(&m_qp->txlock);
-					sleep_usec(10000);
+					sleep_usec(1000);
 					mpxy_lock(&m_qp->txlock);
+					mpxy_lock(&smd->tblock);
 				}
 				if (retries) {
 					mlog(1, " WR stalled: sz %d, hd %d tl %d io %d"
@@ -813,8 +862,10 @@ retry_mr:
 						m_qp->post_sig_cnt, m_qp->comp_cnt, m_qp->wr_pp);
 				}
 
-				if (m_wr->flags & M_SEND_LS)
+				if (m_wr->flags & M_SEND_LS) {
+					mpxy_unlock(&smd->tblock);
 					goto bail;
+				}
 
 				/* prepare the next WR */
 				m_qp->wr_hd = (m_qp->wr_hd + 1) & m_qp->wr_end; /* move hd */
@@ -828,10 +879,21 @@ retry_mr:
 				const_ib_rw(&m_wr->wr, &pmsg->wr, m_sge);
 				m_wr->wr.wr.rdma.remote_addr += total_offset;
 			}
+			mpxy_unlock(&smd->tblock);
+
+			if (!(m_qp->wr_hd % mcm_wr_max)) {
+				write(smd->md->mc->tx_pipe[1], "w", sizeof("w"));
+				write(smd->md->mc->rx_pipe[1], "w", sizeof("w"));
+				mpxy_unlock(&m_qp->txlock);
+				sched_yield();
+				mpxy_lock(&m_qp->txlock);
+			}
 		}
 	}
 	ret = 0;
 bail:
+	write(smd->md->mc->tx_pipe[1], "w", sizeof("w"));
+	write(smd->md->mc->rx_pipe[1], "w", sizeof("w"));
 	mpxy_unlock(&m_qp->txlock);
 	if (ret) {
 		struct dat_mix_wc wc;
@@ -847,11 +909,12 @@ bail:
 	mlog(4, " exit: q_id %d, q_ctx %p, len %d, wr_hd %d pp %d\n",
 		 pmsg->qp_id, (void*)pmsg->qp_ctx, pmsg->len, m_qp->wr_hd, m_qp->wr_pp);
 
+	sched_yield();
 	return ret;
 }
 
 
-/* work completion from remote proxy-in service */
+/* work completion from remote proxy-in service, RX thread */
 void m_po_wc_event(struct mcm_qp *m_qp, struct mcm_wc_rx *wc_rx, int wc_idx)
 {
 	struct mcm_wr *m_wr;
@@ -859,13 +922,11 @@ void m_po_wc_event(struct mcm_qp *m_qp, struct mcm_wc_rx *wc_rx, int wc_idx)
 	struct dat_mix_wc wc_ev;
 	int event = 0;
 
-	mpxy_lock(&m_qp->txlock);
 	m_wr = (struct mcm_wr *)(m_qp->wr_buf + (m_qp->wr_sz * wc_rx->wr_idx));
 
 	if (wc_rx->wr_idx > m_qp->wr_end) {
 		mlog(0," ERR: WC_rx: WR idx out of range %x > %x \n",
 		     wc_rx->wr_idx, m_qp->wr_end);
-		mpxy_unlock(&m_qp->txlock);
 		return;
 	}
 	m_cq = m_qp->m_cq_rx;
@@ -904,7 +965,7 @@ void m_po_wc_event(struct mcm_qp *m_qp, struct mcm_wc_rx *wc_rx, int wc_idx)
 			event++;
 		}
 	}
-	mlog(0x10," [%d:%d:%d] po_pi_RW DONE!: mb_tl %Lx->%x, m_hd %Lx wr_tl %d->%d wr_id %d wr_hd %d"
+	mlog(0x10," [%d:%d:%d] po_pi_RW DONE!: mb_tl %x->%x, m_hd %x wr_tl %d->%d wr_id %d wr_hd %d"
 		  " wc_tl %d->%d - pst %d,%d cmp %d\n",
 		m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid, m_qp->r_entry.tid,
 		m_qp->smd->m_tl, m_wr->m_idx, m_qp->smd->m_hd, m_qp->wr_tl, wc_rx->wr_tl,
@@ -914,11 +975,13 @@ void m_po_wc_event(struct mcm_qp *m_qp, struct mcm_wc_rx *wc_rx, int wc_idx)
 	/* PI peer has copy of data, free buffer slot if saved on MP signal mark */
 	if (m_wr->flags & M_SEND_MP_SIG) {
 #if MCM_PROFILE
-		mcm_qp_prof_ts( m_qp, MCM_QP_PO_PI_RW,
-				m_wr->wr.wr.atomic.swap,
-				m_wr->wr.wr.atomic.rkey,
-				((m_wr->w_idx - m_qp->last_wr_sig) & m_qp->wr_end));
-		m_qp->last_wr_sig = m_wr->w_idx;
+		if ((m_wr->flags & M_SEND_LS) && (m_wr->sg[0].length > 1024)) {
+			mcm_qp_prof_ts( m_qp, MCM_QP_PO_PI_RW,
+					m_wr->wr.wr.atomic.swap,
+					m_wr->wr.wr.atomic.rkey,
+					1);
+			m_qp->last_wr_sig = m_wr->w_idx;
+		}
 #endif
 		m_qp->comp_cnt++;
 		mpxy_lock(&m_qp->smd->tblock);
@@ -926,6 +989,7 @@ void m_po_wc_event(struct mcm_qp *m_qp, struct mcm_wc_rx *wc_rx, int wc_idx)
 		mpxy_unlock(&m_qp->smd->tblock);
 	}
 
+	mpxy_lock(&m_qp->txlock);
 	if (m_qp->wrc_rem.wr_addr)  /* remote MXS: sync PO WR tail with remote PI WR tail */
 		m_qp->wr_tl = wc_rx->wr_tl;
 
@@ -947,22 +1011,28 @@ void m_po_wc_event(struct mcm_qp *m_qp, struct mcm_wc_rx *wc_rx, int wc_idx)
  *
  *
  */
-void m_req_event(struct mcm_cq *m_cq)
+void m_req_event(struct mcm_cq *m_cq, int *events)
 {
 	struct ibv_cq *ib_cq = NULL;
 	struct mcm_qp *m_qp;
 	struct mcm_wr *m_wr;
 	struct mcm_wr_rx *m_wr_rx;
 	void *cq_ctx;
-	int i, ret, num, wr_type, notify = 0;
+	int i, ret, num, wr_type, wc_cnt, notify = 0;
 	struct ibv_wc wc[DAT_MIX_WC_MAX];
 	struct dat_mix_wc wc_ev[DAT_MIX_WC_MAX];
 
-	ibv_get_cq_event(m_cq->ib_ch, &ib_cq, (void *)&cq_ctx);
-	if (ib_cq && (ib_cq != m_cq->ib_cq))
-		mlog(1," WARNING: ib_cq %p != m_cq->ib_cq %p\n", ib_cq, m_cq->ib_cq);
+	ret = ibv_get_cq_event(m_cq->ib_ch, &ib_cq, (void *)&cq_ctx);
+	if (ret == 0)
+		ibv_ack_cq_events(m_cq->ib_cq, 1);
 
+	wc_cnt = 0;
 retry:
+	if (wc_cnt >= mcm_wc_max) {
+		*events += 1;  /* pending */
+		return;
+	}
+
 	ret = ibv_poll_cq(m_cq->ib_cq, DAT_MIX_WC_MAX, wc);
 	if (ret <= 0) {
 		if (!ret && !notify) {
@@ -970,12 +1040,11 @@ retry:
 			notify = 1;
 			goto retry;
 		}
-		if (ib_cq)
-			ibv_ack_cq_events(ib_cq, 1);
 		return;
 	} else
 		notify = 0;
 
+	wc_cnt += ret;
 	num = 0;
 	for (i=0; i<ret; i++) {
 		wr_type = WRID_TYPE(wc[i].wr_id);
@@ -991,7 +1060,6 @@ retry:
 			assert(m_qp);
 			mlog(4," wr_rx_id[%d of %d] wr_rx %p m_qp %p\n",
 				i+1, ret, m_wr_rx, m_qp);
-
 			m_pi_req_event(m_qp, m_wr_rx, &wc[i], wr_type);
 			continue;
 
@@ -1067,7 +1135,7 @@ retry:
 		} else
 			m_qp->wr_pp_rem--;
 
-		mlog(0x10," [%d:%d:%d] mb_tl %Lx hd %Lx: WR tl %d idx %d hd %d: QP pst %d,%d cmp %d - %s\n",
+		mlog(0x10," [%d:%d:%d] mb_tl %x hd %x: WR tl %d idx %d hd %d: QP pst %d,%d cmp %d - %s\n",
 			  m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid, m_qp->r_entry.tid,
 			  m_qp->smd->m_tl, m_qp->smd->m_hd, m_qp->wr_tl, m_wr->w_idx, m_qp->wr_hd,
 			  m_qp->post_cnt, m_qp->post_sig_cnt, m_qp->comp_cnt,
