@@ -34,9 +34,10 @@ short scif_sport = SCIF_OFED_PORT_8;
 int scif_listen_qlen = 240;
 int mix_buffer_mb = 32;
 int mix_buffer_sg = 131072;
-int mix_buffer_sg_po2 = 17;   /* 128 KB */
+int mix_buffer_sg_po2 = DAT_MCM_SEG_PO2;   /* 128 KB */
 int mcm_set_priority = 0; /* set to SCHED_FIFO */
 int mcm_affinity = 1;
+int mcm_op_poll = 0;
 int mcm_affinity_base_mic = 0;
 int mcm_affinity_base_hca = 0;
 int mcm_counters = 0;
@@ -47,6 +48,7 @@ uint64_t system_guid = 0; /* network order */
 extern int mix_max_msg_mb;
 extern int mcm_tx_entries;
 extern int mcm_rx_entries;
+extern int mcm_ib_inline;
 extern char *lock_file;
 extern char *log_file;
 extern char *opts_file;
@@ -196,6 +198,15 @@ static struct ibv_context *open_ib_device(struct mcm_ib_dev *md, char *name, int
 	}
 
 	for (i=0; i < ibcnt; ++i) {
+		/* system GUID set to first IB device GUID */
+		if (!system_guid && iblist[i]->transport_type == IBV_TRANSPORT_IB) {
+			system_guid = ibv_get_device_guid(iblist[i]);
+			mlog(0, "System GUID == %04x:%04x:%04x:%04x\n",
+				(unsigned) (system_guid >> 48) & 0xffff,
+				(unsigned) (system_guid >> 32) & 0xffff,
+				(unsigned) (system_guid >> 16) & 0xffff,
+				(unsigned) (system_guid >>  0) & 0xffff);
+		}
 		if (!strcmp(iblist[i]->name, name)) {
 			ibctx = ibv_open_device(iblist[i]);
 			if (!ibctx) {
@@ -220,12 +231,8 @@ static struct ibv_context *open_ib_device(struct mcm_ib_dev *md, char *name, int
 					     "%s/device/numa_node unreadable\n",
 					      name, md->ibdev->ibdev_path);
 
-				if (!system_guid) /* first time, set system guid */
-#ifdef MPXYD_LOCAL_SUPPORT
-					system_guid = ibv_get_device_guid(md->ibdev);
-#else
-					system_guid = rand();
-#endif
+				if (mcm_ib_inline_data(ibctx) && mcm_ib_inline)
+					md->indata = 1;
 
 				break;
 			}
@@ -481,8 +488,8 @@ static int create_smd_bpool(mcm_scif_dev_t *smd)
 	mlog(8, " IB registered addr=%p,%d, mr_addr=%p handle=0x%x, lkey=0x%x rkey=0x%x \n",
 	     smd->m_buf, smd->m_len, smd->m_mr->addr, smd->m_mr->handle, smd->m_mr->lkey, smd->m_mr->rkey);
 
-	/* SEND WC queue for buffer management, manage empty slots */
-	wcq_size = (((mix_max_msg_mb*1024*1024)/smd->m_seg) * mcm_tx_entries); /* power of 2 */
+	/* SEND WC queue for buffer management, manage empty slots, power of 2 */
+	wcq_size = (((mix_max_msg_mb*1024*1024)/smd->m_seg) * 8);
 	wcq_entries = 1;
 	while (wcq_entries < wcq_size)
 		wcq_entries <<= 1;
@@ -493,11 +500,13 @@ static int create_smd_bpool(mcm_scif_dev_t *smd)
 		mlog(0, "failed to allocate smd m_bu_wc, m_len=%d, ERR: %d\n", wcq_len, ret);
 		return -1;
 	}
+	memset(smd->m_buf_wc, 0, wcq_len);
+	mlog(0x10, " m_buf_wc %p, len %d, entries %d \n",
+		   smd->m_buf_wc, wcq_len, wcq_entries);
+
 	smd->m_buf_hd = 0;
 	smd->m_buf_tl = 0;
-	smd->m_buf_end = (wcq_len/sizeof(mcm_buf_wc_t)) - 1;
-
-	mlog(0x10, " m_buf_wc %p, len %d, entries %d \n", smd->m_buf_wc, wcq_len, wcq_entries);
+	smd->m_buf_end = wcq_entries - 1;
 
 	/* RECEIVE proxy buffers */
 	smd->m_len_r = ((mix_buffer_mb + 8) * (1024 * 1024));
@@ -531,8 +540,8 @@ static int create_smd_bpool(mcm_scif_dev_t *smd)
 	     smd->m_buf_r, smd->m_len_r, smd->m_mr_r->addr,
 	     smd->m_mr_r->handle, smd->m_mr_r->lkey, smd->m_mr_r->rkey);
 
-	/* RECV WC queue for buffer management, manage empty slots */
-	wcq_size = (((mix_max_msg_mb*1024*1024)/smd->m_seg) * mcm_rx_entries); /* power of 2 */
+	/* RECV WC queue for buffer management, manage empty slots, power of 2 */
+	wcq_size = (((mix_max_msg_mb*1024*1024)/smd->m_seg) * 8);
 	wcq_entries = 1;
 	while (wcq_entries < wcq_size)
 		wcq_entries <<= 1;
@@ -543,12 +552,13 @@ static int create_smd_bpool(mcm_scif_dev_t *smd)
 		mlog(0, "failed to allocate smd m_buf_wc_r, m_len=%d, ERR: %d\n", wcq_len);
 		return -1;
 	}
+	memset(smd->m_buf_wc_r, 0, wcq_len);
 	mlog(0x10, " m_buf_wc_r %p, len %d, entries %d \n",
 		   smd->m_buf_wc_r, wcq_len, wcq_entries);
 
 	smd->m_buf_hd_r = 0;
 	smd->m_buf_tl_r = 0;
-	smd->m_buf_end_r = (wcq_len/sizeof(mcm_buf_wc_t)) - 1;
+	smd->m_buf_end_r = wcq_entries - 1;
 	return 0;
 }
 
@@ -802,32 +812,58 @@ void sig_handler( int signum )
 	finished = 1;
 }
 
-static void mpxy_set_thread_priority()
+static int mpxy_set_thread_attrs(struct mcm_client *mc, cpu_set_t *cpu_mask, int rx)
 {
-	int policy;
+	int policy, cpu = 0;
 	struct sched_param params;
-	pthread_t self = pthread_self();
+	pthread_t self;
 
-	if (!mcm_set_priority)
-		return;
+	if (cpu_mask && mcm_affinity) {
+		mpxy_lock(&mcm_cplock);
+		CPU_ZERO(cpu_mask);
+		if (mcm_affinity_base_mic) {
+			if (mcm_affinity == 2)
+				cpu = mcm_affinity_base_mic;
+			else
+				cpu = mcm_affinity_base_mic++;
+		} else {
+			if (mcm_affinity == 2)
+				cpu = mcm_cpumask[mc->numa_node];
+			else
+				cpu = mcm_cpumask[mc->numa_node]++;
+		}
+		CPU_SET(cpu, cpu_mask);
+		if(sched_setaffinity( 0, sizeof(*cpu_mask), cpu_mask) == -1)
+		      mlog(0, "WARNING: setaffinity (%s) ERR, cpu_id=%d\n",
+			      strerror(errno), cpu);
+		mpxy_unlock(&mcm_cplock);
+	} else {
+		cpu = sched_getcpu();
+	}
 
-	/* set all to max for now */
-	params.sched_priority = sched_get_priority_max(SCHED_FIFO) - 50;
+	if (mcm_set_priority) {
+		self = pthread_self();
+		params.sched_priority = sched_get_priority_max(SCHED_FIFO) - 50;
 
-	if (pthread_setschedparam(self, SCHED_FIFO, &params)) {
-		mlog(0, " ERR: setschedparam returned - %s\n", strerror(errno));
-		return;
+		if (pthread_setschedparam(self, SCHED_FIFO, &params)) {
+			mlog(0, " ERR: setschedparam returned - %s\n",
+				strerror(errno));
+			goto bail;
+		}
+		if (pthread_getschedparam(self, &policy, &params)) {
+			mlog(0, " ERR: getschedparam returned - %s\n",
+				strerror(errno));
+			goto bail;
+		}
+		if (policy != SCHED_FIFO) {
+			mlog(0, " ERR: policy != SCHED_FIFO\n");
+			goto bail;
+		}
+		mlog(0, " Thread (%x) policy = SCHED_FIFO, priority = %d\n",
+			pthread_self(), params.sched_priority);
 	}
-	if (pthread_getschedparam(self, &policy, &params)) {
-		mlog(0, " ERR: getschedparam returned - %s\n", strerror(errno));
-		return;
-	}
-	if (policy != SCHED_FIFO) {
-		mlog(0, " ERR: policy != SCHED_FIFO\n");
-		return;
-	}
-	mlog(0, " Thread (%x) policy set to SCHED_FIFO, priority = %d\n",
-		pthread_self(), params.sched_priority);
+bail:
+	return cpu;
 }
 
 void mpxy_tx_thread(void *mic_client)
@@ -838,27 +874,13 @@ void mpxy_tx_thread(void *mic_client)
 	struct mcm_cq *m_cq;
 	struct mcm_qp *m_qp;
 	struct mcm_fd_set *set;
-	int i, time_ms, data, events, cpu_id;
+	int i, time_ms, data, cpu_id;
 	char rbuf[2];
 
-	if (mcm_affinity) {
-		mpxy_lock(&mcm_cplock);
-		CPU_ZERO( &mc->op_mask );
-		if (mcm_affinity_base_mic) /* static config settings */
-			cpu_id = mcm_affinity_base_mic++;
-		else
-			cpu_id = mcm_cpumask[mc->numa_node]++;
+	cpu_id = mpxy_set_thread_attrs(mc, &mc->tx_mask, 0);
 
-		CPU_SET(cpu_id, &mc->op_mask );
-		if(sched_setaffinity( 0, sizeof(mc->op_mask), &mc->op_mask) == -1)
-		      mlog(0, "WARNING: could not set CPU Affinity (%s), continuing...\n", strerror(errno));
-		mpxy_unlock(&mcm_cplock);
-	}
 	mlog(0, "TX thread (%x) MIC node_id %d bound to numa_node %d and cpu_id=%d\n",
- 		pthread_self(), mc->scif_id, mc->numa_node, mcm_affinity ? cpu_id:0 );
-
-	mpxy_set_thread_priority();
-
+ 		pthread_self(), mc->scif_id, mc->numa_node, cpu_id);
 	set = mcm_alloc_fd_set();
 	if (!set)
 		return;
@@ -867,7 +889,7 @@ void mpxy_tx_thread(void *mic_client)
 		mpxy_lock(&mc->txlock);
 		mcm_fd_zero(set);
 		mcm_fd_set(mc->tx_pipe[0], set, POLLIN);
-		data = 0, events = 0;
+		data = 0;
 		for (i=0;i<MCM_IB_MAX;i++) {
 			md = &mc->mdev[i];
 			if (md->ibctx == NULL)
@@ -883,7 +905,7 @@ void mpxy_tx_thread(void *mic_client)
 				mpxy_lock(&smd->cqlock);
 				m_cq = get_head_entry(&smd->cqlist);
 				while (m_cq) {
-					m_req_event(m_cq); /* check completions, PO and PI */
+					m_req_event(m_cq, &data); /* check completions, PO and PI */
 					if (m_cq->ib_ch)
 						mcm_fd_set(m_cq->ib_ch->fd, set, POLLIN);
 					m_cq = get_next_entry(&m_cq->entry, &smd->cqlist);
@@ -894,28 +916,32 @@ void mpxy_tx_thread(void *mic_client)
 				m_qp = get_head_entry(&smd->qptlist);
 				while (m_qp) {
 					if (m_qp->r_entry.tid)
-						m_pi_pending_wc(m_qp, &events);
-					m_po_pending_wr(m_qp, &data, &events); /* proxy-out WR's */
+						m_pi_pending_wc(m_qp, &data);
+					m_po_pending_wr(m_qp, &data); /* proxy-out WR's */
 					m_qp = get_next_entry(&m_qp->t_entry, &smd->qptlist);
 				}
 				mpxy_unlock(&smd->qptlock);
-
+				sched_yield();
 				mpxy_lock(&md->slock);
 				smd->th_ref_cnt--;
 				smd = get_next_entry(&smd->entry, &md->smd_list);
 			}
 			mpxy_unlock(&md->slock);
+			sched_yield();
 		}
-		time_ms = (data || events) ? 0:-1;
+		mc->tx_busy = data;
+		time_ms = (data) ? 0:-1;
 		mpxy_unlock(&mc->txlock);
+		if (time_ms == -1) mlog(0x10," sleep\n");
 		mcm_select(set, time_ms);
+		if (time_ms == -1) mlog(0x10," wake\n");
 		if (mcm_poll(mc->tx_pipe[0], POLLIN) == POLLIN) {
 			int cnt = 0;
 			while (read(mc->tx_pipe[0], rbuf, 1) > 0)
 				cnt++;
 		}
 	}
-	mlog(0, "TX thread exiting\n");
+	mlog(0, "TX thread exiting, cpu=%d\n");
 }
 
 void mpxy_op_thread(void *mic_client)
@@ -925,25 +951,12 @@ void mpxy_op_thread(void *mic_client)
 	struct mcm_ib_dev *md;
 	struct mcm_scif_dev *smd, *next;
 	char rbuf[2];
-	int i, ret, time_ms, cpu_id;
+	int i, ret, time_ms, cpu_id, smd_cnt;
 
-	if (mcm_affinity) {
-		mpxy_lock(&mcm_cplock);
-		CPU_ZERO( &mc->op_mask );
-		if (mcm_affinity_base_mic) /* static config settings */
-			cpu_id = mcm_affinity_base_mic++;
-		else
-			cpu_id = mcm_cpumask[mc->numa_node]++;
+	cpu_id = mpxy_set_thread_attrs(mc, &mc->op_mask, 0);
 
-		CPU_SET(cpu_id, &mc->op_mask );
-		if(sched_setaffinity( 0, sizeof(mc->op_mask), &mc->op_mask) == -1)
-		      mlog(0, "WARNING: could not set CPU Affinity (%s), continuing...\n", strerror(errno));
-		mpxy_unlock(&mcm_cplock);
-	}
 	mlog(0, "OP thread (%x) MIC node_id %d bound to numa_node %d and cpu_id=%d\n",
- 		pthread_self(), mc->scif_id, mc->numa_node, mcm_affinity ? cpu_id:0 );
-
-	mpxy_set_thread_priority();
+ 		pthread_self(), mc->scif_id, mc->numa_node, cpu_id );
 
 	set = mcm_alloc_fd_set();
 	if (!set)
@@ -951,6 +964,7 @@ void mpxy_op_thread(void *mic_client)
 
 	while (!finished) {
 		mpxy_lock(&mc->oplock);
+		smd_cnt = 0;
 		time_ms = -1;
 		mcm_fd_zero(set);
 		mcm_fd_set(mc->op_pipe[0], set, POLLIN);
@@ -965,8 +979,8 @@ void mpxy_op_thread(void *mic_client)
 			mpxy_lock(&md->slock);
 			smd = get_head_entry(&md->smd_list);
 			while (smd && !smd->destroy) {
+				smd_cnt++;
 				smd->th_ref_cnt++;
-				mcm_check_timers(smd, &time_ms);
 				mpxy_unlock(&md->slock);
 
 				ret = mcm_poll(smd->scif_op_ep, POLLIN); /* operations */
@@ -990,11 +1004,14 @@ void mpxy_op_thread(void *mic_client)
 				smd = next;
 			}
 			mpxy_unlock(&md->slock);
+			sched_yield();
 		}
 		mpxy_unlock(&mc->oplock);
-
-		mcm_select(set, time_ms); /* wait, DAPL MCM or SCIF MIX msgs */
-
+		/* data-path, loop if busy or device open & single core */
+		if ((mc->tx_busy || mc->rx_busy) || (smd_cnt && mcm_op_poll))
+			time_ms = 0;
+		mcm_select(set, time_ms);
+		if (time_ms == -1) mlog(0x10," OP wake\n");
 		if (mcm_poll(mc->op_pipe[0], POLLIN) == POLLIN)
 			read(mc->op_pipe[0], rbuf, 2);
 	}
@@ -1008,26 +1025,13 @@ void mpxy_cm_thread(void *mic_client)
 	struct mcm_ib_dev *md;
 	struct mcm_scif_dev *smd, *next;
 	struct pollfd set[MCM_IB_MAX*3];
-	int i, fds, cpu_id, ret, cnt;
+	int i, fds, cpu_id, ret, cnt, time_ms = -1;
 	char rbuf[2];
 
-	if (mcm_affinity) {
-		mpxy_lock(&mcm_cplock);
-		CPU_ZERO( &mc->op_mask );
-		if (mcm_affinity_base_mic) /* static config settings */
-			cpu_id = mcm_affinity_base_mic++;
-		else
-			cpu_id = mcm_cpumask[mc->numa_node]++;
+	cpu_id = mpxy_set_thread_attrs(mc, &mc->cm_mask, 0);
 
-		CPU_SET(cpu_id, &mc->op_mask );
-		if(sched_setaffinity( 0, sizeof(mc->op_mask), &mc->op_mask) == -1)
-		      mlog(0, "WARNING: could not set CPU Affinity (%s), continuing...\n", strerror(errno));
-		mpxy_unlock(&mcm_cplock);
-	}
 	mlog(0, "CM thread (%x) MIC node_id %d bound to numa_node %d and cpu_id=%d\n",
- 		pthread_self(), mc->scif_id, mc->numa_node, mcm_affinity ? cpu_id:0 );
-
-	mpxy_set_thread_priority();
+ 		pthread_self(), mc->scif_id, mc->numa_node, cpu_id);
 
 	while (!finished) {
 		mpxy_lock(&mc->cmlock);
@@ -1051,7 +1055,8 @@ void mpxy_cm_thread(void *mic_client)
 		}
 		mpxy_unlock(&mc->cmlock);
 
-		poll(set, fds, -1);
+		poll(set, fds, time_ms);
+		time_ms = -1;
 
 		mpxy_lock(&mc->cmlock);
 		for (i=0;i<MCM_IB_MAX;i++) {
@@ -1090,6 +1095,14 @@ void mpxy_cm_thread(void *mic_client)
 					mlog(0, " Shutdown MC %p MD %p complete (%d clients)\n", mc, md, cnt);
 				}
 			}
+			mpxy_lock(&md->slock);
+			smd = get_head_entry(&md->smd_list);
+			while (smd && !smd->destroy) {
+				mcm_check_timers(smd, &time_ms);
+				smd = get_next_entry(&smd->entry, &md->smd_list);
+			}
+			mpxy_unlock(&md->slock);
+			sched_yield();
 		}
 		mpxy_unlock(&mc->cmlock);
 	}
@@ -1105,25 +1118,12 @@ void mpxy_rx_thread(void *mic_client)
 	struct mcm_cq *m_cq;
 	struct mcm_fd_set *set;
 	char rbuf[2];
-	int i, data = 0, cpu_id, time_ms;
+	int i, cpu_id, data, time_ms;
 
-	if (mcm_affinity) {
-		mpxy_lock(&mcm_cplock);
-		CPU_ZERO( &mc->op_mask );
-		if (mcm_affinity_base_mic)  /* static config settings */
-			cpu_id = mcm_affinity_base_mic++;
-		else
-			cpu_id = mcm_cpumask[mc->numa_node]++;
+	cpu_id = mpxy_set_thread_attrs(mc, &mc->cm_mask, 1);
 
-		CPU_SET(cpu_id, &mc->op_mask );
-		if(sched_setaffinity( 0, sizeof(mc->op_mask), &mc->op_mask) == -1)
-		      mlog(0, "WARNING: could not set CPU Affinity (%s), continuing...\n", strerror(errno));
-		mpxy_unlock(&mcm_cplock);
-	}
 	mlog(0, "RX thread (%x) MIC node_id %d bound to numa_node %d and cpu_id=%d\n",
- 		pthread_self(), mc->scif_id, mc->numa_node, mcm_affinity ? cpu_id:0 );
-
-	mpxy_set_thread_priority();
+ 		pthread_self(), mc->scif_id, mc->numa_node, cpu_id);
 
 	set = mcm_alloc_fd_set();
 	if (!set)
@@ -1162,16 +1162,20 @@ void mpxy_rx_thread(void *mic_client)
 					m_qp = get_next_entry(&m_qp->r_entry, &smd->qprlist);
 				}
 				mpxy_unlock(&smd->qprlock);
-
+				sched_yield();
 				mpxy_lock(&md->slock);
 				smd->th_ref_cnt--;
 				smd = get_next_entry(&smd->entry, &md->smd_list);
 			}
 			mpxy_unlock(&md->slock);
+			sched_yield();
 		}
+		mc->rx_busy = data;
 		time_ms = data ? 0:-1;
 		mpxy_unlock(&mc->rxlock);
+		if (time_ms == -1) mlog(0x10," RX sleep\n");
 		mcm_select(set, time_ms);
+		if (time_ms == -1) mlog(0x10," RX wake\n");
 		if (mcm_poll(mc->rx_pipe[0], POLLIN) == POLLIN)
 			read(mc->rx_pipe[0], rbuf, 2);
 	}
@@ -1195,7 +1199,7 @@ static void mpxy_server(void)
 {
 	struct pollfd set;
 	struct sigaction act, oldact;
-	int i;
+	int i, cpu_id;
 
 	act.sa_handler = sig_handler;
 	sigemptyset(&act.sa_mask);
@@ -1205,15 +1209,16 @@ static void mpxy_server(void)
 	if (sigaction(SIGTERM, &act, &oldact))
 		mlog(0, "sigaction: %s\n", strerror(errno));
 
-	mpxy_set_thread_priority();
+	cpu_id = mpxy_set_thread_attrs(&mcm_client_list[0], NULL, 0);
 
-	mlog(0, "server started\n");
+	mlog(0, "Server started, cpu_id=%d\n", cpu_id);
 	while (!finished) {
 		set.fd = scif_listen_ep;
 		set.events = POLLIN;
 		set.revents = 0;
+		mlog(0x8, "Server sleep\n");
 		poll(&set, 1, -1); /* sleep */
-
+		mlog(0x8, "Server wake, cpu_id=%d\n");
 		/* process listens */
 		if (mcm_poll(scif_listen_ep, POLLIN) == POLLIN)
 			mix_scif_accept(scif_listen_ep);
@@ -1312,7 +1317,7 @@ int main(int argc, char **argv)
 	logfile = mpxy_open_log();
 	mpxy_log_options();
 
-	mlog(0, "CCL Proxy - SCIF/IB DAPL RDMA Proxy Service, Mix Version %d (Build-%u)\n",
+	mlog(0, "CCL Proxy - SCIF/IB DAPL RDMA Proxy Service, Mix Version %d (Build-%u) v2\n",
 		DAT_MIX_VER, PACKAGE_DATE);
 
 	if (init_scif()) {
@@ -1341,7 +1346,6 @@ int main(int argc, char **argv)
 }
 
 #ifdef MCM_PROFILE
-
 /* Diagnostic helper functions, log client/device/connection states */
 void mcm_qp_log(struct mcm_qp *m_qp, int tx)
 {
@@ -1352,10 +1356,12 @@ void mcm_qp_log(struct mcm_qp *m_qp, int tx)
 		mlog(0, "[%d:%d:%d] PO QPt %p - WR tl %d tl_rf %d hd %d -"
 			" RW pst %d sig %d po_cmp %d, wr_rem %d wr %d - IO %d ACT %d\n",
 			m_qp->smd->md->mc->scif_id, m_qp->smd->entry.tid,
-			m_qp->t_entry.tid, m_qp, m_qp->wr_tl, m_qp->wr_tl_rf, m_qp->wr_hd,
+			m_qp->r_entry.tid, m_qp, m_qp->wr_tl, m_qp->wr_tl_rf, m_qp->wr_hd,
 			m_qp->post_cnt, m_qp->post_sig_cnt, m_qp->comp_cnt,
 			m_qp->wr_pp_rem, m_qp->wr_pp,
 			m_qp->post_cnt, io);
+		if (m_qp->cm)
+			mcm_pr_addrs(0, &m_qp->cm->msg, m_qp->cm->state, 0);
 	} else {
 		io = m_qp->stall_cnt_rr + m_qp->post_cnt_wt + m_qp->pi_rw_cnt;
 		mlog(0,	"[%d:%d:%d] PI QPr %p - WR tl %d tl_wt %d hd %d -"
@@ -1365,9 +1371,9 @@ void mcm_qp_log(struct mcm_qp *m_qp, int tx)
 			m_qp->wr_hd_r, m_qp->post_cnt_rr, m_qp->pi_rr_cnt,
 			m_qp->stall_cnt_rr, m_qp->post_cnt_wt, m_qp->pi_rw_cnt,
 			m_qp->post_cnt_rr, io);
+		if (m_qp->cm)
+			mcm_pr_addrs(0, &m_qp->cm->msg, m_qp->cm->state, 0);
 	}
-	if (m_qp->cm)
-		mcm_pr_addrs(0, &m_qp->cm->msg, m_qp->cm->state, 0);
 }
 
 void mcm_connect_log(struct mcm_scif_dev *smd)
@@ -1375,8 +1381,6 @@ void mcm_connect_log(struct mcm_scif_dev *smd)
 	struct mcm_qp *m_qp_t;
 	struct mcm_qp *m_qp_r;
 
-	mpxy_lock(&smd->qptlock);
-	mpxy_lock(&smd->qprlock);
 	m_qp_t = get_head_entry(&smd->qptlist);
 	m_qp_r = get_head_entry(&smd->qprlist);
 	while (m_qp_t || m_qp_r) {
@@ -1391,8 +1395,6 @@ void mcm_connect_log(struct mcm_scif_dev *smd)
 						&smd->qprlist);
 		}
 	}
-	mpxy_unlock(&smd->qprlock);
-	mpxy_unlock(&smd->qptlock);
 }
 
 void mcm_dat_dev_log(struct mcm_scif_dev *smd)
@@ -1400,13 +1402,14 @@ void mcm_dat_dev_log(struct mcm_scif_dev *smd)
 	int idx;
 	uint32_t now = mcm_ts_us();
 
-	mlog(0, "[%d:%d] SMD %p \n", smd->md->mc->scif_id, smd->entry.tid, smd);
-	mlog(0, "[%d:%d] PO_BUF %p tl 0x%Lx hd 0x%Lx ln %d - WC %p tl %d hd %d ln %d - SEGs %d ACT %u\n",
+	mlog(0, "[%d:%d] PO_BUF %p tl 0x%x hd 0x%x ln %d -"
+		" WC %p tl %d hd %d ln %d - SEGs %d ACT %u\n",
 		smd->md->mc->scif_id, smd->entry.tid, smd->m_buf,
 		smd->m_tl, smd->m_hd, smd->m_len, smd->m_buf_wc,
 		smd->m_buf_tl, smd->m_buf_hd, smd->m_buf_end,
 		smd->m_buf_hd, smd->m_buf_hd - smd->m_buf_tl);
-	mlog(0, "[%d:%d] PI_BUF %p tl 0x%Lx hd 0x%Lx ln %d - WC %p tl %d hd %d ln %d - SEGs %d ACT %u\n",
+	mlog(0, "[%d:%d] PI_BUF %p tl 0x%x hd 0x%x ln %d -"
+		" WC %p tl %d hd %d ln %d - SEGs %d ACT %u\n",
 		smd->md->mc->scif_id, smd->entry.tid, smd->m_buf_r,
 		smd->m_tl_r, smd->m_hd_r, smd->m_len_r, smd->m_buf_wc_r,
  		smd->m_buf_tl_r, smd->m_buf_hd_r, smd->m_buf_end_r,
@@ -1414,19 +1417,59 @@ void mcm_dat_dev_log(struct mcm_scif_dev *smd)
 
 	/* show PO mbuf_wc busy slots */
 	idx = smd->m_buf_tl;
-	while (smd->m_buf_tl != smd->m_buf_hd) {
-		if (smd->m_buf_wc[idx].m_idx) {
-			mlog(0, "[%d:%d] PO: m_wc %p: tl %d hd %d wc[%d].m_idx=0x%x"
-				" %s wr %p %d us\n",
-				smd->md->mc->scif_id, smd->entry.tid, smd->m_buf_wc,
+	while ((smd->m_buf_tl != smd->m_buf_hd) &&
+	       (smd->m_buf_hd - smd->m_buf_tl)) {
+		if ((smd->m_buf_wc[idx].m_idx && !smd->m_buf_wc[idx].done) || 1) {
+			struct mcm_wr *m_wr = NULL;
+			struct mcm_qp *m_qp = NULL;
+			if (smd->m_buf_wc[idx].wr) {
+				m_wr = (struct mcm_wr *)smd->m_buf_wc[idx].wr;
+				m_qp = (struct mcm_qp *)m_wr->context;
+			}
+			mlog(0, "[%d:%d:%d] PO: m_wc - tl %d hd %d wc[%d].m_idx=0x%x %s"
+				" wr[%d] f=%x,m=%x t=%u ref=%d m_tl %x hd %x\n",
+				smd->md->mc->scif_id, smd->entry.tid,
+				m_qp ? m_qp->r_entry.tid:0,
 				smd->m_buf_tl, smd->m_buf_hd, idx,
 				smd->m_buf_wc[idx].m_idx,
 				smd->m_buf_wc[idx].done ? "DONE":"BUSY",
-				smd->m_buf_wc[idx].wr, now - smd->m_buf_wc[idx].ts);
+				m_wr ? m_wr->w_idx:0, m_wr ? m_wr->flags:0,
+				m_wr ? m_wr->m_idx:0, smd->m_buf_wc[idx].done ?
+				smd->m_buf_wc[idx].ts : now - smd->m_buf_wc[idx].ts,
+				smd->m_buf_wc[idx].ref,
+				smd->m_buf_wc[idx].tl, smd->m_buf_wc[idx].hd);
 		}
 		idx = (idx + 1) & smd->m_buf_end;
+		if (idx == (smd->m_buf_hd+2))
+			break;
+	}
 
-		if (idx == smd->m_buf_hd)
+	/* show PI mbuf_wc busy slots, start from tail */
+	idx = smd->m_buf_tl_r;
+	while ((smd->m_buf_tl_r != smd->m_buf_hd_r) &&
+	       (smd->m_buf_hd_r - smd->m_buf_tl_r)) {
+		if (smd->m_buf_wc_r[idx].m_idx || 1) {
+			struct mcm_wr_rx *m_wr = NULL;
+			struct mcm_qp *m_qp = NULL;
+			if (smd->m_buf_wc_r[idx].wr) {
+				m_wr = (struct mcm_wr_rx *)smd->m_buf_wc_r[idx].wr;
+				m_qp = (struct mcm_qp *)m_wr->context;
+			}
+			mlog(0, "[%d:%d:%d] PI: m_wc_r - tl %d hd %d wc[%d].m_idx=0x%x %s "
+				"wr[%d] f=%x,m=%x t=%u ref=%d m_tl %x hd %x\n",
+				smd->md->mc->scif_id, smd->entry.tid,
+				m_qp ? m_qp->r_entry.tid:0,
+				smd->m_buf_tl_r, smd->m_buf_hd_r, idx,
+				smd->m_buf_wc_r[idx].m_idx,
+				smd->m_buf_wc_r[idx].done ? "DONE":"BUSY",
+				m_wr ? m_wr->w_idx:0, m_wr ? m_wr->flags:0,
+				m_wr ? m_wr->m_idx:0, smd->m_buf_wc_r[idx].done ?
+				smd->m_buf_wc_r[idx].ts : now - smd->m_buf_wc_r[idx].ts,
+				smd->m_buf_wc_r[idx].ref,
+				smd->m_buf_wc_r[idx].tl, smd->m_buf_wc_r[idx].hd);
+		}
+		idx = (idx + 1) & smd->m_buf_end_r;
+		if (idx == (smd->m_buf_hd_r+2))
 			break;
 	}
 }
@@ -1441,6 +1484,7 @@ void mcm_ib_dev_log(struct mcm_ib_dev *md)
 		md->addr.ep_map == MIC_SSOCK_DEV ? "MSS":"MXS", md->mc->ver);
 }
 
+static int check_io_run;
 void mcm_check_io()
 {
 	mcm_client_t *mc;
@@ -1448,6 +1492,11 @@ void mcm_check_io()
 	struct mcm_scif_dev *smd;
 	int i, ii;
 
+	if (check_io_run)
+		return;
+	else
+		check_io_run++;
+
 	for (i=0;i<MCM_CLIENT_MAX;i++) {
 		if (mcm_client_list[i].scif_id)
 			mc = &mcm_client_list[i];
@@ -1464,24 +1513,19 @@ void mcm_check_io()
 				continue;
 
 			mcm_ib_dev_log(md); /* ibv_open_device */
-			mpxy_lock(&mc->oplock);
-			mpxy_lock(&mc->cmlock);
-			mpxy_lock(&mc->txlock);
-			mpxy_lock(&mc->rxlock);
 			smd = get_head_entry(&md->smd_list);
 			while (smd && !smd->destroy) {
+				mlog(0, "[%d:%d] SMD %p \n",
+					smd->md->mc->scif_id,
+					smd->entry.tid, smd);
 				mcm_dat_dev_log(smd); /* dat_ia_open */
 				mcm_connect_log(smd); /* dat_connect */
 				smd = get_next_entry(&smd->entry,
 						     &md->smd_list);
 			}
-			mpxy_unlock(&mc->rxlock);
-			mpxy_unlock(&mc->txlock);
-			mpxy_unlock(&mc->cmlock);
-			mpxy_unlock(&mc->oplock);
 		}
 	}
+	assert(0);
 }
-
 #endif
 
