@@ -66,9 +66,7 @@
 #define MIX_MAX DAT_MIX_VER
 /* #define MCM_PROFILE 1 */
 
-
 /* locking */
-
 typedef pthread_mutex_t      mpxy_lock_t;
 #define mpxy_lock_init(x,y)  pthread_mutex_init(x,y)
 #define mpxy_lock_destroy(x) pthread_mutex_destroy(x)
@@ -87,7 +85,7 @@ typedef struct _llist_entry
 } LLIST_ENTRY;
 
 #define MCM_PORT_SPACE 0xffff
-#define MCM_FD_SETSIZE 4096
+#define MCM_FD_SETSIZE 16384
 struct mcm_fd_set {
 	int index;
 	struct pollfd set[MCM_FD_SETSIZE];
@@ -130,6 +128,7 @@ typedef struct mcm_ib_dev {
 	int			rep_time;
 	int			rtu_time;
 	int 			numa_node;
+	int			indata;
 	void			*cntrs;
 
 } mcm_ib_dev_t;
@@ -142,6 +141,9 @@ typedef struct mcm_buf_wc {
 	uint32_t		m_idx;
 	uint32_t		done;
 #ifdef MCM_PROFILE
+	uint32_t		hd;
+	uint32_t		tl;
+	uint32_t		ref;
 	uint32_t		ts;
 	void 			*wr;
 #endif
@@ -154,7 +156,7 @@ enum mcm_prof_type
 	MCM_QP_RF,
 	MCM_QP_IB_RW,
 	MCM_QP_IB_RR,
-	MCM_QP_PI_IO,
+	MCM_QP_PO_POST,
 	MCM_QP_PO_PI_RW,
 	MCM_QP_ALL
 };
@@ -174,7 +176,7 @@ typedef struct mcm_qp_prof {
 	mcm_qp_prof_val_t	rf; /* scif_readfrom  */
 	mcm_qp_prof_val_t	rw; /* IB write */
 	mcm_qp_prof_val_t	rr; /* IB read */
-	mcm_qp_prof_val_t	pi; /* IO proxy-in: FirstSeg to LastSeg */
+	mcm_qp_prof_val_t	ps; /* Mix post_send times */
 	mcm_qp_prof_val_t	po; /* IO proxy-out: FirstSeg to LastSeg */
 } mcm_qp_prof_t;
 
@@ -205,6 +207,8 @@ typedef struct mcm_qp {
 	int 			post_cnt;	/* completion management to avoid WR depletion */
 	int			post_sig_cnt;
 	int			comp_cnt;
+	char			*wr_buf_rx;	/* mcm_wr_rx_t entries, for devices without inline data  */
+	struct ibv_mr		*wr_buf_rx_mr;
 	/* Proxy-in: WR management, remote view from TX side */
 	mcm_wrc_info_t		wrc_rem;	/* WR and WC buffers: remote, in CM req and reply */
 	int			wr_pp_rem;	/* work request pending */
@@ -227,6 +231,8 @@ typedef struct mcm_qp {
 	int			post_cnt_wt;	/* RX WR, proxy-in, scif_writeto post pending */
 	int 			pi_rr_cnt;	/* RX WR, proxy-in, IB Rdma Reads pending */
 	int 			stall_cnt_rr;	/* RX WR, proxy-in, IB Rdma Reads stalled */
+	char			*wc_buf_rx;	/* mcm_wc_rx_t entries, for devices without inline data  */
+	struct ibv_mr		*wc_buf_rx_mr;
 	/* Proxy-in: send/recv message queue, local on RX side */
 	char			*sr_buf;	/* Send-Recv message work request queue */
 	int 			sr_hd;		/* SR WR head */
@@ -239,6 +245,8 @@ typedef struct mcm_qp {
 	mcm_qp_prof_t		ts;
 	uint32_t		last_wr_sig;
 	uint32_t		last_wr_pst;
+	uint32_t		last_rf;
+	uint32_t		last_ps;
 #endif
 
 } mcm_qp_t;
@@ -325,8 +333,8 @@ typedef struct mcm_scif_dev {
 	char			*m_buf;		/* MIC TX proxy buffer, SCIF and IB  */
 	struct ibv_mr		*m_mr;		/* ib registration */
 	off_t 			m_offset;	/* SCIF registration */
-	off_t			m_hd;		/* buffer pool head */
-	off_t			m_tl;		/* buffer pool tail */
+	int			m_hd;		/* buffer pool head */
+	int			m_tl;		/* buffer pool tail */
 	int			m_len;		/* TX buffer size */
 	int			m_seg;		/* segment size, same for TX and RX proxy */
 	struct 	mcm_buf_wc	*m_buf_wc;	/* Proxy Buffer work completion queue */
@@ -338,14 +346,20 @@ typedef struct mcm_scif_dev {
 	char			*m_buf_r;	/* MIC RX proxy buffer, SCIF and IB  */
 	struct ibv_mr		*m_mr_r;	/* Rcv proxy buffer, ib registration */
 	off_t 			m_offset_r;	/* Rcv proxy buffer, SCIF registration */
-	off_t			m_hd_r;		/* Rcv buffer pool head */
-	off_t			m_tl_r;		/* Rcv buffer pool tail */
+	int			m_hd_r;		/* Rcv buffer pool head */
+	int			m_tl_r;		/* Rcv buffer pool tail */
 	int			m_len_r;	/* Rcv proxy buffer size */
 	struct 	mcm_buf_wc	*m_buf_wc_r;	/* Proxy Buffer work completion queue */
 	int			m_buf_tl_r;	/* Proxy Buffer WC queue tl */
 	int			m_buf_hd_r;	/* Proxy Buffer WC queue hd */
 	int			m_buf_end_r;	/* Proxy Buffer WC queue end */
 	char			*cmd_buf;	/* operation command buffer  */
+#ifdef MCM_PROFILE
+	uint16_t		m_hd_ro;	/* HD,TL tracking */
+	uint16_t		m_tl_ro;
+	uint16_t		m_hd_ro_r;
+	uint16_t		m_tl_ro_r;
+#endif
 
 } mcm_scif_dev_t;
 
@@ -360,6 +374,8 @@ typedef struct mcm_client {
 	int tx_pipe[2];
 	int rx_pipe[2];
 	int cm_pipe[2];
+	int tx_busy;
+	int rx_busy;
 	cpu_set_t op_mask;
 	cpu_set_t tx_mask;
 	cpu_set_t rx_mask;
@@ -569,10 +585,10 @@ void md_cntr_log(mcm_ib_dev_t *md, int counter, int reset);
 /* mpxy_out.c, proxy_out services, rdma write, message req side */
 void m_po_destroy_bpool(struct mcm_qp *m_qp);
 int m_po_create_bpool(struct mcm_qp *m_qp, int max_req_wr);
-void m_po_pending_wr(struct mcm_qp *m_qp, int *data, int *events);
+void m_po_pending_wr(struct mcm_qp *m_qp, int *data);
 int m_po_proxy_data(mcm_scif_dev_t *smd, dat_mix_sr_t *pmsg, struct mcm_qp *m_qp);
 void m_po_wc_event(struct mcm_qp *m_qp, struct mcm_wc_rx *wc_rx, int wc_idx);
-void m_req_event(struct mcm_cq *m_cq);
+void m_req_event(struct mcm_cq *m_cq, int *events);
 int m_po_buf_hd(mcm_scif_dev_t *smd, int m_idx, struct mcm_wr *wr);
 
 /* mpxy_in.c, proxy_in services, rdma write, message rcv side */
@@ -669,6 +685,7 @@ static inline void remove_entry(LLIST_ENTRY *entry)
         entry->tid = 0;
 }
 
+extern int mcm_affinity;
 static inline void sleep_usec(int usec)
 {
 	struct timespec sleep, remain;
@@ -676,6 +693,7 @@ static inline void sleep_usec(int usec)
 	sleep.tv_sec = 0;
 	sleep.tv_nsec = usec * 1000;
 	nanosleep(&sleep, &remain);
+	sched_yield();
 }
 
 static inline uint32_t mcm_ts_us(void)
@@ -695,6 +713,16 @@ static inline uint64_t mcm_time_us(void)
 }
 #define mcm_time_ms() (mcm_time_us() / 1000)
 
+static inline int pb_full(int start, int end, int head, int tail, int len, int size)
+{
+	if (((start < tail) && (end >= tail)) ||
+	   (((head - tail) < 0) && (ALIGN_64(head) + len) >= tail) ||
+	   (((ALIGN_64(head) + len) > size) && start >= tail))
+		return 1;
+	else
+		return 0;
+}
+
 static inline void mcm_free_port(uint64_t *p_port, uint16_t port)
 {
 	p_port[port] = 0;
@@ -825,6 +853,45 @@ static inline void mcm_pr_addrs(int lvl, struct dat_mcm_msg *msg, int state, int
 	}
 }
 
+static inline int mcm_ib_inline_data (IN struct ibv_context *ib_ctx)
+{
+	struct ibv_qp_init_attr qp_create;
+	struct ibv_pd *ib_pd = NULL;
+	struct ibv_cq *ib_cq = NULL;
+	struct ibv_qp *ib_qp = NULL;
+	int ret = 0;
+
+	ib_pd = ibv_alloc_pd(ib_ctx);
+	if (!ib_pd)
+		goto bail;
+
+	ib_cq = ibv_create_cq(ib_ctx, 10, ib_ctx, NULL, 0);
+	if (!ib_cq)
+		goto bail;
+
+	memset((void *)&qp_create, 0, sizeof(qp_create));
+	qp_create.qp_type = IBV_QPT_RC;
+	qp_create.send_cq = ib_cq;
+	qp_create.recv_cq = ib_cq;
+	qp_create.cap.max_send_wr = 1;
+	qp_create.cap.max_send_sge = 1;
+	qp_create.cap.max_inline_data = MCM_IB_INLINE;
+	qp_create.qp_context = (void *)ib_ctx;
+
+	ib_qp = ibv_create_qp(ib_pd, &qp_create);
+	if (ib_qp && qp_create.cap.max_inline_data >= MCM_IB_INLINE)
+		ret = 1;
+bail:
+	if (ib_qp)
+		ibv_destroy_qp(ib_qp);
+	if (ib_cq)
+		ibv_destroy_cq(ib_cq);
+	if (ib_pd)
+		ibv_dealloc_pd(ib_pd);
+
+	return ret;
+}
+
 #endif /*  _MPXYD_H_ */
 
 
