@@ -72,6 +72,8 @@ extern int mcm_ib_inline;
 extern int mcm_rw_signal;
 extern int mcm_rr_signal;
 extern int mcm_rr_max;
+extern int mcm_wr_max;
+extern int mcm_wc_max;
 extern int mcm_tx_entries;
 extern int mcm_rx_entries;
 extern int mcm_rx_cq_size;
@@ -85,6 +87,7 @@ extern int mix_buffer_sg_po2;
 extern int mcm_affinity;
 extern int mcm_affinity_base_mic;
 extern int mcm_affinity_base_hca;
+extern int mcm_op_poll;
 extern int mcm_counters;
 
 /* Counters */
@@ -284,11 +287,15 @@ int mcm_select(struct mcm_fd_set *set, int time_ms)
 {
 	int ret, i;
 
+	if (time_ms != -1)
+		sched_yield();
+
 	mlog(0x20, " sleep, fds=%d\n", set->index);
 	ret = poll(set->set, set->index, time_ms);
 	if (ret > 0) {
 		for (i=0;i<set->index;i++) {
-			mlog(0x20, " wakeup, set[%d].fd %d = %d\n", i, set->set[i].fd, set->set[i].revents);
+			mlog(0x20, " wakeup, set[%d].fd %d = %d\n",
+				   i, set->set[i].fd, set->set[i].revents);
 		}
 	}
 	return ret;
@@ -397,20 +404,28 @@ void mpxy_set_options( int debug_mode )
 			mcm_rw_signal = atoi(value);
 		else if (!strcasecmp("mcm_rr_max_pending", opt))
 			mcm_rr_max = atoi(value);
+		else if (!strcasecmp("mcm_wr_max", opt))
+			mcm_wr_max = atoi(value);
+		else if (!strcasecmp("mcm_wc_max", opt))
+			mcm_wc_max = atoi(value);
 		else if (!strcasecmp("mcm_req_timeout_ms", opt))
 			mcm_rep_ms = atoi(value);
 		else if (!strcasecmp("mcm_rep_timeout_ms", opt))
 			mcm_rtu_ms = atoi(value);
 		else if (!strcasecmp("mcm_retry_count", opt))
 			mcm_retry = atoi(value);
-		else if (!strcasecmp("mcm_affinity", opt))
+		else if (!strcasecmp("mcm_affinity", opt)) {
 			mcm_affinity = atoi(value);
-		else if (!strcasecmp("mcm_affinity_base_hca", opt))
+			if (mcm_affinity == 2)
+				mcm_op_poll = 1;
+		} else if (!strcasecmp("mcm_affinity_base_hca", opt))
 			mcm_affinity_base_hca = atoi(value);
 		else if (!strcasecmp("mcm_affinity_base_mic", opt))
 			mcm_affinity_base_mic = atoi(value);
 		else if (!strcasecmp("mcm_ib_inline", opt))
-			mcm_ib_inline = max(atoi(value), MCM_IB_INLINE);
+			mcm_ib_inline = atoi(value);
+			if (mcm_ib_inline)
+				mcm_ib_inline = max(atoi(value), MCM_IB_INLINE);
 		else if (!strcasecmp("mcm_perf_profile", opt))
 			mcm_profile = atoi(value);
 		else if (!strcasecmp("mcm_eager_completion", opt))
@@ -419,6 +434,8 @@ void mpxy_set_options( int debug_mode )
 			mcm_counters = atoi(value);
 		else if (!strcasecmp("mcm_proxy_in", opt))
 			mcm_proxy_in = atoi(value);
+		else if (!strcasecmp("mcm_op_poll", opt))
+			mcm_op_poll = atoi(value);
 		else if (!strcasecmp("proxy_tx_depth", opt))
 		{
 			int tsize = atoi(value); /* power of 2 */
@@ -451,8 +468,8 @@ void mpxy_log_options(void)
 	mlog(0, "SCIF server_port %d%s\n", scif_sport, scif_sport?"":"(auto)");
 	mlog(0, "SCIF listen queue length %d\n", scif_listen_qlen);
 	mlog(0, "CPU affinity enabled %d\n", mcm_affinity);
-	mlog(0, "CPU affinity base core_id for HCA %d\n", mcm_affinity_base_hca);
-	mlog(0, "CPU affinity base core_id for MIC %d\n", mcm_affinity_base_mic);
+	mlog(0, "CPU affinity base core_id for MICs %d\n", mcm_affinity_base_mic);
+	mlog(0, "OP thread polling %s\n", mcm_op_poll ? "enabled":"disabled");
 	mlog(0, "RDMA buffer pool size %d MB\n", mix_buffer_mb);
 	mlog(0, "RDMA buffer segment size %d\n", mix_buffer_sg);
 	mlog(0, "RDMA buffer alignment %d\n", mix_align);
@@ -462,6 +479,8 @@ void mpxy_log_options(void)
 	mlog(0, "RDMA IB inline threshold %d\n", mcm_ib_inline);
 	mlog(0, "RDMA eager completion %d\n", mix_eager_completion);
 	mlog(0, "RDMA proxy-out signal rate %d\n", mcm_rw_signal);
+	mlog(0, "RDMA proxy-out/in max WR processing %d\n", mcm_wr_max);
+	mlog(0, "RDMA proxy-out/in max WC processing %d\n", mcm_wc_max);
 	mlog(0, "RDMA proxy-in %s\n", mcm_proxy_in ? "enabled":"disabled");
 	if (mcm_proxy_in) {
 		mlog(0, "RDMA proxy-in signal rate %d\n", mcm_rr_signal);
@@ -614,14 +633,14 @@ void mcm_qp_prof_pr(struct mcm_qp *m_qp, int type)
 			m_qp->ts.rr.stop - m_qp->ts.rr.start,
 			(m_qp->ts.rr.stop - m_qp->ts.rr.start)/m_qp->ts.rr.cnt);
 
-	if ((m_qp->ts.pi.min) &&
-	    (type == MCM_QP_PI_IO || type == MCM_QP_ALL))
-		mlog(0, " QP (%p) PERF IO: Proxy-in: FS to LS times"
+	if ((m_qp->ts.ps.min) &&
+	    (type == MCM_QP_PO_POST || type == MCM_QP_ALL))
+		mlog(0, " QP (%p) PERF send_post() times:"
 			" (usecs): max %u min %u avg %u ttime %u cnt %u - et %u avg %u\n",
-			m_qp, m_qp->ts.pi.max, m_qp->ts.pi.min,
-			m_qp->ts.pi.avg, m_qp->ts.pi.all, m_qp->ts.pi.cnt,
-			m_qp->ts.pi.stop - m_qp->ts.pi.start,
-			(m_qp->ts.pi.stop - m_qp->ts.pi.start)/m_qp->ts.pi.cnt);
+			m_qp, m_qp->ts.ps.max, m_qp->ts.ps.min,
+			m_qp->ts.ps.avg, m_qp->ts.ps.all, m_qp->ts.ps.cnt,
+			m_qp->ts.ps.stop - m_qp->ts.ps.start,
+			(m_qp->ts.ps.stop - m_qp->ts.ps.start)/m_qp->ts.ps.cnt);
 
 	if ((m_qp->ts.po.min) &&
 	    (type == MCM_QP_PO_PI_RW || type == MCM_QP_ALL))
@@ -688,15 +707,15 @@ void mcm_qp_prof_ts(struct mcm_qp *m_qp, int type, uint32_t start, uint32_t que,
 		end = &m_qp->ts.rr.stop;
 		diff = diff/qcnt;
 		break;
-	case MCM_QP_PI_IO:
-		type_str = "MCM_QP_PI_IO";
-		min = &m_qp->ts.pi.min;
-		max = &m_qp->ts.pi.max;
-		avg = &m_qp->ts.pi.avg;
-		all = &m_qp->ts.pi.all;
-		cnt = &m_qp->ts.pi.cnt;
-		begin = &m_qp->ts.pi.start;
-		end = &m_qp->ts.pi.stop;
+	case MCM_QP_PO_POST:
+		type_str = "MCM_QP_PO_POST";
+		min = &m_qp->ts.ps.min;
+		max = &m_qp->ts.ps.max;
+		avg = &m_qp->ts.ps.avg;
+		all = &m_qp->ts.ps.all;
+		cnt = &m_qp->ts.ps.cnt;
+		begin = &m_qp->ts.ps.start;
+		end = &m_qp->ts.ps.stop;
 		break;
 	case MCM_QP_PO_PI_RW:
 		type_str = "MCM_QP_PO_PI_RW";
@@ -717,7 +736,7 @@ void mcm_qp_prof_ts(struct mcm_qp *m_qp, int type, uint32_t start, uint32_t que,
 
 	diff = diff/ccnt;
 	*cnt += ccnt;
-	*all += stop - start;
+	*all += diff;
 	*end = stop;
 	if (*min == 0 || diff < *min)
 		*min = diff;
@@ -726,13 +745,21 @@ void mcm_qp_prof_ts(struct mcm_qp *m_qp, int type, uint32_t start, uint32_t que,
 	if (*avg == 0)
 		*avg = diff;
 	else
-		*avg = (diff + *avg)/2;
+		*avg = (*all / *cnt);
+
+	if (log_level == 0x10)
+		mlog(0x10, "us(%u-%u=%u,%d,%d) %s: io %d: mx %u mn %u av %u tt %u et %u "
+			   "TX tl %d hd %d RX tl %d w_tl %d hd %d rr %d wt %d\n",
+			start, stop, diff, qcnt, ccnt, type_str, *cnt, *max, *min, *avg, *all,
+			*end - *begin, m_qp->wr_tl, m_qp->wr_hd, m_qp->wr_tl_r,
+			m_qp->wr_tl_r_wt, m_qp->wr_hd_r, m_qp->pi_rr_cnt, m_qp->post_cnt_wt);
+	else if (log_level == 0x1 && !(*cnt % 100))
+		mlog(0x1, "us(%u-%u=%u,%d,%d) %s: io %d: mx %u mn %u av %u tt %u et %u "
+			   "TX tl %d hd %d RX tl %d w_tl %d hd %d rr %d wt %d\n",
+			start, stop, diff, qcnt, ccnt, type_str, *cnt, *max, *min, *avg, *all,
+			*end - *begin, m_qp->wr_tl, m_qp->wr_hd, m_qp->wr_tl_r,
+			m_qp->wr_tl_r_wt, m_qp->wr_hd_r, m_qp->pi_rr_cnt, m_qp->post_cnt_wt);
 
-	mlog(0x10, "us(%u-%u=%u,%d,%d) %s: io %d: mx %u mn %u av %u tt %u et %u "
-		   "TX tl %d hd %d RX tl %d w_tl %d hd %d rr %d wt %d\n",
-		start, stop, diff, qcnt, ccnt, type_str, *cnt, *max, *min, *avg, *all,
-		*end - *begin, m_qp->wr_tl, m_qp->wr_hd, m_qp->wr_tl_r,
-		m_qp->wr_tl_r_wt, m_qp->wr_hd_r, m_qp->pi_rr_cnt, m_qp->post_cnt_wt);
 }
 #endif
 
